// Copyright (C) 2022 by Camden Mannett.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE or copy at https://www.boost.org/LICENSE_1_0.txt)

#pragma once

#include "arg_router/utility/utf8/code_point.hpp"

#include <functional>

namespace arg_router::utility::utf8
{
/** Grapheme cluster break classes, and their values in the encoded code points in
 * grapheme_cluster_break_table.
 *
 * Do not change the order or value as they need to match scripts/unicode_table_generators.py.
 */
enum class grapheme_cluster_break_class : std::uint8_t {
    any,
    CR,
    LF,
    control,
    extend,
    ZWJ,
    regional_indicator,
    prepend,
    spacing_mark,
    L,
    V,
    T,
    LV,
    LVT,
    extended_pictographic,
};

/** Namespace for grapheme cluster and line break rules.
 *
 * These rules define when @em not to break.
 */
namespace no_break_rules
{
/** Rule GB3, do not split between CR and LF.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB3(const std::array<grapheme_cluster_break_class, N>& trailing_window,
                                 grapheme_cluster_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (trailing_window[0] == grapheme_cluster_break_class::CR) &&
           (next_class == grapheme_cluster_break_class::LF);
}

/** Rule GB6, do not break Hangul syllable sequences.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB6(const std::array<grapheme_cluster_break_class, N>& trailing_window,
                                 grapheme_cluster_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (trailing_window[0] == grapheme_cluster_break_class::L) &&
           ((next_class == grapheme_cluster_break_class::L) ||
            (next_class == grapheme_cluster_break_class::V) ||
            (next_class == grapheme_cluster_break_class::LV) ||
            (next_class == grapheme_cluster_break_class::LVT));
};

/** Rule GB7, do not break Hangul syllable sequences.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB7(const std::array<grapheme_cluster_break_class, N>& trailing_window,
                                 grapheme_cluster_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return ((trailing_window[0] == grapheme_cluster_break_class::LV) ||
            (trailing_window[0] == grapheme_cluster_break_class::V)) &&
           ((next_class == grapheme_cluster_break_class::V) ||
            (next_class == grapheme_cluster_break_class::T));
};

/** Rule GB8, do not break Hangul syllable sequences.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB8(const std::array<grapheme_cluster_break_class, N>& trailing_window,
                                 grapheme_cluster_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return ((trailing_window[0] == grapheme_cluster_break_class::LVT) ||
            (trailing_window[0] == grapheme_cluster_break_class::T)) &&
           (next_class == grapheme_cluster_break_class::T);
};

/** Rule GB9, do not break before extending characters or ZWJ.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB9(
    [[maybe_unused]] const std::array<grapheme_cluster_break_class, N>& trailing_window,
    grapheme_cluster_break_class next_class) noexcept
{
    return (next_class == grapheme_cluster_break_class::extend) ||
           (next_class == grapheme_cluster_break_class::ZWJ);
};

/** Rule GB9a, do not break before SpacingMarks, or after Prepend characters.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB9a(
    [[maybe_unused]] const std::array<grapheme_cluster_break_class, N>& trailing_window,
    grapheme_cluster_break_class next_class) noexcept
{
    return next_class == grapheme_cluster_break_class::spacing_mark;
};

/** Rule GB9b, do not break before SpacingMarks, or after Prepend characters.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB9b(
    const std::array<grapheme_cluster_break_class, N>& trailing_window,
    [[maybe_unused]] grapheme_cluster_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return trailing_window[0] == grapheme_cluster_break_class::prepend;
};

/** Rule GB11, do not break within emoji modifier sequences or emoji zwj sequences.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB11(
    const std::array<grapheme_cluster_break_class, N>& trailing_window,
    grapheme_cluster_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (trailing_window[0] == grapheme_cluster_break_class::ZWJ) &&
           (next_class == grapheme_cluster_break_class::extended_pictographic);
};

/** Rule GB12 and GB13, do not break within emoji flag sequences.
 *
 * From https://www.unicode.org/reports/tr29/#Grapheme_Cluster_Boundary_Rules.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool GB12_13(
    const std::array<grapheme_cluster_break_class, N>& trailing_window,
    grapheme_cluster_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");

    // Count the preceding RIs, if there's an odd number - do not break
    auto i = 0u;
    for (; i < trailing_window.size(); ++i) {
        if (trailing_window[i] != grapheme_cluster_break_class::regional_indicator) {
            break;
        }
    }

    return ((i % 2) != 0) && (next_class == grapheme_cluster_break_class::regional_indicator);
};

/** Array of rule function pointers, just to ease iteration.
 *
 * @tparam N Size of trailing window
 */
template <std::size_t N>
constexpr auto grapheme_cluster = std::array{&GB3<N>,
                                             &GB6<N>,
                                             &GB7<N>,
                                             &GB8<N>,
                                             &GB9<N>,
                                             &GB9a<N>,
                                             &GB9b<N>,
                                             &GB11<N>,
                                             &GB12_13<N>};
}  // namespace no_break_rules

/** Grapheme cluster break class table.
 *
 * Each entry is an inclusive range of code points, with each range's break class value encoded as
 * metadata.
 *
 * This table is generated using scripts/unicode_table_generators.py from
 * https://www.unicode.org/Public/UCD/latest/ucd/auxiliary/GraphemeBreakProperty.txt v14.0.0 and
 * https://www.unicode.org/Public/14.0.0/ucd/emoji/emoji-data.txt v14.0.0
 */
constexpr auto grapheme_cluster_break_table = std::array<code_point::range, 1878>{{
    {0x000000, 0x000009, 0x000003}, {0x00000A, 0x00000A, 0x000002}, {0x00000B, 0x00000C, 0x000003},
    {0x00000D, 0x00000D, 0x000001}, {0x00000E, 0x00001F, 0x000003}, {0x00007F, 0x00009F, 0x000003},
    {0x0000A9, 0x0000A9, 0x00000E}, {0x0000AD, 0x0000AD, 0x000003}, {0x0000AE, 0x0000AE, 0x00000E},
    {0x000300, 0x00036F, 0x000004}, {0x000483, 0x000487, 0x000004}, {0x000488, 0x000489, 0x000004},
    {0x000591, 0x0005BD, 0x000004}, {0x0005BF, 0x0005BF, 0x000004}, {0x0005C1, 0x0005C2, 0x000004},
    {0x0005C4, 0x0005C5, 0x000004}, {0x0005C7, 0x0005C7, 0x000004}, {0x000600, 0x000605, 0x000007},
    {0x000610, 0x00061A, 0x000004}, {0x00061C, 0x00061C, 0x000003}, {0x00064B, 0x00065F, 0x000004},
    {0x000670, 0x000670, 0x000004}, {0x0006D6, 0x0006DC, 0x000004}, {0x0006DD, 0x0006DD, 0x000007},
    {0x0006DF, 0x0006E4, 0x000004}, {0x0006E7, 0x0006E8, 0x000004}, {0x0006EA, 0x0006ED, 0x000004},
    {0x00070F, 0x00070F, 0x000007}, {0x000711, 0x000711, 0x000004}, {0x000730, 0x00074A, 0x000004},
    {0x0007A6, 0x0007B0, 0x000004}, {0x0007EB, 0x0007F3, 0x000004}, {0x0007FD, 0x0007FD, 0x000004},
    {0x000816, 0x000819, 0x000004}, {0x00081B, 0x000823, 0x000004}, {0x000825, 0x000827, 0x000004},
    {0x000829, 0x00082D, 0x000004}, {0x000859, 0x00085B, 0x000004}, {0x000890, 0x000891, 0x000007},
    {0x000898, 0x00089F, 0x000004}, {0x0008CA, 0x0008E1, 0x000004}, {0x0008E2, 0x0008E2, 0x000007},
    {0x0008E3, 0x000902, 0x000004}, {0x000903, 0x000903, 0x000008}, {0x00093A, 0x00093A, 0x000004},
    {0x00093B, 0x00093B, 0x000008}, {0x00093C, 0x00093C, 0x000004}, {0x00093E, 0x000940, 0x000008},
    {0x000941, 0x000948, 0x000004}, {0x000949, 0x00094C, 0x000008}, {0x00094D, 0x00094D, 0x000004},
    {0x00094E, 0x00094F, 0x000008}, {0x000951, 0x000957, 0x000004}, {0x000962, 0x000963, 0x000004},
    {0x000981, 0x000981, 0x000004}, {0x000982, 0x000983, 0x000008}, {0x0009BC, 0x0009BC, 0x000004},
    {0x0009BE, 0x0009BE, 0x000004}, {0x0009BF, 0x0009C0, 0x000008}, {0x0009C1, 0x0009C4, 0x000004},
    {0x0009C7, 0x0009C8, 0x000008}, {0x0009CB, 0x0009CC, 0x000008}, {0x0009CD, 0x0009CD, 0x000004},
    {0x0009D7, 0x0009D7, 0x000004}, {0x0009E2, 0x0009E3, 0x000004}, {0x0009FE, 0x0009FE, 0x000004},
    {0x000A01, 0x000A02, 0x000004}, {0x000A03, 0x000A03, 0x000008}, {0x000A3C, 0x000A3C, 0x000004},
    {0x000A3E, 0x000A40, 0x000008}, {0x000A41, 0x000A42, 0x000004}, {0x000A47, 0x000A48, 0x000004},
    {0x000A4B, 0x000A4D, 0x000004}, {0x000A51, 0x000A51, 0x000004}, {0x000A70, 0x000A71, 0x000004},
    {0x000A75, 0x000A75, 0x000004}, {0x000A81, 0x000A82, 0x000004}, {0x000A83, 0x000A83, 0x000008},
    {0x000ABC, 0x000ABC, 0x000004}, {0x000ABE, 0x000AC0, 0x000008}, {0x000AC1, 0x000AC5, 0x000004},
    {0x000AC7, 0x000AC8, 0x000004}, {0x000AC9, 0x000AC9, 0x000008}, {0x000ACB, 0x000ACC, 0x000008},
    {0x000ACD, 0x000ACD, 0x000004}, {0x000AE2, 0x000AE3, 0x000004}, {0x000AFA, 0x000AFF, 0x000004},
    {0x000B01, 0x000B01, 0x000004}, {0x000B02, 0x000B03, 0x000008}, {0x000B3C, 0x000B3C, 0x000004},
    {0x000B3E, 0x000B3E, 0x000004}, {0x000B3F, 0x000B3F, 0x000004}, {0x000B40, 0x000B40, 0x000008},
    {0x000B41, 0x000B44, 0x000004}, {0x000B47, 0x000B48, 0x000008}, {0x000B4B, 0x000B4C, 0x000008},
    {0x000B4D, 0x000B4D, 0x000004}, {0x000B55, 0x000B56, 0x000004}, {0x000B57, 0x000B57, 0x000004},
    {0x000B62, 0x000B63, 0x000004}, {0x000B82, 0x000B82, 0x000004}, {0x000BBE, 0x000BBE, 0x000004},
    {0x000BBF, 0x000BBF, 0x000008}, {0x000BC0, 0x000BC0, 0x000004}, {0x000BC1, 0x000BC2, 0x000008},
    {0x000BC6, 0x000BC8, 0x000008}, {0x000BCA, 0x000BCC, 0x000008}, {0x000BCD, 0x000BCD, 0x000004},
    {0x000BD7, 0x000BD7, 0x000004}, {0x000C00, 0x000C00, 0x000004}, {0x000C01, 0x000C03, 0x000008},
    {0x000C04, 0x000C04, 0x000004}, {0x000C3C, 0x000C3C, 0x000004}, {0x000C3E, 0x000C40, 0x000004},
    {0x000C41, 0x000C44, 0x000008}, {0x000C46, 0x000C48, 0x000004}, {0x000C4A, 0x000C4D, 0x000004},
    {0x000C55, 0x000C56, 0x000004}, {0x000C62, 0x000C63, 0x000004}, {0x000C81, 0x000C81, 0x000004},
    {0x000C82, 0x000C83, 0x000008}, {0x000CBC, 0x000CBC, 0x000004}, {0x000CBE, 0x000CBE, 0x000008},
    {0x000CBF, 0x000CBF, 0x000004}, {0x000CC0, 0x000CC1, 0x000008}, {0x000CC2, 0x000CC2, 0x000004},
    {0x000CC3, 0x000CC4, 0x000008}, {0x000CC6, 0x000CC6, 0x000004}, {0x000CC7, 0x000CC8, 0x000008},
    {0x000CCA, 0x000CCB, 0x000008}, {0x000CCC, 0x000CCD, 0x000004}, {0x000CD5, 0x000CD6, 0x000004},
    {0x000CE2, 0x000CE3, 0x000004}, {0x000D00, 0x000D01, 0x000004}, {0x000D02, 0x000D03, 0x000008},
    {0x000D3B, 0x000D3C, 0x000004}, {0x000D3E, 0x000D3E, 0x000004}, {0x000D3F, 0x000D40, 0x000008},
    {0x000D41, 0x000D44, 0x000004}, {0x000D46, 0x000D48, 0x000008}, {0x000D4A, 0x000D4C, 0x000008},
    {0x000D4D, 0x000D4D, 0x000004}, {0x000D4E, 0x000D4E, 0x000007}, {0x000D57, 0x000D57, 0x000004},
    {0x000D62, 0x000D63, 0x000004}, {0x000D81, 0x000D81, 0x000004}, {0x000D82, 0x000D83, 0x000008},
    {0x000DCA, 0x000DCA, 0x000004}, {0x000DCF, 0x000DCF, 0x000004}, {0x000DD0, 0x000DD1, 0x000008},
    {0x000DD2, 0x000DD4, 0x000004}, {0x000DD6, 0x000DD6, 0x000004}, {0x000DD8, 0x000DDE, 0x000008},
    {0x000DDF, 0x000DDF, 0x000004}, {0x000DF2, 0x000DF3, 0x000008}, {0x000E31, 0x000E31, 0x000004},
    {0x000E33, 0x000E33, 0x000008}, {0x000E34, 0x000E3A, 0x000004}, {0x000E47, 0x000E4E, 0x000004},
    {0x000EB1, 0x000EB1, 0x000004}, {0x000EB3, 0x000EB3, 0x000008}, {0x000EB4, 0x000EBC, 0x000004},
    {0x000EC8, 0x000ECD, 0x000004}, {0x000F18, 0x000F19, 0x000004}, {0x000F35, 0x000F35, 0x000004},
    {0x000F37, 0x000F37, 0x000004}, {0x000F39, 0x000F39, 0x000004}, {0x000F3E, 0x000F3F, 0x000008},
    {0x000F71, 0x000F7E, 0x000004}, {0x000F7F, 0x000F7F, 0x000008}, {0x000F80, 0x000F84, 0x000004},
    {0x000F86, 0x000F87, 0x000004}, {0x000F8D, 0x000F97, 0x000004}, {0x000F99, 0x000FBC, 0x000004},
    {0x000FC6, 0x000FC6, 0x000004}, {0x00102D, 0x001030, 0x000004}, {0x001031, 0x001031, 0x000008},
    {0x001032, 0x001037, 0x000004}, {0x001039, 0x00103A, 0x000004}, {0x00103B, 0x00103C, 0x000008},
    {0x00103D, 0x00103E, 0x000004}, {0x001056, 0x001057, 0x000008}, {0x001058, 0x001059, 0x000004},
    {0x00105E, 0x001060, 0x000004}, {0x001071, 0x001074, 0x000004}, {0x001082, 0x001082, 0x000004},
    {0x001084, 0x001084, 0x000008}, {0x001085, 0x001086, 0x000004}, {0x00108D, 0x00108D, 0x000004},
    {0x00109D, 0x00109D, 0x000004}, {0x001100, 0x00115F, 0x000009}, {0x001160, 0x0011A7, 0x00000A},
    {0x0011A8, 0x0011FF, 0x00000B}, {0x00135D, 0x00135F, 0x000004}, {0x001712, 0x001714, 0x000004},
    {0x001715, 0x001715, 0x000008}, {0x001732, 0x001733, 0x000004}, {0x001734, 0x001734, 0x000008},
    {0x001752, 0x001753, 0x000004}, {0x001772, 0x001773, 0x000004}, {0x0017B4, 0x0017B5, 0x000004},
    {0x0017B6, 0x0017B6, 0x000008}, {0x0017B7, 0x0017BD, 0x000004}, {0x0017BE, 0x0017C5, 0x000008},
    {0x0017C6, 0x0017C6, 0x000004}, {0x0017C7, 0x0017C8, 0x000008}, {0x0017C9, 0x0017D3, 0x000004},
    {0x0017DD, 0x0017DD, 0x000004}, {0x00180B, 0x00180D, 0x000004}, {0x00180E, 0x00180E, 0x000003},
    {0x00180F, 0x00180F, 0x000004}, {0x001885, 0x001886, 0x000004}, {0x0018A9, 0x0018A9, 0x000004},
    {0x001920, 0x001922, 0x000004}, {0x001923, 0x001926, 0x000008}, {0x001927, 0x001928, 0x000004},
    {0x001929, 0x00192B, 0x000008}, {0x001930, 0x001931, 0x000008}, {0x001932, 0x001932, 0x000004},
    {0x001933, 0x001938, 0x000008}, {0x001939, 0x00193B, 0x000004}, {0x001A17, 0x001A18, 0x000004},
    {0x001A19, 0x001A1A, 0x000008}, {0x001A1B, 0x001A1B, 0x000004}, {0x001A55, 0x001A55, 0x000008},
    {0x001A56, 0x001A56, 0x000004}, {0x001A57, 0x001A57, 0x000008}, {0x001A58, 0x001A5E, 0x000004},
    {0x001A60, 0x001A60, 0x000004}, {0x001A62, 0x001A62, 0x000004}, {0x001A65, 0x001A6C, 0x000004},
    {0x001A6D, 0x001A72, 0x000008}, {0x001A73, 0x001A7C, 0x000004}, {0x001A7F, 0x001A7F, 0x000004},
    {0x001AB0, 0x001ABD, 0x000004}, {0x001ABE, 0x001ABE, 0x000004}, {0x001ABF, 0x001ACE, 0x000004},
    {0x001B00, 0x001B03, 0x000004}, {0x001B04, 0x001B04, 0x000008}, {0x001B34, 0x001B34, 0x000004},
    {0x001B35, 0x001B35, 0x000004}, {0x001B36, 0x001B3A, 0x000004}, {0x001B3B, 0x001B3B, 0x000008},
    {0x001B3C, 0x001B3C, 0x000004}, {0x001B3D, 0x001B41, 0x000008}, {0x001B42, 0x001B42, 0x000004},
    {0x001B43, 0x001B44, 0x000008}, {0x001B6B, 0x001B73, 0x000004}, {0x001B80, 0x001B81, 0x000004},
    {0x001B82, 0x001B82, 0x000008}, {0x001BA1, 0x001BA1, 0x000008}, {0x001BA2, 0x001BA5, 0x000004},
    {0x001BA6, 0x001BA7, 0x000008}, {0x001BA8, 0x001BA9, 0x000004}, {0x001BAA, 0x001BAA, 0x000008},
    {0x001BAB, 0x001BAD, 0x000004}, {0x001BE6, 0x001BE6, 0x000004}, {0x001BE7, 0x001BE7, 0x000008},
    {0x001BE8, 0x001BE9, 0x000004}, {0x001BEA, 0x001BEC, 0x000008}, {0x001BED, 0x001BED, 0x000004},
    {0x001BEE, 0x001BEE, 0x000008}, {0x001BEF, 0x001BF1, 0x000004}, {0x001BF2, 0x001BF3, 0x000008},
    {0x001C24, 0x001C2B, 0x000008}, {0x001C2C, 0x001C33, 0x000004}, {0x001C34, 0x001C35, 0x000008},
    {0x001C36, 0x001C37, 0x000004}, {0x001CD0, 0x001CD2, 0x000004}, {0x001CD4, 0x001CE0, 0x000004},
    {0x001CE1, 0x001CE1, 0x000008}, {0x001CE2, 0x001CE8, 0x000004}, {0x001CED, 0x001CED, 0x000004},
    {0x001CF4, 0x001CF4, 0x000004}, {0x001CF7, 0x001CF7, 0x000008}, {0x001CF8, 0x001CF9, 0x000004},
    {0x001DC0, 0x001DFF, 0x000004}, {0x00200B, 0x00200B, 0x000003}, {0x00200C, 0x00200C, 0x000004},
    {0x00200D, 0x00200D, 0x000005}, {0x00200E, 0x00200F, 0x000003}, {0x002028, 0x002028, 0x000003},
    {0x002029, 0x002029, 0x000003}, {0x00202A, 0x00202E, 0x000003}, {0x00203C, 0x00203C, 0x00000E},
    {0x002049, 0x002049, 0x00000E}, {0x002060, 0x002064, 0x000003}, {0x002065, 0x002065, 0x000003},
    {0x002066, 0x00206F, 0x000003}, {0x0020D0, 0x0020DC, 0x000004}, {0x0020DD, 0x0020E0, 0x000004},
    {0x0020E1, 0x0020E1, 0x000004}, {0x0020E2, 0x0020E4, 0x000004}, {0x0020E5, 0x0020F0, 0x000004},
    {0x002122, 0x002122, 0x00000E}, {0x002139, 0x002139, 0x00000E}, {0x002194, 0x002199, 0x00000E},
    {0x0021A9, 0x0021AA, 0x00000E}, {0x00231A, 0x00231B, 0x00000E}, {0x002328, 0x002328, 0x00000E},
    {0x002388, 0x002388, 0x00000E}, {0x0023CF, 0x0023CF, 0x00000E}, {0x0023E9, 0x0023EC, 0x00000E},
    {0x0023ED, 0x0023EE, 0x00000E}, {0x0023EF, 0x0023EF, 0x00000E}, {0x0023F0, 0x0023F0, 0x00000E},
    {0x0023F1, 0x0023F2, 0x00000E}, {0x0023F3, 0x0023F3, 0x00000E}, {0x0023F8, 0x0023FA, 0x00000E},
    {0x0024C2, 0x0024C2, 0x00000E}, {0x0025AA, 0x0025AB, 0x00000E}, {0x0025B6, 0x0025B6, 0x00000E},
    {0x0025C0, 0x0025C0, 0x00000E}, {0x0025FB, 0x0025FE, 0x00000E}, {0x002600, 0x002601, 0x00000E},
    {0x002602, 0x002603, 0x00000E}, {0x002604, 0x002604, 0x00000E}, {0x002605, 0x002605, 0x00000E},
    {0x002607, 0x00260D, 0x00000E}, {0x00260E, 0x00260E, 0x00000E}, {0x00260F, 0x002610, 0x00000E},
    {0x002611, 0x002611, 0x00000E}, {0x002612, 0x002612, 0x00000E}, {0x002614, 0x002615, 0x00000E},
    {0x002616, 0x002617, 0x00000E}, {0x002618, 0x002618, 0x00000E}, {0x002619, 0x00261C, 0x00000E},
    {0x00261D, 0x00261D, 0x00000E}, {0x00261E, 0x00261F, 0x00000E}, {0x002620, 0x002620, 0x00000E},
    {0x002621, 0x002621, 0x00000E}, {0x002622, 0x002623, 0x00000E}, {0x002624, 0x002625, 0x00000E},
    {0x002626, 0x002626, 0x00000E}, {0x002627, 0x002629, 0x00000E}, {0x00262A, 0x00262A, 0x00000E},
    {0x00262B, 0x00262D, 0x00000E}, {0x00262E, 0x00262E, 0x00000E}, {0x00262F, 0x00262F, 0x00000E},
    {0x002630, 0x002637, 0x00000E}, {0x002638, 0x002639, 0x00000E}, {0x00263A, 0x00263A, 0x00000E},
    {0x00263B, 0x00263F, 0x00000E}, {0x002640, 0x002640, 0x00000E}, {0x002641, 0x002641, 0x00000E},
    {0x002642, 0x002642, 0x00000E}, {0x002643, 0x002647, 0x00000E}, {0x002648, 0x002653, 0x00000E},
    {0x002654, 0x00265E, 0x00000E}, {0x00265F, 0x00265F, 0x00000E}, {0x002660, 0x002660, 0x00000E},
    {0x002661, 0x002662, 0x00000E}, {0x002663, 0x002663, 0x00000E}, {0x002664, 0x002664, 0x00000E},
    {0x002665, 0x002666, 0x00000E}, {0x002667, 0x002667, 0x00000E}, {0x002668, 0x002668, 0x00000E},
    {0x002669, 0x00267A, 0x00000E}, {0x00267B, 0x00267B, 0x00000E}, {0x00267C, 0x00267D, 0x00000E},
    {0x00267E, 0x00267E, 0x00000E}, {0x00267F, 0x00267F, 0x00000E}, {0x002680, 0x002685, 0x00000E},
    {0x002690, 0x002691, 0x00000E}, {0x002692, 0x002692, 0x00000E}, {0x002693, 0x002693, 0x00000E},
    {0x002694, 0x002694, 0x00000E}, {0x002695, 0x002695, 0x00000E}, {0x002696, 0x002697, 0x00000E},
    {0x002698, 0x002698, 0x00000E}, {0x002699, 0x002699, 0x00000E}, {0x00269A, 0x00269A, 0x00000E},
    {0x00269B, 0x00269C, 0x00000E}, {0x00269D, 0x00269F, 0x00000E}, {0x0026A0, 0x0026A1, 0x00000E},
    {0x0026A2, 0x0026A6, 0x00000E}, {0x0026A7, 0x0026A7, 0x00000E}, {0x0026A8, 0x0026A9, 0x00000E},
    {0x0026AA, 0x0026AB, 0x00000E}, {0x0026AC, 0x0026AF, 0x00000E}, {0x0026B0, 0x0026B1, 0x00000E},
    {0x0026B2, 0x0026BC, 0x00000E}, {0x0026BD, 0x0026BE, 0x00000E}, {0x0026BF, 0x0026C3, 0x00000E},
    {0x0026C4, 0x0026C5, 0x00000E}, {0x0026C6, 0x0026C7, 0x00000E}, {0x0026C8, 0x0026C8, 0x00000E},
    {0x0026C9, 0x0026CD, 0x00000E}, {0x0026CE, 0x0026CE, 0x00000E}, {0x0026CF, 0x0026CF, 0x00000E},
    {0x0026D0, 0x0026D0, 0x00000E}, {0x0026D1, 0x0026D1, 0x00000E}, {0x0026D2, 0x0026D2, 0x00000E},
    {0x0026D3, 0x0026D3, 0x00000E}, {0x0026D4, 0x0026D4, 0x00000E}, {0x0026D5, 0x0026E8, 0x00000E},
    {0x0026E9, 0x0026E9, 0x00000E}, {0x0026EA, 0x0026EA, 0x00000E}, {0x0026EB, 0x0026EF, 0x00000E},
    {0x0026F0, 0x0026F1, 0x00000E}, {0x0026F2, 0x0026F3, 0x00000E}, {0x0026F4, 0x0026F4, 0x00000E},
    {0x0026F5, 0x0026F5, 0x00000E}, {0x0026F6, 0x0026F6, 0x00000E}, {0x0026F7, 0x0026F9, 0x00000E},
    {0x0026FA, 0x0026FA, 0x00000E}, {0x0026FB, 0x0026FC, 0x00000E}, {0x0026FD, 0x0026FD, 0x00000E},
    {0x0026FE, 0x002701, 0x00000E}, {0x002702, 0x002702, 0x00000E}, {0x002703, 0x002704, 0x00000E},
    {0x002705, 0x002705, 0x00000E}, {0x002708, 0x00270C, 0x00000E}, {0x00270D, 0x00270D, 0x00000E},
    {0x00270E, 0x00270E, 0x00000E}, {0x00270F, 0x00270F, 0x00000E}, {0x002710, 0x002711, 0x00000E},
    {0x002712, 0x002712, 0x00000E}, {0x002714, 0x002714, 0x00000E}, {0x002716, 0x002716, 0x00000E},
    {0x00271D, 0x00271D, 0x00000E}, {0x002721, 0x002721, 0x00000E}, {0x002728, 0x002728, 0x00000E},
    {0x002733, 0x002734, 0x00000E}, {0x002744, 0x002744, 0x00000E}, {0x002747, 0x002747, 0x00000E},
    {0x00274C, 0x00274C, 0x00000E}, {0x00274E, 0x00274E, 0x00000E}, {0x002753, 0x002755, 0x00000E},
    {0x002757, 0x002757, 0x00000E}, {0x002763, 0x002763, 0x00000E}, {0x002764, 0x002764, 0x00000E},
    {0x002765, 0x002767, 0x00000E}, {0x002795, 0x002797, 0x00000E}, {0x0027A1, 0x0027A1, 0x00000E},
    {0x0027B0, 0x0027B0, 0x00000E}, {0x0027BF, 0x0027BF, 0x00000E}, {0x002934, 0x002935, 0x00000E},
    {0x002B05, 0x002B07, 0x00000E}, {0x002B1B, 0x002B1C, 0x00000E}, {0x002B50, 0x002B50, 0x00000E},
    {0x002B55, 0x002B55, 0x00000E}, {0x002CEF, 0x002CF1, 0x000004}, {0x002D7F, 0x002D7F, 0x000004},
    {0x002DE0, 0x002DFF, 0x000004}, {0x00302A, 0x00302D, 0x000004}, {0x00302E, 0x00302F, 0x000004},
    {0x003030, 0x003030, 0x00000E}, {0x00303D, 0x00303D, 0x00000E}, {0x003099, 0x00309A, 0x000004},
    {0x003297, 0x003297, 0x00000E}, {0x003299, 0x003299, 0x00000E}, {0x00A66F, 0x00A66F, 0x000004},
    {0x00A670, 0x00A672, 0x000004}, {0x00A674, 0x00A67D, 0x000004}, {0x00A69E, 0x00A69F, 0x000004},
    {0x00A6F0, 0x00A6F1, 0x000004}, {0x00A802, 0x00A802, 0x000004}, {0x00A806, 0x00A806, 0x000004},
    {0x00A80B, 0x00A80B, 0x000004}, {0x00A823, 0x00A824, 0x000008}, {0x00A825, 0x00A826, 0x000004},
    {0x00A827, 0x00A827, 0x000008}, {0x00A82C, 0x00A82C, 0x000004}, {0x00A880, 0x00A881, 0x000008},
    {0x00A8B4, 0x00A8C3, 0x000008}, {0x00A8C4, 0x00A8C5, 0x000004}, {0x00A8E0, 0x00A8F1, 0x000004},
    {0x00A8FF, 0x00A8FF, 0x000004}, {0x00A926, 0x00A92D, 0x000004}, {0x00A947, 0x00A951, 0x000004},
    {0x00A952, 0x00A953, 0x000008}, {0x00A960, 0x00A97C, 0x000009}, {0x00A980, 0x00A982, 0x000004},
    {0x00A983, 0x00A983, 0x000008}, {0x00A9B3, 0x00A9B3, 0x000004}, {0x00A9B4, 0x00A9B5, 0x000008},
    {0x00A9B6, 0x00A9B9, 0x000004}, {0x00A9BA, 0x00A9BB, 0x000008}, {0x00A9BC, 0x00A9BD, 0x000004},
    {0x00A9BE, 0x00A9C0, 0x000008}, {0x00A9E5, 0x00A9E5, 0x000004}, {0x00AA29, 0x00AA2E, 0x000004},
    {0x00AA2F, 0x00AA30, 0x000008}, {0x00AA31, 0x00AA32, 0x000004}, {0x00AA33, 0x00AA34, 0x000008},
    {0x00AA35, 0x00AA36, 0x000004}, {0x00AA43, 0x00AA43, 0x000004}, {0x00AA4C, 0x00AA4C, 0x000004},
    {0x00AA4D, 0x00AA4D, 0x000008}, {0x00AA7C, 0x00AA7C, 0x000004}, {0x00AAB0, 0x00AAB0, 0x000004},
    {0x00AAB2, 0x00AAB4, 0x000004}, {0x00AAB7, 0x00AAB8, 0x000004}, {0x00AABE, 0x00AABF, 0x000004},
    {0x00AAC1, 0x00AAC1, 0x000004}, {0x00AAEB, 0x00AAEB, 0x000008}, {0x00AAEC, 0x00AAED, 0x000004},
    {0x00AAEE, 0x00AAEF, 0x000008}, {0x00AAF5, 0x00AAF5, 0x000008}, {0x00AAF6, 0x00AAF6, 0x000004},
    {0x00ABE3, 0x00ABE4, 0x000008}, {0x00ABE5, 0x00ABE5, 0x000004}, {0x00ABE6, 0x00ABE7, 0x000008},
    {0x00ABE8, 0x00ABE8, 0x000004}, {0x00ABE9, 0x00ABEA, 0x000008}, {0x00ABEC, 0x00ABEC, 0x000008},
    {0x00ABED, 0x00ABED, 0x000004}, {0x00AC00, 0x00AC00, 0x00000C}, {0x00AC01, 0x00AC1B, 0x00000D},
    {0x00AC1C, 0x00AC1C, 0x00000C}, {0x00AC1D, 0x00AC37, 0x00000D}, {0x00AC38, 0x00AC38, 0x00000C},
    {0x00AC39, 0x00AC53, 0x00000D}, {0x00AC54, 0x00AC54, 0x00000C}, {0x00AC55, 0x00AC6F, 0x00000D},
    {0x00AC70, 0x00AC70, 0x00000C}, {0x00AC71, 0x00AC8B, 0x00000D}, {0x00AC8C, 0x00AC8C, 0x00000C},
    {0x00AC8D, 0x00ACA7, 0x00000D}, {0x00ACA8, 0x00ACA8, 0x00000C}, {0x00ACA9, 0x00ACC3, 0x00000D},
    {0x00ACC4, 0x00ACC4, 0x00000C}, {0x00ACC5, 0x00ACDF, 0x00000D}, {0x00ACE0, 0x00ACE0, 0x00000C},
    {0x00ACE1, 0x00ACFB, 0x00000D}, {0x00ACFC, 0x00ACFC, 0x00000C}, {0x00ACFD, 0x00AD17, 0x00000D},
    {0x00AD18, 0x00AD18, 0x00000C}, {0x00AD19, 0x00AD33, 0x00000D}, {0x00AD34, 0x00AD34, 0x00000C},
    {0x00AD35, 0x00AD4F, 0x00000D}, {0x00AD50, 0x00AD50, 0x00000C}, {0x00AD51, 0x00AD6B, 0x00000D},
    {0x00AD6C, 0x00AD6C, 0x00000C}, {0x00AD6D, 0x00AD87, 0x00000D}, {0x00AD88, 0x00AD88, 0x00000C},
    {0x00AD89, 0x00ADA3, 0x00000D}, {0x00ADA4, 0x00ADA4, 0x00000C}, {0x00ADA5, 0x00ADBF, 0x00000D},
    {0x00ADC0, 0x00ADC0, 0x00000C}, {0x00ADC1, 0x00ADDB, 0x00000D}, {0x00ADDC, 0x00ADDC, 0x00000C},
    {0x00ADDD, 0x00ADF7, 0x00000D}, {0x00ADF8, 0x00ADF8, 0x00000C}, {0x00ADF9, 0x00AE13, 0x00000D},
    {0x00AE14, 0x00AE14, 0x00000C}, {0x00AE15, 0x00AE2F, 0x00000D}, {0x00AE30, 0x00AE30, 0x00000C},
    {0x00AE31, 0x00AE4B, 0x00000D}, {0x00AE4C, 0x00AE4C, 0x00000C}, {0x00AE4D, 0x00AE67, 0x00000D},
    {0x00AE68, 0x00AE68, 0x00000C}, {0x00AE69, 0x00AE83, 0x00000D}, {0x00AE84, 0x00AE84, 0x00000C},
    {0x00AE85, 0x00AE9F, 0x00000D}, {0x00AEA0, 0x00AEA0, 0x00000C}, {0x00AEA1, 0x00AEBB, 0x00000D},
    {0x00AEBC, 0x00AEBC, 0x00000C}, {0x00AEBD, 0x00AED7, 0x00000D}, {0x00AED8, 0x00AED8, 0x00000C},
    {0x00AED9, 0x00AEF3, 0x00000D}, {0x00AEF4, 0x00AEF4, 0x00000C}, {0x00AEF5, 0x00AF0F, 0x00000D},
    {0x00AF10, 0x00AF10, 0x00000C}, {0x00AF11, 0x00AF2B, 0x00000D}, {0x00AF2C, 0x00AF2C, 0x00000C},
    {0x00AF2D, 0x00AF47, 0x00000D}, {0x00AF48, 0x00AF48, 0x00000C}, {0x00AF49, 0x00AF63, 0x00000D},
    {0x00AF64, 0x00AF64, 0x00000C}, {0x00AF65, 0x00AF7F, 0x00000D}, {0x00AF80, 0x00AF80, 0x00000C},
    {0x00AF81, 0x00AF9B, 0x00000D}, {0x00AF9C, 0x00AF9C, 0x00000C}, {0x00AF9D, 0x00AFB7, 0x00000D},
    {0x00AFB8, 0x00AFB8, 0x00000C}, {0x00AFB9, 0x00AFD3, 0x00000D}, {0x00AFD4, 0x00AFD4, 0x00000C},
    {0x00AFD5, 0x00AFEF, 0x00000D}, {0x00AFF0, 0x00AFF0, 0x00000C}, {0x00AFF1, 0x00B00B, 0x00000D},
    {0x00B00C, 0x00B00C, 0x00000C}, {0x00B00D, 0x00B027, 0x00000D}, {0x00B028, 0x00B028, 0x00000C},
    {0x00B029, 0x00B043, 0x00000D}, {0x00B044, 0x00B044, 0x00000C}, {0x00B045, 0x00B05F, 0x00000D},
    {0x00B060, 0x00B060, 0x00000C}, {0x00B061, 0x00B07B, 0x00000D}, {0x00B07C, 0x00B07C, 0x00000C},
    {0x00B07D, 0x00B097, 0x00000D}, {0x00B098, 0x00B098, 0x00000C}, {0x00B099, 0x00B0B3, 0x00000D},
    {0x00B0B4, 0x00B0B4, 0x00000C}, {0x00B0B5, 0x00B0CF, 0x00000D}, {0x00B0D0, 0x00B0D0, 0x00000C},
    {0x00B0D1, 0x00B0EB, 0x00000D}, {0x00B0EC, 0x00B0EC, 0x00000C}, {0x00B0ED, 0x00B107, 0x00000D},
    {0x00B108, 0x00B108, 0x00000C}, {0x00B109, 0x00B123, 0x00000D}, {0x00B124, 0x00B124, 0x00000C},
    {0x00B125, 0x00B13F, 0x00000D}, {0x00B140, 0x00B140, 0x00000C}, {0x00B141, 0x00B15B, 0x00000D},
    {0x00B15C, 0x00B15C, 0x00000C}, {0x00B15D, 0x00B177, 0x00000D}, {0x00B178, 0x00B178, 0x00000C},
    {0x00B179, 0x00B193, 0x00000D}, {0x00B194, 0x00B194, 0x00000C}, {0x00B195, 0x00B1AF, 0x00000D},
    {0x00B1B0, 0x00B1B0, 0x00000C}, {0x00B1B1, 0x00B1CB, 0x00000D}, {0x00B1CC, 0x00B1CC, 0x00000C},
    {0x00B1CD, 0x00B1E7, 0x00000D}, {0x00B1E8, 0x00B1E8, 0x00000C}, {0x00B1E9, 0x00B203, 0x00000D},
    {0x00B204, 0x00B204, 0x00000C}, {0x00B205, 0x00B21F, 0x00000D}, {0x00B220, 0x00B220, 0x00000C},
    {0x00B221, 0x00B23B, 0x00000D}, {0x00B23C, 0x00B23C, 0x00000C}, {0x00B23D, 0x00B257, 0x00000D},
    {0x00B258, 0x00B258, 0x00000C}, {0x00B259, 0x00B273, 0x00000D}, {0x00B274, 0x00B274, 0x00000C},
    {0x00B275, 0x00B28F, 0x00000D}, {0x00B290, 0x00B290, 0x00000C}, {0x00B291, 0x00B2AB, 0x00000D},
    {0x00B2AC, 0x00B2AC, 0x00000C}, {0x00B2AD, 0x00B2C7, 0x00000D}, {0x00B2C8, 0x00B2C8, 0x00000C},
    {0x00B2C9, 0x00B2E3, 0x00000D}, {0x00B2E4, 0x00B2E4, 0x00000C}, {0x00B2E5, 0x00B2FF, 0x00000D},
    {0x00B300, 0x00B300, 0x00000C}, {0x00B301, 0x00B31B, 0x00000D}, {0x00B31C, 0x00B31C, 0x00000C},
    {0x00B31D, 0x00B337, 0x00000D}, {0x00B338, 0x00B338, 0x00000C}, {0x00B339, 0x00B353, 0x00000D},
    {0x00B354, 0x00B354, 0x00000C}, {0x00B355, 0x00B36F, 0x00000D}, {0x00B370, 0x00B370, 0x00000C},
    {0x00B371, 0x00B38B, 0x00000D}, {0x00B38C, 0x00B38C, 0x00000C}, {0x00B38D, 0x00B3A7, 0x00000D},
    {0x00B3A8, 0x00B3A8, 0x00000C}, {0x00B3A9, 0x00B3C3, 0x00000D}, {0x00B3C4, 0x00B3C4, 0x00000C},
    {0x00B3C5, 0x00B3DF, 0x00000D}, {0x00B3E0, 0x00B3E0, 0x00000C}, {0x00B3E1, 0x00B3FB, 0x00000D},
    {0x00B3FC, 0x00B3FC, 0x00000C}, {0x00B3FD, 0x00B417, 0x00000D}, {0x00B418, 0x00B418, 0x00000C},
    {0x00B419, 0x00B433, 0x00000D}, {0x00B434, 0x00B434, 0x00000C}, {0x00B435, 0x00B44F, 0x00000D},
    {0x00B450, 0x00B450, 0x00000C}, {0x00B451, 0x00B46B, 0x00000D}, {0x00B46C, 0x00B46C, 0x00000C},
    {0x00B46D, 0x00B487, 0x00000D}, {0x00B488, 0x00B488, 0x00000C}, {0x00B489, 0x00B4A3, 0x00000D},
    {0x00B4A4, 0x00B4A4, 0x00000C}, {0x00B4A5, 0x00B4BF, 0x00000D}, {0x00B4C0, 0x00B4C0, 0x00000C},
    {0x00B4C1, 0x00B4DB, 0x00000D}, {0x00B4DC, 0x00B4DC, 0x00000C}, {0x00B4DD, 0x00B4F7, 0x00000D},
    {0x00B4F8, 0x00B4F8, 0x00000C}, {0x00B4F9, 0x00B513, 0x00000D}, {0x00B514, 0x00B514, 0x00000C},
    {0x00B515, 0x00B52F, 0x00000D}, {0x00B530, 0x00B530, 0x00000C}, {0x00B531, 0x00B54B, 0x00000D},
    {0x00B54C, 0x00B54C, 0x00000C}, {0x00B54D, 0x00B567, 0x00000D}, {0x00B568, 0x00B568, 0x00000C},
    {0x00B569, 0x00B583, 0x00000D}, {0x00B584, 0x00B584, 0x00000C}, {0x00B585, 0x00B59F, 0x00000D},
    {0x00B5A0, 0x00B5A0, 0x00000C}, {0x00B5A1, 0x00B5BB, 0x00000D}, {0x00B5BC, 0x00B5BC, 0x00000C},
    {0x00B5BD, 0x00B5D7, 0x00000D}, {0x00B5D8, 0x00B5D8, 0x00000C}, {0x00B5D9, 0x00B5F3, 0x00000D},
    {0x00B5F4, 0x00B5F4, 0x00000C}, {0x00B5F5, 0x00B60F, 0x00000D}, {0x00B610, 0x00B610, 0x00000C},
    {0x00B611, 0x00B62B, 0x00000D}, {0x00B62C, 0x00B62C, 0x00000C}, {0x00B62D, 0x00B647, 0x00000D},
    {0x00B648, 0x00B648, 0x00000C}, {0x00B649, 0x00B663, 0x00000D}, {0x00B664, 0x00B664, 0x00000C},
    {0x00B665, 0x00B67F, 0x00000D}, {0x00B680, 0x00B680, 0x00000C}, {0x00B681, 0x00B69B, 0x00000D},
    {0x00B69C, 0x00B69C, 0x00000C}, {0x00B69D, 0x00B6B7, 0x00000D}, {0x00B6B8, 0x00B6B8, 0x00000C},
    {0x00B6B9, 0x00B6D3, 0x00000D}, {0x00B6D4, 0x00B6D4, 0x00000C}, {0x00B6D5, 0x00B6EF, 0x00000D},
    {0x00B6F0, 0x00B6F0, 0x00000C}, {0x00B6F1, 0x00B70B, 0x00000D}, {0x00B70C, 0x00B70C, 0x00000C},
    {0x00B70D, 0x00B727, 0x00000D}, {0x00B728, 0x00B728, 0x00000C}, {0x00B729, 0x00B743, 0x00000D},
    {0x00B744, 0x00B744, 0x00000C}, {0x00B745, 0x00B75F, 0x00000D}, {0x00B760, 0x00B760, 0x00000C},
    {0x00B761, 0x00B77B, 0x00000D}, {0x00B77C, 0x00B77C, 0x00000C}, {0x00B77D, 0x00B797, 0x00000D},
    {0x00B798, 0x00B798, 0x00000C}, {0x00B799, 0x00B7B3, 0x00000D}, {0x00B7B4, 0x00B7B4, 0x00000C},
    {0x00B7B5, 0x00B7CF, 0x00000D}, {0x00B7D0, 0x00B7D0, 0x00000C}, {0x00B7D1, 0x00B7EB, 0x00000D},
    {0x00B7EC, 0x00B7EC, 0x00000C}, {0x00B7ED, 0x00B807, 0x00000D}, {0x00B808, 0x00B808, 0x00000C},
    {0x00B809, 0x00B823, 0x00000D}, {0x00B824, 0x00B824, 0x00000C}, {0x00B825, 0x00B83F, 0x00000D},
    {0x00B840, 0x00B840, 0x00000C}, {0x00B841, 0x00B85B, 0x00000D}, {0x00B85C, 0x00B85C, 0x00000C},
    {0x00B85D, 0x00B877, 0x00000D}, {0x00B878, 0x00B878, 0x00000C}, {0x00B879, 0x00B893, 0x00000D},
    {0x00B894, 0x00B894, 0x00000C}, {0x00B895, 0x00B8AF, 0x00000D}, {0x00B8B0, 0x00B8B0, 0x00000C},
    {0x00B8B1, 0x00B8CB, 0x00000D}, {0x00B8CC, 0x00B8CC, 0x00000C}, {0x00B8CD, 0x00B8E7, 0x00000D},
    {0x00B8E8, 0x00B8E8, 0x00000C}, {0x00B8E9, 0x00B903, 0x00000D}, {0x00B904, 0x00B904, 0x00000C},
    {0x00B905, 0x00B91F, 0x00000D}, {0x00B920, 0x00B920, 0x00000C}, {0x00B921, 0x00B93B, 0x00000D},
    {0x00B93C, 0x00B93C, 0x00000C}, {0x00B93D, 0x00B957, 0x00000D}, {0x00B958, 0x00B958, 0x00000C},
    {0x00B959, 0x00B973, 0x00000D}, {0x00B974, 0x00B974, 0x00000C}, {0x00B975, 0x00B98F, 0x00000D},
    {0x00B990, 0x00B990, 0x00000C}, {0x00B991, 0x00B9AB, 0x00000D}, {0x00B9AC, 0x00B9AC, 0x00000C},
    {0x00B9AD, 0x00B9C7, 0x00000D}, {0x00B9C8, 0x00B9C8, 0x00000C}, {0x00B9C9, 0x00B9E3, 0x00000D},
    {0x00B9E4, 0x00B9E4, 0x00000C}, {0x00B9E5, 0x00B9FF, 0x00000D}, {0x00BA00, 0x00BA00, 0x00000C},
    {0x00BA01, 0x00BA1B, 0x00000D}, {0x00BA1C, 0x00BA1C, 0x00000C}, {0x00BA1D, 0x00BA37, 0x00000D},
    {0x00BA38, 0x00BA38, 0x00000C}, {0x00BA39, 0x00BA53, 0x00000D}, {0x00BA54, 0x00BA54, 0x00000C},
    {0x00BA55, 0x00BA6F, 0x00000D}, {0x00BA70, 0x00BA70, 0x00000C}, {0x00BA71, 0x00BA8B, 0x00000D},
    {0x00BA8C, 0x00BA8C, 0x00000C}, {0x00BA8D, 0x00BAA7, 0x00000D}, {0x00BAA8, 0x00BAA8, 0x00000C},
    {0x00BAA9, 0x00BAC3, 0x00000D}, {0x00BAC4, 0x00BAC4, 0x00000C}, {0x00BAC5, 0x00BADF, 0x00000D},
    {0x00BAE0, 0x00BAE0, 0x00000C}, {0x00BAE1, 0x00BAFB, 0x00000D}, {0x00BAFC, 0x00BAFC, 0x00000C},
    {0x00BAFD, 0x00BB17, 0x00000D}, {0x00BB18, 0x00BB18, 0x00000C}, {0x00BB19, 0x00BB33, 0x00000D},
    {0x00BB34, 0x00BB34, 0x00000C}, {0x00BB35, 0x00BB4F, 0x00000D}, {0x00BB50, 0x00BB50, 0x00000C},
    {0x00BB51, 0x00BB6B, 0x00000D}, {0x00BB6C, 0x00BB6C, 0x00000C}, {0x00BB6D, 0x00BB87, 0x00000D},
    {0x00BB88, 0x00BB88, 0x00000C}, {0x00BB89, 0x00BBA3, 0x00000D}, {0x00BBA4, 0x00BBA4, 0x00000C},
    {0x00BBA5, 0x00BBBF, 0x00000D}, {0x00BBC0, 0x00BBC0, 0x00000C}, {0x00BBC1, 0x00BBDB, 0x00000D},
    {0x00BBDC, 0x00BBDC, 0x00000C}, {0x00BBDD, 0x00BBF7, 0x00000D}, {0x00BBF8, 0x00BBF8, 0x00000C},
    {0x00BBF9, 0x00BC13, 0x00000D}, {0x00BC14, 0x00BC14, 0x00000C}, {0x00BC15, 0x00BC2F, 0x00000D},
    {0x00BC30, 0x00BC30, 0x00000C}, {0x00BC31, 0x00BC4B, 0x00000D}, {0x00BC4C, 0x00BC4C, 0x00000C},
    {0x00BC4D, 0x00BC67, 0x00000D}, {0x00BC68, 0x00BC68, 0x00000C}, {0x00BC69, 0x00BC83, 0x00000D},
    {0x00BC84, 0x00BC84, 0x00000C}, {0x00BC85, 0x00BC9F, 0x00000D}, {0x00BCA0, 0x00BCA0, 0x00000C},
    {0x00BCA1, 0x00BCBB, 0x00000D}, {0x00BCBC, 0x00BCBC, 0x00000C}, {0x00BCBD, 0x00BCD7, 0x00000D},
    {0x00BCD8, 0x00BCD8, 0x00000C}, {0x00BCD9, 0x00BCF3, 0x00000D}, {0x00BCF4, 0x00BCF4, 0x00000C},
    {0x00BCF5, 0x00BD0F, 0x00000D}, {0x00BD10, 0x00BD10, 0x00000C}, {0x00BD11, 0x00BD2B, 0x00000D},
    {0x00BD2C, 0x00BD2C, 0x00000C}, {0x00BD2D, 0x00BD47, 0x00000D}, {0x00BD48, 0x00BD48, 0x00000C},
    {0x00BD49, 0x00BD63, 0x00000D}, {0x00BD64, 0x00BD64, 0x00000C}, {0x00BD65, 0x00BD7F, 0x00000D},
    {0x00BD80, 0x00BD80, 0x00000C}, {0x00BD81, 0x00BD9B, 0x00000D}, {0x00BD9C, 0x00BD9C, 0x00000C},
    {0x00BD9D, 0x00BDB7, 0x00000D}, {0x00BDB8, 0x00BDB8, 0x00000C}, {0x00BDB9, 0x00BDD3, 0x00000D},
    {0x00BDD4, 0x00BDD4, 0x00000C}, {0x00BDD5, 0x00BDEF, 0x00000D}, {0x00BDF0, 0x00BDF0, 0x00000C},
    {0x00BDF1, 0x00BE0B, 0x00000D}, {0x00BE0C, 0x00BE0C, 0x00000C}, {0x00BE0D, 0x00BE27, 0x00000D},
    {0x00BE28, 0x00BE28, 0x00000C}, {0x00BE29, 0x00BE43, 0x00000D}, {0x00BE44, 0x00BE44, 0x00000C},
    {0x00BE45, 0x00BE5F, 0x00000D}, {0x00BE60, 0x00BE60, 0x00000C}, {0x00BE61, 0x00BE7B, 0x00000D},
    {0x00BE7C, 0x00BE7C, 0x00000C}, {0x00BE7D, 0x00BE97, 0x00000D}, {0x00BE98, 0x00BE98, 0x00000C},
    {0x00BE99, 0x00BEB3, 0x00000D}, {0x00BEB4, 0x00BEB4, 0x00000C}, {0x00BEB5, 0x00BECF, 0x00000D},
    {0x00BED0, 0x00BED0, 0x00000C}, {0x00BED1, 0x00BEEB, 0x00000D}, {0x00BEEC, 0x00BEEC, 0x00000C},
    {0x00BEED, 0x00BF07, 0x00000D}, {0x00BF08, 0x00BF08, 0x00000C}, {0x00BF09, 0x00BF23, 0x00000D},
    {0x00BF24, 0x00BF24, 0x00000C}, {0x00BF25, 0x00BF3F, 0x00000D}, {0x00BF40, 0x00BF40, 0x00000C},
    {0x00BF41, 0x00BF5B, 0x00000D}, {0x00BF5C, 0x00BF5C, 0x00000C}, {0x00BF5D, 0x00BF77, 0x00000D},
    {0x00BF78, 0x00BF78, 0x00000C}, {0x00BF79, 0x00BF93, 0x00000D}, {0x00BF94, 0x00BF94, 0x00000C},
    {0x00BF95, 0x00BFAF, 0x00000D}, {0x00BFB0, 0x00BFB0, 0x00000C}, {0x00BFB1, 0x00BFCB, 0x00000D},
    {0x00BFCC, 0x00BFCC, 0x00000C}, {0x00BFCD, 0x00BFE7, 0x00000D}, {0x00BFE8, 0x00BFE8, 0x00000C},
    {0x00BFE9, 0x00C003, 0x00000D}, {0x00C004, 0x00C004, 0x00000C}, {0x00C005, 0x00C01F, 0x00000D},
    {0x00C020, 0x00C020, 0x00000C}, {0x00C021, 0x00C03B, 0x00000D}, {0x00C03C, 0x00C03C, 0x00000C},
    {0x00C03D, 0x00C057, 0x00000D}, {0x00C058, 0x00C058, 0x00000C}, {0x00C059, 0x00C073, 0x00000D},
    {0x00C074, 0x00C074, 0x00000C}, {0x00C075, 0x00C08F, 0x00000D}, {0x00C090, 0x00C090, 0x00000C},
    {0x00C091, 0x00C0AB, 0x00000D}, {0x00C0AC, 0x00C0AC, 0x00000C}, {0x00C0AD, 0x00C0C7, 0x00000D},
    {0x00C0C8, 0x00C0C8, 0x00000C}, {0x00C0C9, 0x00C0E3, 0x00000D}, {0x00C0E4, 0x00C0E4, 0x00000C},
    {0x00C0E5, 0x00C0FF, 0x00000D}, {0x00C100, 0x00C100, 0x00000C}, {0x00C101, 0x00C11B, 0x00000D},
    {0x00C11C, 0x00C11C, 0x00000C}, {0x00C11D, 0x00C137, 0x00000D}, {0x00C138, 0x00C138, 0x00000C},
    {0x00C139, 0x00C153, 0x00000D}, {0x00C154, 0x00C154, 0x00000C}, {0x00C155, 0x00C16F, 0x00000D},
    {0x00C170, 0x00C170, 0x00000C}, {0x00C171, 0x00C18B, 0x00000D}, {0x00C18C, 0x00C18C, 0x00000C},
    {0x00C18D, 0x00C1A7, 0x00000D}, {0x00C1A8, 0x00C1A8, 0x00000C}, {0x00C1A9, 0x00C1C3, 0x00000D},
    {0x00C1C4, 0x00C1C4, 0x00000C}, {0x00C1C5, 0x00C1DF, 0x00000D}, {0x00C1E0, 0x00C1E0, 0x00000C},
    {0x00C1E1, 0x00C1FB, 0x00000D}, {0x00C1FC, 0x00C1FC, 0x00000C}, {0x00C1FD, 0x00C217, 0x00000D},
    {0x00C218, 0x00C218, 0x00000C}, {0x00C219, 0x00C233, 0x00000D}, {0x00C234, 0x00C234, 0x00000C},
    {0x00C235, 0x00C24F, 0x00000D}, {0x00C250, 0x00C250, 0x00000C}, {0x00C251, 0x00C26B, 0x00000D},
    {0x00C26C, 0x00C26C, 0x00000C}, {0x00C26D, 0x00C287, 0x00000D}, {0x00C288, 0x00C288, 0x00000C},
    {0x00C289, 0x00C2A3, 0x00000D}, {0x00C2A4, 0x00C2A4, 0x00000C}, {0x00C2A5, 0x00C2BF, 0x00000D},
    {0x00C2C0, 0x00C2C0, 0x00000C}, {0x00C2C1, 0x00C2DB, 0x00000D}, {0x00C2DC, 0x00C2DC, 0x00000C},
    {0x00C2DD, 0x00C2F7, 0x00000D}, {0x00C2F8, 0x00C2F8, 0x00000C}, {0x00C2F9, 0x00C313, 0x00000D},
    {0x00C314, 0x00C314, 0x00000C}, {0x00C315, 0x00C32F, 0x00000D}, {0x00C330, 0x00C330, 0x00000C},
    {0x00C331, 0x00C34B, 0x00000D}, {0x00C34C, 0x00C34C, 0x00000C}, {0x00C34D, 0x00C367, 0x00000D},
    {0x00C368, 0x00C368, 0x00000C}, {0x00C369, 0x00C383, 0x00000D}, {0x00C384, 0x00C384, 0x00000C},
    {0x00C385, 0x00C39F, 0x00000D}, {0x00C3A0, 0x00C3A0, 0x00000C}, {0x00C3A1, 0x00C3BB, 0x00000D},
    {0x00C3BC, 0x00C3BC, 0x00000C}, {0x00C3BD, 0x00C3D7, 0x00000D}, {0x00C3D8, 0x00C3D8, 0x00000C},
    {0x00C3D9, 0x00C3F3, 0x00000D}, {0x00C3F4, 0x00C3F4, 0x00000C}, {0x00C3F5, 0x00C40F, 0x00000D},
    {0x00C410, 0x00C410, 0x00000C}, {0x00C411, 0x00C42B, 0x00000D}, {0x00C42C, 0x00C42C, 0x00000C},
    {0x00C42D, 0x00C447, 0x00000D}, {0x00C448, 0x00C448, 0x00000C}, {0x00C449, 0x00C463, 0x00000D},
    {0x00C464, 0x00C464, 0x00000C}, {0x00C465, 0x00C47F, 0x00000D}, {0x00C480, 0x00C480, 0x00000C},
    {0x00C481, 0x00C49B, 0x00000D}, {0x00C49C, 0x00C49C, 0x00000C}, {0x00C49D, 0x00C4B7, 0x00000D},
    {0x00C4B8, 0x00C4B8, 0x00000C}, {0x00C4B9, 0x00C4D3, 0x00000D}, {0x00C4D4, 0x00C4D4, 0x00000C},
    {0x00C4D5, 0x00C4EF, 0x00000D}, {0x00C4F0, 0x00C4F0, 0x00000C}, {0x00C4F1, 0x00C50B, 0x00000D},
    {0x00C50C, 0x00C50C, 0x00000C}, {0x00C50D, 0x00C527, 0x00000D}, {0x00C528, 0x00C528, 0x00000C},
    {0x00C529, 0x00C543, 0x00000D}, {0x00C544, 0x00C544, 0x00000C}, {0x00C545, 0x00C55F, 0x00000D},
    {0x00C560, 0x00C560, 0x00000C}, {0x00C561, 0x00C57B, 0x00000D}, {0x00C57C, 0x00C57C, 0x00000C},
    {0x00C57D, 0x00C597, 0x00000D}, {0x00C598, 0x00C598, 0x00000C}, {0x00C599, 0x00C5B3, 0x00000D},
    {0x00C5B4, 0x00C5B4, 0x00000C}, {0x00C5B5, 0x00C5CF, 0x00000D}, {0x00C5D0, 0x00C5D0, 0x00000C},
    {0x00C5D1, 0x00C5EB, 0x00000D}, {0x00C5EC, 0x00C5EC, 0x00000C}, {0x00C5ED, 0x00C607, 0x00000D},
    {0x00C608, 0x00C608, 0x00000C}, {0x00C609, 0x00C623, 0x00000D}, {0x00C624, 0x00C624, 0x00000C},
    {0x00C625, 0x00C63F, 0x00000D}, {0x00C640, 0x00C640, 0x00000C}, {0x00C641, 0x00C65B, 0x00000D},
    {0x00C65C, 0x00C65C, 0x00000C}, {0x00C65D, 0x00C677, 0x00000D}, {0x00C678, 0x00C678, 0x00000C},
    {0x00C679, 0x00C693, 0x00000D}, {0x00C694, 0x00C694, 0x00000C}, {0x00C695, 0x00C6AF, 0x00000D},
    {0x00C6B0, 0x00C6B0, 0x00000C}, {0x00C6B1, 0x00C6CB, 0x00000D}, {0x00C6CC, 0x00C6CC, 0x00000C},
    {0x00C6CD, 0x00C6E7, 0x00000D}, {0x00C6E8, 0x00C6E8, 0x00000C}, {0x00C6E9, 0x00C703, 0x00000D},
    {0x00C704, 0x00C704, 0x00000C}, {0x00C705, 0x00C71F, 0x00000D}, {0x00C720, 0x00C720, 0x00000C},
    {0x00C721, 0x00C73B, 0x00000D}, {0x00C73C, 0x00C73C, 0x00000C}, {0x00C73D, 0x00C757, 0x00000D},
    {0x00C758, 0x00C758, 0x00000C}, {0x00C759, 0x00C773, 0x00000D}, {0x00C774, 0x00C774, 0x00000C},
    {0x00C775, 0x00C78F, 0x00000D}, {0x00C790, 0x00C790, 0x00000C}, {0x00C791, 0x00C7AB, 0x00000D},
    {0x00C7AC, 0x00C7AC, 0x00000C}, {0x00C7AD, 0x00C7C7, 0x00000D}, {0x00C7C8, 0x00C7C8, 0x00000C},
    {0x00C7C9, 0x00C7E3, 0x00000D}, {0x00C7E4, 0x00C7E4, 0x00000C}, {0x00C7E5, 0x00C7FF, 0x00000D},
    {0x00C800, 0x00C800, 0x00000C}, {0x00C801, 0x00C81B, 0x00000D}, {0x00C81C, 0x00C81C, 0x00000C},
    {0x00C81D, 0x00C837, 0x00000D}, {0x00C838, 0x00C838, 0x00000C}, {0x00C839, 0x00C853, 0x00000D},
    {0x00C854, 0x00C854, 0x00000C}, {0x00C855, 0x00C86F, 0x00000D}, {0x00C870, 0x00C870, 0x00000C},
    {0x00C871, 0x00C88B, 0x00000D}, {0x00C88C, 0x00C88C, 0x00000C}, {0x00C88D, 0x00C8A7, 0x00000D},
    {0x00C8A8, 0x00C8A8, 0x00000C}, {0x00C8A9, 0x00C8C3, 0x00000D}, {0x00C8C4, 0x00C8C4, 0x00000C},
    {0x00C8C5, 0x00C8DF, 0x00000D}, {0x00C8E0, 0x00C8E0, 0x00000C}, {0x00C8E1, 0x00C8FB, 0x00000D},
    {0x00C8FC, 0x00C8FC, 0x00000C}, {0x00C8FD, 0x00C917, 0x00000D}, {0x00C918, 0x00C918, 0x00000C},
    {0x00C919, 0x00C933, 0x00000D}, {0x00C934, 0x00C934, 0x00000C}, {0x00C935, 0x00C94F, 0x00000D},
    {0x00C950, 0x00C950, 0x00000C}, {0x00C951, 0x00C96B, 0x00000D}, {0x00C96C, 0x00C96C, 0x00000C},
    {0x00C96D, 0x00C987, 0x00000D}, {0x00C988, 0x00C988, 0x00000C}, {0x00C989, 0x00C9A3, 0x00000D},
    {0x00C9A4, 0x00C9A4, 0x00000C}, {0x00C9A5, 0x00C9BF, 0x00000D}, {0x00C9C0, 0x00C9C0, 0x00000C},
    {0x00C9C1, 0x00C9DB, 0x00000D}, {0x00C9DC, 0x00C9DC, 0x00000C}, {0x00C9DD, 0x00C9F7, 0x00000D},
    {0x00C9F8, 0x00C9F8, 0x00000C}, {0x00C9F9, 0x00CA13, 0x00000D}, {0x00CA14, 0x00CA14, 0x00000C},
    {0x00CA15, 0x00CA2F, 0x00000D}, {0x00CA30, 0x00CA30, 0x00000C}, {0x00CA31, 0x00CA4B, 0x00000D},
    {0x00CA4C, 0x00CA4C, 0x00000C}, {0x00CA4D, 0x00CA67, 0x00000D}, {0x00CA68, 0x00CA68, 0x00000C},
    {0x00CA69, 0x00CA83, 0x00000D}, {0x00CA84, 0x00CA84, 0x00000C}, {0x00CA85, 0x00CA9F, 0x00000D},
    {0x00CAA0, 0x00CAA0, 0x00000C}, {0x00CAA1, 0x00CABB, 0x00000D}, {0x00CABC, 0x00CABC, 0x00000C},
    {0x00CABD, 0x00CAD7, 0x00000D}, {0x00CAD8, 0x00CAD8, 0x00000C}, {0x00CAD9, 0x00CAF3, 0x00000D},
    {0x00CAF4, 0x00CAF4, 0x00000C}, {0x00CAF5, 0x00CB0F, 0x00000D}, {0x00CB10, 0x00CB10, 0x00000C},
    {0x00CB11, 0x00CB2B, 0x00000D}, {0x00CB2C, 0x00CB2C, 0x00000C}, {0x00CB2D, 0x00CB47, 0x00000D},
    {0x00CB48, 0x00CB48, 0x00000C}, {0x00CB49, 0x00CB63, 0x00000D}, {0x00CB64, 0x00CB64, 0x00000C},
    {0x00CB65, 0x00CB7F, 0x00000D}, {0x00CB80, 0x00CB80, 0x00000C}, {0x00CB81, 0x00CB9B, 0x00000D},
    {0x00CB9C, 0x00CB9C, 0x00000C}, {0x00CB9D, 0x00CBB7, 0x00000D}, {0x00CBB8, 0x00CBB8, 0x00000C},
    {0x00CBB9, 0x00CBD3, 0x00000D}, {0x00CBD4, 0x00CBD4, 0x00000C}, {0x00CBD5, 0x00CBEF, 0x00000D},
    {0x00CBF0, 0x00CBF0, 0x00000C}, {0x00CBF1, 0x00CC0B, 0x00000D}, {0x00CC0C, 0x00CC0C, 0x00000C},
    {0x00CC0D, 0x00CC27, 0x00000D}, {0x00CC28, 0x00CC28, 0x00000C}, {0x00CC29, 0x00CC43, 0x00000D},
    {0x00CC44, 0x00CC44, 0x00000C}, {0x00CC45, 0x00CC5F, 0x00000D}, {0x00CC60, 0x00CC60, 0x00000C},
    {0x00CC61, 0x00CC7B, 0x00000D}, {0x00CC7C, 0x00CC7C, 0x00000C}, {0x00CC7D, 0x00CC97, 0x00000D},
    {0x00CC98, 0x00CC98, 0x00000C}, {0x00CC99, 0x00CCB3, 0x00000D}, {0x00CCB4, 0x00CCB4, 0x00000C},
    {0x00CCB5, 0x00CCCF, 0x00000D}, {0x00CCD0, 0x00CCD0, 0x00000C}, {0x00CCD1, 0x00CCEB, 0x00000D},
    {0x00CCEC, 0x00CCEC, 0x00000C}, {0x00CCED, 0x00CD07, 0x00000D}, {0x00CD08, 0x00CD08, 0x00000C},
    {0x00CD09, 0x00CD23, 0x00000D}, {0x00CD24, 0x00CD24, 0x00000C}, {0x00CD25, 0x00CD3F, 0x00000D},
    {0x00CD40, 0x00CD40, 0x00000C}, {0x00CD41, 0x00CD5B, 0x00000D}, {0x00CD5C, 0x00CD5C, 0x00000C},
    {0x00CD5D, 0x00CD77, 0x00000D}, {0x00CD78, 0x00CD78, 0x00000C}, {0x00CD79, 0x00CD93, 0x00000D},
    {0x00CD94, 0x00CD94, 0x00000C}, {0x00CD95, 0x00CDAF, 0x00000D}, {0x00CDB0, 0x00CDB0, 0x00000C},
    {0x00CDB1, 0x00CDCB, 0x00000D}, {0x00CDCC, 0x00CDCC, 0x00000C}, {0x00CDCD, 0x00CDE7, 0x00000D},
    {0x00CDE8, 0x00CDE8, 0x00000C}, {0x00CDE9, 0x00CE03, 0x00000D}, {0x00CE04, 0x00CE04, 0x00000C},
    {0x00CE05, 0x00CE1F, 0x00000D}, {0x00CE20, 0x00CE20, 0x00000C}, {0x00CE21, 0x00CE3B, 0x00000D},
    {0x00CE3C, 0x00CE3C, 0x00000C}, {0x00CE3D, 0x00CE57, 0x00000D}, {0x00CE58, 0x00CE58, 0x00000C},
    {0x00CE59, 0x00CE73, 0x00000D}, {0x00CE74, 0x00CE74, 0x00000C}, {0x00CE75, 0x00CE8F, 0x00000D},
    {0x00CE90, 0x00CE90, 0x00000C}, {0x00CE91, 0x00CEAB, 0x00000D}, {0x00CEAC, 0x00CEAC, 0x00000C},
    {0x00CEAD, 0x00CEC7, 0x00000D}, {0x00CEC8, 0x00CEC8, 0x00000C}, {0x00CEC9, 0x00CEE3, 0x00000D},
    {0x00CEE4, 0x00CEE4, 0x00000C}, {0x00CEE5, 0x00CEFF, 0x00000D}, {0x00CF00, 0x00CF00, 0x00000C},
    {0x00CF01, 0x00CF1B, 0x00000D}, {0x00CF1C, 0x00CF1C, 0x00000C}, {0x00CF1D, 0x00CF37, 0x00000D},
    {0x00CF38, 0x00CF38, 0x00000C}, {0x00CF39, 0x00CF53, 0x00000D}, {0x00CF54, 0x00CF54, 0x00000C},
    {0x00CF55, 0x00CF6F, 0x00000D}, {0x00CF70, 0x00CF70, 0x00000C}, {0x00CF71, 0x00CF8B, 0x00000D},
    {0x00CF8C, 0x00CF8C, 0x00000C}, {0x00CF8D, 0x00CFA7, 0x00000D}, {0x00CFA8, 0x00CFA8, 0x00000C},
    {0x00CFA9, 0x00CFC3, 0x00000D}, {0x00CFC4, 0x00CFC4, 0x00000C}, {0x00CFC5, 0x00CFDF, 0x00000D},
    {0x00CFE0, 0x00CFE0, 0x00000C}, {0x00CFE1, 0x00CFFB, 0x00000D}, {0x00CFFC, 0x00CFFC, 0x00000C},
    {0x00CFFD, 0x00D017, 0x00000D}, {0x00D018, 0x00D018, 0x00000C}, {0x00D019, 0x00D033, 0x00000D},
    {0x00D034, 0x00D034, 0x00000C}, {0x00D035, 0x00D04F, 0x00000D}, {0x00D050, 0x00D050, 0x00000C},
    {0x00D051, 0x00D06B, 0x00000D}, {0x00D06C, 0x00D06C, 0x00000C}, {0x00D06D, 0x00D087, 0x00000D},
    {0x00D088, 0x00D088, 0x00000C}, {0x00D089, 0x00D0A3, 0x00000D}, {0x00D0A4, 0x00D0A4, 0x00000C},
    {0x00D0A5, 0x00D0BF, 0x00000D}, {0x00D0C0, 0x00D0C0, 0x00000C}, {0x00D0C1, 0x00D0DB, 0x00000D},
    {0x00D0DC, 0x00D0DC, 0x00000C}, {0x00D0DD, 0x00D0F7, 0x00000D}, {0x00D0F8, 0x00D0F8, 0x00000C},
    {0x00D0F9, 0x00D113, 0x00000D}, {0x00D114, 0x00D114, 0x00000C}, {0x00D115, 0x00D12F, 0x00000D},
    {0x00D130, 0x00D130, 0x00000C}, {0x00D131, 0x00D14B, 0x00000D}, {0x00D14C, 0x00D14C, 0x00000C},
    {0x00D14D, 0x00D167, 0x00000D}, {0x00D168, 0x00D168, 0x00000C}, {0x00D169, 0x00D183, 0x00000D},
    {0x00D184, 0x00D184, 0x00000C}, {0x00D185, 0x00D19F, 0x00000D}, {0x00D1A0, 0x00D1A0, 0x00000C},
    {0x00D1A1, 0x00D1BB, 0x00000D}, {0x00D1BC, 0x00D1BC, 0x00000C}, {0x00D1BD, 0x00D1D7, 0x00000D},
    {0x00D1D8, 0x00D1D8, 0x00000C}, {0x00D1D9, 0x00D1F3, 0x00000D}, {0x00D1F4, 0x00D1F4, 0x00000C},
    {0x00D1F5, 0x00D20F, 0x00000D}, {0x00D210, 0x00D210, 0x00000C}, {0x00D211, 0x00D22B, 0x00000D},
    {0x00D22C, 0x00D22C, 0x00000C}, {0x00D22D, 0x00D247, 0x00000D}, {0x00D248, 0x00D248, 0x00000C},
    {0x00D249, 0x00D263, 0x00000D}, {0x00D264, 0x00D264, 0x00000C}, {0x00D265, 0x00D27F, 0x00000D},
    {0x00D280, 0x00D280, 0x00000C}, {0x00D281, 0x00D29B, 0x00000D}, {0x00D29C, 0x00D29C, 0x00000C},
    {0x00D29D, 0x00D2B7, 0x00000D}, {0x00D2B8, 0x00D2B8, 0x00000C}, {0x00D2B9, 0x00D2D3, 0x00000D},
    {0x00D2D4, 0x00D2D4, 0x00000C}, {0x00D2D5, 0x00D2EF, 0x00000D}, {0x00D2F0, 0x00D2F0, 0x00000C},
    {0x00D2F1, 0x00D30B, 0x00000D}, {0x00D30C, 0x00D30C, 0x00000C}, {0x00D30D, 0x00D327, 0x00000D},
    {0x00D328, 0x00D328, 0x00000C}, {0x00D329, 0x00D343, 0x00000D}, {0x00D344, 0x00D344, 0x00000C},
    {0x00D345, 0x00D35F, 0x00000D}, {0x00D360, 0x00D360, 0x00000C}, {0x00D361, 0x00D37B, 0x00000D},
    {0x00D37C, 0x00D37C, 0x00000C}, {0x00D37D, 0x00D397, 0x00000D}, {0x00D398, 0x00D398, 0x00000C},
    {0x00D399, 0x00D3B3, 0x00000D}, {0x00D3B4, 0x00D3B4, 0x00000C}, {0x00D3B5, 0x00D3CF, 0x00000D},
    {0x00D3D0, 0x00D3D0, 0x00000C}, {0x00D3D1, 0x00D3EB, 0x00000D}, {0x00D3EC, 0x00D3EC, 0x00000C},
    {0x00D3ED, 0x00D407, 0x00000D}, {0x00D408, 0x00D408, 0x00000C}, {0x00D409, 0x00D423, 0x00000D},
    {0x00D424, 0x00D424, 0x00000C}, {0x00D425, 0x00D43F, 0x00000D}, {0x00D440, 0x00D440, 0x00000C},
    {0x00D441, 0x00D45B, 0x00000D}, {0x00D45C, 0x00D45C, 0x00000C}, {0x00D45D, 0x00D477, 0x00000D},
    {0x00D478, 0x00D478, 0x00000C}, {0x00D479, 0x00D493, 0x00000D}, {0x00D494, 0x00D494, 0x00000C},
    {0x00D495, 0x00D4AF, 0x00000D}, {0x00D4B0, 0x00D4B0, 0x00000C}, {0x00D4B1, 0x00D4CB, 0x00000D},
    {0x00D4CC, 0x00D4CC, 0x00000C}, {0x00D4CD, 0x00D4E7, 0x00000D}, {0x00D4E8, 0x00D4E8, 0x00000C},
    {0x00D4E9, 0x00D503, 0x00000D}, {0x00D504, 0x00D504, 0x00000C}, {0x00D505, 0x00D51F, 0x00000D},
    {0x00D520, 0x00D520, 0x00000C}, {0x00D521, 0x00D53B, 0x00000D}, {0x00D53C, 0x00D53C, 0x00000C},
    {0x00D53D, 0x00D557, 0x00000D}, {0x00D558, 0x00D558, 0x00000C}, {0x00D559, 0x00D573, 0x00000D},
    {0x00D574, 0x00D574, 0x00000C}, {0x00D575, 0x00D58F, 0x00000D}, {0x00D590, 0x00D590, 0x00000C},
    {0x00D591, 0x00D5AB, 0x00000D}, {0x00D5AC, 0x00D5AC, 0x00000C}, {0x00D5AD, 0x00D5C7, 0x00000D},
    {0x00D5C8, 0x00D5C8, 0x00000C}, {0x00D5C9, 0x00D5E3, 0x00000D}, {0x00D5E4, 0x00D5E4, 0x00000C},
    {0x00D5E5, 0x00D5FF, 0x00000D}, {0x00D600, 0x00D600, 0x00000C}, {0x00D601, 0x00D61B, 0x00000D},
    {0x00D61C, 0x00D61C, 0x00000C}, {0x00D61D, 0x00D637, 0x00000D}, {0x00D638, 0x00D638, 0x00000C},
    {0x00D639, 0x00D653, 0x00000D}, {0x00D654, 0x00D654, 0x00000C}, {0x00D655, 0x00D66F, 0x00000D},
    {0x00D670, 0x00D670, 0x00000C}, {0x00D671, 0x00D68B, 0x00000D}, {0x00D68C, 0x00D68C, 0x00000C},
    {0x00D68D, 0x00D6A7, 0x00000D}, {0x00D6A8, 0x00D6A8, 0x00000C}, {0x00D6A9, 0x00D6C3, 0x00000D},
    {0x00D6C4, 0x00D6C4, 0x00000C}, {0x00D6C5, 0x00D6DF, 0x00000D}, {0x00D6E0, 0x00D6E0, 0x00000C},
    {0x00D6E1, 0x00D6FB, 0x00000D}, {0x00D6FC, 0x00D6FC, 0x00000C}, {0x00D6FD, 0x00D717, 0x00000D},
    {0x00D718, 0x00D718, 0x00000C}, {0x00D719, 0x00D733, 0x00000D}, {0x00D734, 0x00D734, 0x00000C},
    {0x00D735, 0x00D74F, 0x00000D}, {0x00D750, 0x00D750, 0x00000C}, {0x00D751, 0x00D76B, 0x00000D},
    {0x00D76C, 0x00D76C, 0x00000C}, {0x00D76D, 0x00D787, 0x00000D}, {0x00D788, 0x00D788, 0x00000C},
    {0x00D789, 0x00D7A3, 0x00000D}, {0x00D7B0, 0x00D7C6, 0x00000A}, {0x00D7CB, 0x00D7FB, 0x00000B},
    {0x00FB1E, 0x00FB1E, 0x000004}, {0x00FE00, 0x00FE0F, 0x000004}, {0x00FE20, 0x00FE2F, 0x000004},
    {0x00FEFF, 0x00FEFF, 0x000003}, {0x00FF9E, 0x00FF9F, 0x000004}, {0x00FFF0, 0x00FFF8, 0x000003},
    {0x00FFF9, 0x00FFFB, 0x000003}, {0x0101FD, 0x0101FD, 0x000004}, {0x0102E0, 0x0102E0, 0x000004},
    {0x010376, 0x01037A, 0x000004}, {0x010A01, 0x010A03, 0x000004}, {0x010A05, 0x010A06, 0x000004},
    {0x010A0C, 0x010A0F, 0x000004}, {0x010A38, 0x010A3A, 0x000004}, {0x010A3F, 0x010A3F, 0x000004},
    {0x010AE5, 0x010AE6, 0x000004}, {0x010D24, 0x010D27, 0x000004}, {0x010EAB, 0x010EAC, 0x000004},
    {0x010F46, 0x010F50, 0x000004}, {0x010F82, 0x010F85, 0x000004}, {0x011000, 0x011000, 0x000008},
    {0x011001, 0x011001, 0x000004}, {0x011002, 0x011002, 0x000008}, {0x011038, 0x011046, 0x000004},
    {0x011070, 0x011070, 0x000004}, {0x011073, 0x011074, 0x000004}, {0x01107F, 0x011081, 0x000004},
    {0x011082, 0x011082, 0x000008}, {0x0110B0, 0x0110B2, 0x000008}, {0x0110B3, 0x0110B6, 0x000004},
    {0x0110B7, 0x0110B8, 0x000008}, {0x0110B9, 0x0110BA, 0x000004}, {0x0110BD, 0x0110BD, 0x000007},
    {0x0110C2, 0x0110C2, 0x000004}, {0x0110CD, 0x0110CD, 0x000007}, {0x011100, 0x011102, 0x000004},
    {0x011127, 0x01112B, 0x000004}, {0x01112C, 0x01112C, 0x000008}, {0x01112D, 0x011134, 0x000004},
    {0x011145, 0x011146, 0x000008}, {0x011173, 0x011173, 0x000004}, {0x011180, 0x011181, 0x000004},
    {0x011182, 0x011182, 0x000008}, {0x0111B3, 0x0111B5, 0x000008}, {0x0111B6, 0x0111BE, 0x000004},
    {0x0111BF, 0x0111C0, 0x000008}, {0x0111C2, 0x0111C3, 0x000007}, {0x0111C9, 0x0111CC, 0x000004},
    {0x0111CE, 0x0111CE, 0x000008}, {0x0111CF, 0x0111CF, 0x000004}, {0x01122C, 0x01122E, 0x000008},
    {0x01122F, 0x011231, 0x000004}, {0x011232, 0x011233, 0x000008}, {0x011234, 0x011234, 0x000004},
    {0x011235, 0x011235, 0x000008}, {0x011236, 0x011237, 0x000004}, {0x01123E, 0x01123E, 0x000004},
    {0x0112DF, 0x0112DF, 0x000004}, {0x0112E0, 0x0112E2, 0x000008}, {0x0112E3, 0x0112EA, 0x000004},
    {0x011300, 0x011301, 0x000004}, {0x011302, 0x011303, 0x000008}, {0x01133B, 0x01133C, 0x000004},
    {0x01133E, 0x01133E, 0x000004}, {0x01133F, 0x01133F, 0x000008}, {0x011340, 0x011340, 0x000004},
    {0x011341, 0x011344, 0x000008}, {0x011347, 0x011348, 0x000008}, {0x01134B, 0x01134D, 0x000008},
    {0x011357, 0x011357, 0x000004}, {0x011362, 0x011363, 0x000008}, {0x011366, 0x01136C, 0x000004},
    {0x011370, 0x011374, 0x000004}, {0x011435, 0x011437, 0x000008}, {0x011438, 0x01143F, 0x000004},
    {0x011440, 0x011441, 0x000008}, {0x011442, 0x011444, 0x000004}, {0x011445, 0x011445, 0x000008},
    {0x011446, 0x011446, 0x000004}, {0x01145E, 0x01145E, 0x000004}, {0x0114B0, 0x0114B0, 0x000004},
    {0x0114B1, 0x0114B2, 0x000008}, {0x0114B3, 0x0114B8, 0x000004}, {0x0114B9, 0x0114B9, 0x000008},
    {0x0114BA, 0x0114BA, 0x000004}, {0x0114BB, 0x0114BC, 0x000008}, {0x0114BD, 0x0114BD, 0x000004},
    {0x0114BE, 0x0114BE, 0x000008}, {0x0114BF, 0x0114C0, 0x000004}, {0x0114C1, 0x0114C1, 0x000008},
    {0x0114C2, 0x0114C3, 0x000004}, {0x0115AF, 0x0115AF, 0x000004}, {0x0115B0, 0x0115B1, 0x000008},
    {0x0115B2, 0x0115B5, 0x000004}, {0x0115B8, 0x0115BB, 0x000008}, {0x0115BC, 0x0115BD, 0x000004},
    {0x0115BE, 0x0115BE, 0x000008}, {0x0115BF, 0x0115C0, 0x000004}, {0x0115DC, 0x0115DD, 0x000004},
    {0x011630, 0x011632, 0x000008}, {0x011633, 0x01163A, 0x000004}, {0x01163B, 0x01163C, 0x000008},
    {0x01163D, 0x01163D, 0x000004}, {0x01163E, 0x01163E, 0x000008}, {0x01163F, 0x011640, 0x000004},
    {0x0116AB, 0x0116AB, 0x000004}, {0x0116AC, 0x0116AC, 0x000008}, {0x0116AD, 0x0116AD, 0x000004},
    {0x0116AE, 0x0116AF, 0x000008}, {0x0116B0, 0x0116B5, 0x000004}, {0x0116B6, 0x0116B6, 0x000008},
    {0x0116B7, 0x0116B7, 0x000004}, {0x01171D, 0x01171F, 0x000004}, {0x011722, 0x011725, 0x000004},
    {0x011726, 0x011726, 0x000008}, {0x011727, 0x01172B, 0x000004}, {0x01182C, 0x01182E, 0x000008},
    {0x01182F, 0x011837, 0x000004}, {0x011838, 0x011838, 0x000008}, {0x011839, 0x01183A, 0x000004},
    {0x011930, 0x011930, 0x000004}, {0x011931, 0x011935, 0x000008}, {0x011937, 0x011938, 0x000008},
    {0x01193B, 0x01193C, 0x000004}, {0x01193D, 0x01193D, 0x000008}, {0x01193E, 0x01193E, 0x000004},
    {0x01193F, 0x01193F, 0x000007}, {0x011940, 0x011940, 0x000008}, {0x011941, 0x011941, 0x000007},
    {0x011942, 0x011942, 0x000008}, {0x011943, 0x011943, 0x000004}, {0x0119D1, 0x0119D3, 0x000008},
    {0x0119D4, 0x0119D7, 0x000004}, {0x0119DA, 0x0119DB, 0x000004}, {0x0119DC, 0x0119DF, 0x000008},
    {0x0119E0, 0x0119E0, 0x000004}, {0x0119E4, 0x0119E4, 0x000008}, {0x011A01, 0x011A0A, 0x000004},
    {0x011A33, 0x011A38, 0x000004}, {0x011A39, 0x011A39, 0x000008}, {0x011A3A, 0x011A3A, 0x000007},
    {0x011A3B, 0x011A3E, 0x000004}, {0x011A47, 0x011A47, 0x000004}, {0x011A51, 0x011A56, 0x000004},
    {0x011A57, 0x011A58, 0x000008}, {0x011A59, 0x011A5B, 0x000004}, {0x011A84, 0x011A89, 0x000007},
    {0x011A8A, 0x011A96, 0x000004}, {0x011A97, 0x011A97, 0x000008}, {0x011A98, 0x011A99, 0x000004},
    {0x011C2F, 0x011C2F, 0x000008}, {0x011C30, 0x011C36, 0x000004}, {0x011C38, 0x011C3D, 0x000004},
    {0x011C3E, 0x011C3E, 0x000008}, {0x011C3F, 0x011C3F, 0x000004}, {0x011C92, 0x011CA7, 0x000004},
    {0x011CA9, 0x011CA9, 0x000008}, {0x011CAA, 0x011CB0, 0x000004}, {0x011CB1, 0x011CB1, 0x000008},
    {0x011CB2, 0x011CB3, 0x000004}, {0x011CB4, 0x011CB4, 0x000008}, {0x011CB5, 0x011CB6, 0x000004},
    {0x011D31, 0x011D36, 0x000004}, {0x011D3A, 0x011D3A, 0x000004}, {0x011D3C, 0x011D3D, 0x000004},
    {0x011D3F, 0x011D45, 0x000004}, {0x011D46, 0x011D46, 0x000007}, {0x011D47, 0x011D47, 0x000004},
    {0x011D8A, 0x011D8E, 0x000008}, {0x011D90, 0x011D91, 0x000004}, {0x011D93, 0x011D94, 0x000008},
    {0x011D95, 0x011D95, 0x000004}, {0x011D96, 0x011D96, 0x000008}, {0x011D97, 0x011D97, 0x000004},
    {0x011EF3, 0x011EF4, 0x000004}, {0x011EF5, 0x011EF6, 0x000008}, {0x013430, 0x013438, 0x000003},
    {0x016AF0, 0x016AF4, 0x000004}, {0x016B30, 0x016B36, 0x000004}, {0x016F4F, 0x016F4F, 0x000004},
    {0x016F51, 0x016F87, 0x000008}, {0x016F8F, 0x016F92, 0x000004}, {0x016FE4, 0x016FE4, 0x000004},
    {0x016FF0, 0x016FF1, 0x000008}, {0x01BC9D, 0x01BC9E, 0x000004}, {0x01BCA0, 0x01BCA3, 0x000003},
    {0x01CF00, 0x01CF2D, 0x000004}, {0x01CF30, 0x01CF46, 0x000004}, {0x01D165, 0x01D165, 0x000004},
    {0x01D166, 0x01D166, 0x000008}, {0x01D167, 0x01D169, 0x000004}, {0x01D16D, 0x01D16D, 0x000008},
    {0x01D16E, 0x01D172, 0x000004}, {0x01D173, 0x01D17A, 0x000003}, {0x01D17B, 0x01D182, 0x000004},
    {0x01D185, 0x01D18B, 0x000004}, {0x01D1AA, 0x01D1AD, 0x000004}, {0x01D242, 0x01D244, 0x000004},
    {0x01DA00, 0x01DA36, 0x000004}, {0x01DA3B, 0x01DA6C, 0x000004}, {0x01DA75, 0x01DA75, 0x000004},
    {0x01DA84, 0x01DA84, 0x000004}, {0x01DA9B, 0x01DA9F, 0x000004}, {0x01DAA1, 0x01DAAF, 0x000004},
    {0x01E000, 0x01E006, 0x000004}, {0x01E008, 0x01E018, 0x000004}, {0x01E01B, 0x01E021, 0x000004},
    {0x01E023, 0x01E024, 0x000004}, {0x01E026, 0x01E02A, 0x000004}, {0x01E130, 0x01E136, 0x000004},
    {0x01E2AE, 0x01E2AE, 0x000004}, {0x01E2EC, 0x01E2EF, 0x000004}, {0x01E8D0, 0x01E8D6, 0x000004},
    {0x01E944, 0x01E94A, 0x000004}, {0x01F000, 0x01F003, 0x00000E}, {0x01F004, 0x01F004, 0x00000E},
    {0x01F005, 0x01F0CE, 0x00000E}, {0x01F0CF, 0x01F0CF, 0x00000E}, {0x01F0D0, 0x01F0FF, 0x00000E},
    {0x01F10D, 0x01F10F, 0x00000E}, {0x01F12F, 0x01F12F, 0x00000E}, {0x01F16C, 0x01F16F, 0x00000E},
    {0x01F170, 0x01F171, 0x00000E}, {0x01F17E, 0x01F17F, 0x00000E}, {0x01F18E, 0x01F18E, 0x00000E},
    {0x01F191, 0x01F19A, 0x00000E}, {0x01F1AD, 0x01F1E5, 0x00000E}, {0x01F1E6, 0x01F1FF, 0x000006},
    {0x01F201, 0x01F202, 0x00000E}, {0x01F203, 0x01F20F, 0x00000E}, {0x01F21A, 0x01F21A, 0x00000E},
    {0x01F22F, 0x01F22F, 0x00000E}, {0x01F232, 0x01F23A, 0x00000E}, {0x01F23C, 0x01F23F, 0x00000E},
    {0x01F249, 0x01F24F, 0x00000E}, {0x01F250, 0x01F251, 0x00000E}, {0x01F252, 0x01F2FF, 0x00000E},
    {0x01F300, 0x01F30C, 0x00000E}, {0x01F30D, 0x01F30E, 0x00000E}, {0x01F30F, 0x01F30F, 0x00000E},
    {0x01F310, 0x01F310, 0x00000E}, {0x01F311, 0x01F311, 0x00000E}, {0x01F312, 0x01F312, 0x00000E},
    {0x01F313, 0x01F315, 0x00000E}, {0x01F316, 0x01F318, 0x00000E}, {0x01F319, 0x01F319, 0x00000E},
    {0x01F31A, 0x01F31A, 0x00000E}, {0x01F31B, 0x01F31B, 0x00000E}, {0x01F31C, 0x01F31C, 0x00000E},
    {0x01F31D, 0x01F31E, 0x00000E}, {0x01F31F, 0x01F320, 0x00000E}, {0x01F321, 0x01F321, 0x00000E},
    {0x01F322, 0x01F323, 0x00000E}, {0x01F324, 0x01F32C, 0x00000E}, {0x01F32D, 0x01F32F, 0x00000E},
    {0x01F330, 0x01F331, 0x00000E}, {0x01F332, 0x01F333, 0x00000E}, {0x01F334, 0x01F335, 0x00000E},
    {0x01F336, 0x01F336, 0x00000E}, {0x01F337, 0x01F34A, 0x00000E}, {0x01F34B, 0x01F34B, 0x00000E},
    {0x01F34C, 0x01F34F, 0x00000E}, {0x01F350, 0x01F350, 0x00000E}, {0x01F351, 0x01F37B, 0x00000E},
    {0x01F37C, 0x01F37C, 0x00000E}, {0x01F37D, 0x01F37D, 0x00000E}, {0x01F37E, 0x01F37F, 0x00000E},
    {0x01F380, 0x01F393, 0x00000E}, {0x01F394, 0x01F395, 0x00000E}, {0x01F396, 0x01F397, 0x00000E},
    {0x01F398, 0x01F398, 0x00000E}, {0x01F399, 0x01F39B, 0x00000E}, {0x01F39C, 0x01F39D, 0x00000E},
    {0x01F39E, 0x01F39F, 0x00000E}, {0x01F3A0, 0x01F3C4, 0x00000E}, {0x01F3C5, 0x01F3C5, 0x00000E},
    {0x01F3C6, 0x01F3C6, 0x00000E}, {0x01F3C7, 0x01F3C7, 0x00000E}, {0x01F3C8, 0x01F3C8, 0x00000E},
    {0x01F3C9, 0x01F3C9, 0x00000E}, {0x01F3CA, 0x01F3CA, 0x00000E}, {0x01F3CB, 0x01F3CE, 0x00000E},
    {0x01F3CF, 0x01F3D3, 0x00000E}, {0x01F3D4, 0x01F3DF, 0x00000E}, {0x01F3E0, 0x01F3E3, 0x00000E},
    {0x01F3E4, 0x01F3E4, 0x00000E}, {0x01F3E5, 0x01F3F0, 0x00000E}, {0x01F3F1, 0x01F3F2, 0x00000E},
    {0x01F3F3, 0x01F3F3, 0x00000E}, {0x01F3F4, 0x01F3F4, 0x00000E}, {0x01F3F5, 0x01F3F5, 0x00000E},
    {0x01F3F6, 0x01F3F6, 0x00000E}, {0x01F3F7, 0x01F3F7, 0x00000E}, {0x01F3F8, 0x01F3FA, 0x00000E},
    {0x01F3FB, 0x01F3FF, 0x000004}, {0x01F400, 0x01F407, 0x00000E}, {0x01F408, 0x01F408, 0x00000E},
    {0x01F409, 0x01F40B, 0x00000E}, {0x01F40C, 0x01F40E, 0x00000E}, {0x01F40F, 0x01F410, 0x00000E},
    {0x01F411, 0x01F412, 0x00000E}, {0x01F413, 0x01F413, 0x00000E}, {0x01F414, 0x01F414, 0x00000E},
    {0x01F415, 0x01F415, 0x00000E}, {0x01F416, 0x01F416, 0x00000E}, {0x01F417, 0x01F429, 0x00000E},
    {0x01F42A, 0x01F42A, 0x00000E}, {0x01F42B, 0x01F43E, 0x00000E}, {0x01F43F, 0x01F43F, 0x00000E},
    {0x01F440, 0x01F440, 0x00000E}, {0x01F441, 0x01F441, 0x00000E}, {0x01F442, 0x01F464, 0x00000E},
    {0x01F465, 0x01F465, 0x00000E}, {0x01F466, 0x01F46B, 0x00000E}, {0x01F46C, 0x01F46D, 0x00000E},
    {0x01F46E, 0x01F4AC, 0x00000E}, {0x01F4AD, 0x01F4AD, 0x00000E}, {0x01F4AE, 0x01F4B5, 0x00000E},
    {0x01F4B6, 0x01F4B7, 0x00000E}, {0x01F4B8, 0x01F4EB, 0x00000E}, {0x01F4EC, 0x01F4ED, 0x00000E},
    {0x01F4EE, 0x01F4EE, 0x00000E}, {0x01F4EF, 0x01F4EF, 0x00000E}, {0x01F4F0, 0x01F4F4, 0x00000E},
    {0x01F4F5, 0x01F4F5, 0x00000E}, {0x01F4F6, 0x01F4F7, 0x00000E}, {0x01F4F8, 0x01F4F8, 0x00000E},
    {0x01F4F9, 0x01F4FC, 0x00000E}, {0x01F4FD, 0x01F4FD, 0x00000E}, {0x01F4FE, 0x01F4FE, 0x00000E},
    {0x01F4FF, 0x01F502, 0x00000E}, {0x01F503, 0x01F503, 0x00000E}, {0x01F504, 0x01F507, 0x00000E},
    {0x01F508, 0x01F508, 0x00000E}, {0x01F509, 0x01F509, 0x00000E}, {0x01F50A, 0x01F514, 0x00000E},
    {0x01F515, 0x01F515, 0x00000E}, {0x01F516, 0x01F52B, 0x00000E}, {0x01F52C, 0x01F52D, 0x00000E},
    {0x01F52E, 0x01F53D, 0x00000E}, {0x01F546, 0x01F548, 0x00000E}, {0x01F549, 0x01F54A, 0x00000E},
    {0x01F54B, 0x01F54E, 0x00000E}, {0x01F54F, 0x01F54F, 0x00000E}, {0x01F550, 0x01F55B, 0x00000E},
    {0x01F55C, 0x01F567, 0x00000E}, {0x01F568, 0x01F56E, 0x00000E}, {0x01F56F, 0x01F570, 0x00000E},
    {0x01F571, 0x01F572, 0x00000E}, {0x01F573, 0x01F579, 0x00000E}, {0x01F57A, 0x01F57A, 0x00000E},
    {0x01F57B, 0x01F586, 0x00000E}, {0x01F587, 0x01F587, 0x00000E}, {0x01F588, 0x01F589, 0x00000E},
    {0x01F58A, 0x01F58D, 0x00000E}, {0x01F58E, 0x01F58F, 0x00000E}, {0x01F590, 0x01F590, 0x00000E},
    {0x01F591, 0x01F594, 0x00000E}, {0x01F595, 0x01F596, 0x00000E}, {0x01F597, 0x01F5A3, 0x00000E},
    {0x01F5A4, 0x01F5A4, 0x00000E}, {0x01F5A5, 0x01F5A5, 0x00000E}, {0x01F5A6, 0x01F5A7, 0x00000E},
    {0x01F5A8, 0x01F5A8, 0x00000E}, {0x01F5A9, 0x01F5B0, 0x00000E}, {0x01F5B1, 0x01F5B2, 0x00000E},
    {0x01F5B3, 0x01F5BB, 0x00000E}, {0x01F5BC, 0x01F5BC, 0x00000E}, {0x01F5BD, 0x01F5C1, 0x00000E},
    {0x01F5C2, 0x01F5C4, 0x00000E}, {0x01F5C5, 0x01F5D0, 0x00000E}, {0x01F5D1, 0x01F5D3, 0x00000E},
    {0x01F5D4, 0x01F5DB, 0x00000E}, {0x01F5DC, 0x01F5DE, 0x00000E}, {0x01F5DF, 0x01F5E0, 0x00000E},
    {0x01F5E1, 0x01F5E1, 0x00000E}, {0x01F5E2, 0x01F5E2, 0x00000E}, {0x01F5E3, 0x01F5E3, 0x00000E},
    {0x01F5E4, 0x01F5E7, 0x00000E}, {0x01F5E8, 0x01F5E8, 0x00000E}, {0x01F5E9, 0x01F5EE, 0x00000E},
    {0x01F5EF, 0x01F5EF, 0x00000E}, {0x01F5F0, 0x01F5F2, 0x00000E}, {0x01F5F3, 0x01F5F3, 0x00000E},
    {0x01F5F4, 0x01F5F9, 0x00000E}, {0x01F5FA, 0x01F5FA, 0x00000E}, {0x01F5FB, 0x01F5FF, 0x00000E},
    {0x01F600, 0x01F600, 0x00000E}, {0x01F601, 0x01F606, 0x00000E}, {0x01F607, 0x01F608, 0x00000E},
    {0x01F609, 0x01F60D, 0x00000E}, {0x01F60E, 0x01F60E, 0x00000E}, {0x01F60F, 0x01F60F, 0x00000E},
    {0x01F610, 0x01F610, 0x00000E}, {0x01F611, 0x01F611, 0x00000E}, {0x01F612, 0x01F614, 0x00000E},
    {0x01F615, 0x01F615, 0x00000E}, {0x01F616, 0x01F616, 0x00000E}, {0x01F617, 0x01F617, 0x00000E},
    {0x01F618, 0x01F618, 0x00000E}, {0x01F619, 0x01F619, 0x00000E}, {0x01F61A, 0x01F61A, 0x00000E},
    {0x01F61B, 0x01F61B, 0x00000E}, {0x01F61C, 0x01F61E, 0x00000E}, {0x01F61F, 0x01F61F, 0x00000E},
    {0x01F620, 0x01F625, 0x00000E}, {0x01F626, 0x01F627, 0x00000E}, {0x01F628, 0x01F62B, 0x00000E},
    {0x01F62C, 0x01F62C, 0x00000E}, {0x01F62D, 0x01F62D, 0x00000E}, {0x01F62E, 0x01F62F, 0x00000E},
    {0x01F630, 0x01F633, 0x00000E}, {0x01F634, 0x01F634, 0x00000E}, {0x01F635, 0x01F635, 0x00000E},
    {0x01F636, 0x01F636, 0x00000E}, {0x01F637, 0x01F640, 0x00000E}, {0x01F641, 0x01F644, 0x00000E},
    {0x01F645, 0x01F64F, 0x00000E}, {0x01F680, 0x01F680, 0x00000E}, {0x01F681, 0x01F682, 0x00000E},
    {0x01F683, 0x01F685, 0x00000E}, {0x01F686, 0x01F686, 0x00000E}, {0x01F687, 0x01F687, 0x00000E},
    {0x01F688, 0x01F688, 0x00000E}, {0x01F689, 0x01F689, 0x00000E}, {0x01F68A, 0x01F68B, 0x00000E},
    {0x01F68C, 0x01F68C, 0x00000E}, {0x01F68D, 0x01F68D, 0x00000E}, {0x01F68E, 0x01F68E, 0x00000E},
    {0x01F68F, 0x01F68F, 0x00000E}, {0x01F690, 0x01F690, 0x00000E}, {0x01F691, 0x01F693, 0x00000E},
    {0x01F694, 0x01F694, 0x00000E}, {0x01F695, 0x01F695, 0x00000E}, {0x01F696, 0x01F696, 0x00000E},
    {0x01F697, 0x01F697, 0x00000E}, {0x01F698, 0x01F698, 0x00000E}, {0x01F699, 0x01F69A, 0x00000E},
    {0x01F69B, 0x01F6A1, 0x00000E}, {0x01F6A2, 0x01F6A2, 0x00000E}, {0x01F6A3, 0x01F6A3, 0x00000E},
    {0x01F6A4, 0x01F6A5, 0x00000E}, {0x01F6A6, 0x01F6A6, 0x00000E}, {0x01F6A7, 0x01F6AD, 0x00000E},
    {0x01F6AE, 0x01F6B1, 0x00000E}, {0x01F6B2, 0x01F6B2, 0x00000E}, {0x01F6B3, 0x01F6B5, 0x00000E},
    {0x01F6B6, 0x01F6B6, 0x00000E}, {0x01F6B7, 0x01F6B8, 0x00000E}, {0x01F6B9, 0x01F6BE, 0x00000E},
    {0x01F6BF, 0x01F6BF, 0x00000E}, {0x01F6C0, 0x01F6C0, 0x00000E}, {0x01F6C1, 0x01F6C5, 0x00000E},
    {0x01F6C6, 0x01F6CA, 0x00000E}, {0x01F6CB, 0x01F6CB, 0x00000E}, {0x01F6CC, 0x01F6CC, 0x00000E},
    {0x01F6CD, 0x01F6CF, 0x00000E}, {0x01F6D0, 0x01F6D0, 0x00000E}, {0x01F6D1, 0x01F6D2, 0x00000E},
    {0x01F6D3, 0x01F6D4, 0x00000E}, {0x01F6D5, 0x01F6D5, 0x00000E}, {0x01F6D6, 0x01F6D7, 0x00000E},
    {0x01F6D8, 0x01F6DC, 0x00000E}, {0x01F6DD, 0x01F6DF, 0x00000E}, {0x01F6E0, 0x01F6E5, 0x00000E},
    {0x01F6E6, 0x01F6E8, 0x00000E}, {0x01F6E9, 0x01F6E9, 0x00000E}, {0x01F6EA, 0x01F6EA, 0x00000E},
    {0x01F6EB, 0x01F6EC, 0x00000E}, {0x01F6ED, 0x01F6EF, 0x00000E}, {0x01F6F0, 0x01F6F0, 0x00000E},
    {0x01F6F1, 0x01F6F2, 0x00000E}, {0x01F6F3, 0x01F6F3, 0x00000E}, {0x01F6F4, 0x01F6F6, 0x00000E},
    {0x01F6F7, 0x01F6F8, 0x00000E}, {0x01F6F9, 0x01F6F9, 0x00000E}, {0x01F6FA, 0x01F6FA, 0x00000E},
    {0x01F6FB, 0x01F6FC, 0x00000E}, {0x01F6FD, 0x01F6FF, 0x00000E}, {0x01F774, 0x01F77F, 0x00000E},
    {0x01F7D5, 0x01F7DF, 0x00000E}, {0x01F7E0, 0x01F7EB, 0x00000E}, {0x01F7EC, 0x01F7EF, 0x00000E},
    {0x01F7F0, 0x01F7F0, 0x00000E}, {0x01F7F1, 0x01F7FF, 0x00000E}, {0x01F80C, 0x01F80F, 0x00000E},
    {0x01F848, 0x01F84F, 0x00000E}, {0x01F85A, 0x01F85F, 0x00000E}, {0x01F888, 0x01F88F, 0x00000E},
    {0x01F8AE, 0x01F8FF, 0x00000E}, {0x01F90C, 0x01F90C, 0x00000E}, {0x01F90D, 0x01F90F, 0x00000E},
    {0x01F910, 0x01F918, 0x00000E}, {0x01F919, 0x01F91E, 0x00000E}, {0x01F91F, 0x01F91F, 0x00000E},
    {0x01F920, 0x01F927, 0x00000E}, {0x01F928, 0x01F92F, 0x00000E}, {0x01F930, 0x01F930, 0x00000E},
    {0x01F931, 0x01F932, 0x00000E}, {0x01F933, 0x01F93A, 0x00000E}, {0x01F93C, 0x01F93E, 0x00000E},
    {0x01F93F, 0x01F93F, 0x00000E}, {0x01F940, 0x01F945, 0x00000E}, {0x01F947, 0x01F94B, 0x00000E},
    {0x01F94C, 0x01F94C, 0x00000E}, {0x01F94D, 0x01F94F, 0x00000E}, {0x01F950, 0x01F95E, 0x00000E},
    {0x01F95F, 0x01F96B, 0x00000E}, {0x01F96C, 0x01F970, 0x00000E}, {0x01F971, 0x01F971, 0x00000E},
    {0x01F972, 0x01F972, 0x00000E}, {0x01F973, 0x01F976, 0x00000E}, {0x01F977, 0x01F978, 0x00000E},
    {0x01F979, 0x01F979, 0x00000E}, {0x01F97A, 0x01F97A, 0x00000E}, {0x01F97B, 0x01F97B, 0x00000E},
    {0x01F97C, 0x01F97F, 0x00000E}, {0x01F980, 0x01F984, 0x00000E}, {0x01F985, 0x01F991, 0x00000E},
    {0x01F992, 0x01F997, 0x00000E}, {0x01F998, 0x01F9A2, 0x00000E}, {0x01F9A3, 0x01F9A4, 0x00000E},
    {0x01F9A5, 0x01F9AA, 0x00000E}, {0x01F9AB, 0x01F9AD, 0x00000E}, {0x01F9AE, 0x01F9AF, 0x00000E},
    {0x01F9B0, 0x01F9B9, 0x00000E}, {0x01F9BA, 0x01F9BF, 0x00000E}, {0x01F9C0, 0x01F9C0, 0x00000E},
    {0x01F9C1, 0x01F9C2, 0x00000E}, {0x01F9C3, 0x01F9CA, 0x00000E}, {0x01F9CB, 0x01F9CB, 0x00000E},
    {0x01F9CC, 0x01F9CC, 0x00000E}, {0x01F9CD, 0x01F9CF, 0x00000E}, {0x01F9D0, 0x01F9E6, 0x00000E},
    {0x01F9E7, 0x01F9FF, 0x00000E}, {0x01FA00, 0x01FA6F, 0x00000E}, {0x01FA70, 0x01FA73, 0x00000E},
    {0x01FA74, 0x01FA74, 0x00000E}, {0x01FA75, 0x01FA77, 0x00000E}, {0x01FA78, 0x01FA7A, 0x00000E},
    {0x01FA7B, 0x01FA7C, 0x00000E}, {0x01FA7D, 0x01FA7F, 0x00000E}, {0x01FA80, 0x01FA82, 0x00000E},
    {0x01FA83, 0x01FA86, 0x00000E}, {0x01FA87, 0x01FA8F, 0x00000E}, {0x01FA90, 0x01FA95, 0x00000E},
    {0x01FA96, 0x01FAA8, 0x00000E}, {0x01FAA9, 0x01FAAC, 0x00000E}, {0x01FAAD, 0x01FAAF, 0x00000E},
    {0x01FAB0, 0x01FAB6, 0x00000E}, {0x01FAB7, 0x01FABA, 0x00000E}, {0x01FABB, 0x01FABF, 0x00000E},
    {0x01FAC0, 0x01FAC2, 0x00000E}, {0x01FAC3, 0x01FAC5, 0x00000E}, {0x01FAC6, 0x01FACF, 0x00000E},
    {0x01FAD0, 0x01FAD6, 0x00000E}, {0x01FAD7, 0x01FAD9, 0x00000E}, {0x01FADA, 0x01FADF, 0x00000E},
    {0x01FAE0, 0x01FAE7, 0x00000E}, {0x01FAE8, 0x01FAEF, 0x00000E}, {0x01FAF0, 0x01FAF6, 0x00000E},
    {0x01FAF7, 0x01FAFF, 0x00000E}, {0x01FC00, 0x01FFFD, 0x00000E}, {0x0E0000, 0x0E0000, 0x000003},
    {0x0E0001, 0x0E0001, 0x000003}, {0x0E0002, 0x0E001F, 0x000003}, {0x0E0020, 0x0E007F, 0x000004},
    {0x0E0080, 0x0E00FF, 0x000003}, {0x0E0100, 0x0E01EF, 0x000004}, {0x0E01F0, 0x0E0FFF, 0x000003},
}};
}  // namespace arg_router::utility::utf8
