// Copyright (C) 2022 by Camden Mannett.
// Distributed under the Boost Software License, Version 1.0.
// (See accompanying file LICENSE or copy at https://www.boost.org/LICENSE_1_0.txt)

#pragma once

#include "arg_router/utility/utf8/code_point.hpp"

namespace arg_router::utility::utf8
{
/** Line break classes, and their values in the encoded code points in line_break_table.
 *
 * Do not change the order or value as they need to match scripts/unicode_table_generators.py.
 */
enum class line_break_class : std::uint8_t {
    any,
    AL,
    BA,
    BB,
    B2,
    BK,
    CB,
    CL,
    CM,
    CP,
    CR,
    EB,
    EM,
    EX,
    GL,
    H2,
    H3,
    HY,
    ID,
    HL,
    IN_,  // Trailing underscore due to Windows having defined IN...
    IS,
    JL,
    JT,
    JV,
    LF,
    NL,
    NS,
    NU,
    OP,
    PO,
    PR,
    QU,
    RI,
    SP,
    SY,
    WJ,
    ZW,
    ZWJ,
};

namespace no_break_rules
{
/** Rule LB6, do not break before hard line breaks.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB6(
    [[maybe_unused]] const std::array<line_break_class, N>& trailing_window,
    line_break_class next_class) noexcept
{
    return (next_class == line_break_class::BK) ||  //
           (next_class == line_break_class::CR) ||  //
           (next_class == line_break_class::LF) ||  //
           (next_class == line_break_class::NL);
}

/** Rule LB7, do not break before spaces or zero width space.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB7(
    [[maybe_unused]] const std::array<line_break_class, N>& trailing_window,
    line_break_class next_class) noexcept
{
    return (next_class == line_break_class::SP) ||  //
           (next_class == line_break_class::ZW);
}

/** Rule LB8a and LB9, do not break after a combining character sequence or zero width joiner.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB8a_9(const std::array<line_break_class, N>& trailing_window,
                                    [[maybe_unused]] line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (trailing_window[0] == line_break_class::CM) ||
           (trailing_window[0] == line_break_class::ZWJ);
}

/** Rule LB11, do not break before or after Word joiner and related characters.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB11(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (trailing_window[0] == line_break_class::WJ) ||  //
           (next_class == line_break_class::WJ);
}

/** Rule LB12, do not break after NBSP and related characters.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB12(const std::array<line_break_class, N>& trailing_window,
                                  [[maybe_unused]] line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return trailing_window[0] == line_break_class::GL;
}

/** Rule LB12a, do not break before NBSP and related characters, except after spaces and hyphens.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB12a(const std::array<line_break_class, N>& trailing_window,
                                   line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return ((trailing_window[0] != line_break_class::SP) &&
            (trailing_window[0] != line_break_class::BA) &&
            (trailing_window[0] != line_break_class::HY)) &&
           (next_class == line_break_class::GL);
}

/** Rule LB13, do not break before ‘]’ or ‘!’ or ‘;’ or ‘/’, even after spaces.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB13(
    [[maybe_unused]] const std::array<line_break_class, N>& trailing_window,
    line_break_class next_class) noexcept
{
    return (next_class == line_break_class::CL) ||  //
           (next_class == line_break_class::CP) ||  //
           (next_class == line_break_class::EX) ||  //
           (next_class == line_break_class::IS) ||  //
           (next_class == line_break_class::SY);
}

/** Rule LB14, do not break after ‘[’, even after spaces.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB14(const std::array<line_break_class, N>& trailing_window,
                                  [[maybe_unused]] line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");

    auto it = trailing_window.begin();
    for (; it != trailing_window.end(); ++it) {
        if (*it == line_break_class::SP) {
            continue;
        }

        return *it == line_break_class::OP;
    }

    return false;
}

/** Rule LB15, do not break within ‘”[’, even with intervening spaces.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB15(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");

    auto it = trailing_window.begin();
    for (; it != trailing_window.end(); ++it) {
        if (*it == line_break_class::SP) {
            continue;
        }

        return (*it == line_break_class::QU) &&  //
               (next_class == line_break_class::OP);
    }

    return false;
}

/** Rule LB16, do not break between closing punctuation and a nonstarter (<TT>NS</TT>), even with
 * intervening spaces.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB16(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");

    auto it = trailing_window.begin();
    for (; it != trailing_window.end(); ++it) {
        if (*it == line_break_class::SP) {
            continue;
        }

        return ((*it == line_break_class::CL) || (*it == line_break_class::CP)) &&
               (next_class == line_break_class::NS);
    }

    return false;
}

/** Rule LB17, do not break within ‘——’, even with intervening spaces.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB17(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");

    auto it = trailing_window.begin();
    for (; it != trailing_window.end(); ++it) {
        if (*it == line_break_class::SP) {
            continue;
        }

        return (*it == line_break_class::B2) &&  //
               (next_class == line_break_class::B2);
    }

    return false;
}

/** Rule LB19, do not break before or after quotation marks, such as ‘”’.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB19(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (next_class == line_break_class::QU) ||  //
           (trailing_window[0] == line_break_class::QU);
}

/** Rule LB21, do not break before hyphen-minus, other hyphens, fixed-width spaces, small kana, and
 * other non-starters, or after acute accents.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB21(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (next_class == line_break_class::BA) ||  //
           (next_class == line_break_class::HY) ||  //
           (next_class == line_break_class::NS) ||  //
           (trailing_window[0] == line_break_class::BB);
}

/** Rule LB21a, do not break after Hebrew + Hyphen.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB21a(const std::array<line_break_class, N>& trailing_window,
                                   [[maybe_unused]] line_break_class next_class) noexcept
{
    static_assert(N > 1, "Trailing window must be at least 2 elements");
    return (trailing_window[1] == line_break_class::HL) &&
           ((trailing_window[0] == line_break_class::HY) ||
            (trailing_window[0] == line_break_class::BA));
}

/** Rule LB21b, do not break between Solidus and Hebrew letters.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB21b(const std::array<line_break_class, N>& trailing_window,
                                   line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (trailing_window[0] == line_break_class::SY) &&  //
           (next_class == line_break_class::HL);
}

/** Rule LB22, do not break before ellipses.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB22(
    [[maybe_unused]] const std::array<line_break_class, N>& trailing_window,
    line_break_class next_class) noexcept
{
    return next_class == line_break_class::IN_;
}

/** Rule LB23, do not break between digits and letters.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB23(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (((trailing_window[0] == line_break_class::AL) ||
             (trailing_window[0] == line_break_class::HL)) &&
            (next_class == line_break_class::NU)) ||
           ((trailing_window[0] == line_break_class::NU) &&
            ((next_class == line_break_class::AL) ||  //
             (next_class == line_break_class::HL)));
}

/** Rule LB23a, do not break between numeric prefixes and ideographs, or between ideographs and
 * numeric postfixes.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB23a(const std::array<line_break_class, N>& trailing_window,
                                   line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return ((trailing_window[0] == line_break_class::PR) &&
            ((next_class == line_break_class::ID) ||          //
             (next_class == line_break_class::EB) ||          //
             (next_class == line_break_class::EM))) ||        //
           (((trailing_window[0] == line_break_class::ID) ||  //
             (trailing_window[0] == line_break_class::EB) ||  //
             (trailing_window[0] == line_break_class::EM)) &&
            (next_class == line_break_class::PO));
}

/** Rule LB24, do not break between numeric prefixes and ideographs, or between ideographs and
 * numeric postfixes.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB24(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (((trailing_window[0] == line_break_class::PR) ||
             (trailing_window[0] == line_break_class::PO)) &&
            ((next_class == line_break_class::AL) ||  //
             (next_class == line_break_class::HL))) ||
           (((trailing_window[0] == line_break_class::AL) ||
             (trailing_window[0] == line_break_class::HL)) &&
            ((next_class == line_break_class::PR) ||  //
             (next_class == line_break_class::PO)));
}

/** Rule LB25, do not break between these pairs of classes relevant to numbers.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB25(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");

    constexpr auto pairs = std::array{
        std::array{line_break_class::CL, line_break_class::PO},
        std::array{line_break_class::CP, line_break_class::PO},
        std::array{line_break_class::CL, line_break_class::PR},
        std::array{line_break_class::CP, line_break_class::PR},
        std::array{line_break_class::NU, line_break_class::PO},
        std::array{line_break_class::NU, line_break_class::PR},
        std::array{line_break_class::PO, line_break_class::OP},
        std::array{line_break_class::PO, line_break_class::NU},
        std::array{line_break_class::PR, line_break_class::OP},
        std::array{line_break_class::PR, line_break_class::NU},
        std::array{line_break_class::HY, line_break_class::NU},
        std::array{line_break_class::IS, line_break_class::NU},
        std::array{line_break_class::NU, line_break_class::NU},
        std::array{line_break_class::SY, line_break_class::NU},
    };

    for (const auto pr : pairs) {
        if ((trailing_window[0] == pr[0]) && (next_class == pr[1])) {
            return true;
        }
    }

    return false;
}

/** Rule LB26, do not break a Korean syllable.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB26(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return ((trailing_window[0] == line_break_class::JL) &&  //
            ((next_class == line_break_class::JL) ||         //
             (next_class == line_break_class::JV) ||         //
             (next_class == line_break_class::H2) ||         //
             (next_class == line_break_class::H3))) ||
           (((trailing_window[0] == line_break_class::JV) ||  //
             (trailing_window[0] == line_break_class::H2)) &&
            ((next_class == line_break_class::JV) ||  //
             (next_class == line_break_class::JT))) ||
           (((trailing_window[0] == line_break_class::JT) ||   //
             (trailing_window[0] == line_break_class::H3)) &&  //
            (next_class == line_break_class::JT));
}

/** Rule LB27, treat a Korean Syllable Block the same as ID.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB27(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (((trailing_window[0] == line_break_class::JL) ||  //
             (trailing_window[0] == line_break_class::JV) ||  //
             (trailing_window[0] == line_break_class::JT) ||  //
             (trailing_window[0] == line_break_class::H2) ||  //
             (trailing_window[0] == line_break_class::H3)) &&
            (next_class == line_break_class::PO)) ||
           ((trailing_window[0] == line_break_class::PR) &&  //
            ((next_class == line_break_class::JL) ||         //
             (next_class == line_break_class::JV) ||         //
             (next_class == line_break_class::JT) ||         //
             (next_class == line_break_class::H2) ||         //
             (next_class == line_break_class::H3)));
}

/** Rule LB28, do not break between alphabetics (“at”).
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB28(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return ((trailing_window[0] == line_break_class::AL) ||  //
            (trailing_window[0] == line_break_class::HL)) &&
           ((next_class == line_break_class::AL) ||  //
            (next_class == line_break_class::HL));
}

/** Rule LB29, do not break between numeric punctuation and alphabetics (“e.g.”).
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB29(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (trailing_window[0] == line_break_class::IS) &&
           ((next_class == line_break_class::AL) ||  //
            (next_class == line_break_class::HL));
}

/** Rule LB30, do not break between letters, numbers, or ordinary symbols and opening or closing
 * parentheses.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 *
 * @note This is not a true LB30 because it doesn't take into account far east asian parentheses
 * styles, so it can cause a break between a latin alphanumeric character and a far east asian
 * open parentheses for example
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB30(const std::array<line_break_class, N>& trailing_window,
                                  line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (((trailing_window[0] == line_break_class::AL) ||  //
             (trailing_window[0] == line_break_class::HL) ||  //
             (trailing_window[0] == line_break_class::NU)) &&
            (next_class == line_break_class::OP)) ||         //
           ((trailing_window[0] == line_break_class::CP) &&  //
            ((next_class == line_break_class::AL) ||         //
             (next_class == line_break_class::HL) ||         //
             (next_class == line_break_class::NU)));
}

/** Rule LB30a, break between two regional indicator symbols if and only if there are an even number
 * of regional indicators preceding the position of the break.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB30a(const std::array<line_break_class, N>& trailing_window,
                                   line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");

    // Count the preceding RIs, if there's an odd number - do not break
    auto i = 0u;
    for (; i < trailing_window.size(); ++i) {
        if (trailing_window[i] != line_break_class::RI) {
            break;
        }
    }

    return ((i % 2) != 0) && (next_class == line_break_class::RI);
}

/** Rule LB30b, do not break between an emoji base (or potential emoji) and an emoji modifier.
 *
 * From https://www.unicode.org/reports/tr14/#Algorithm.
 *
 * @note This is not a true LB30b as it doesn't take into account extended pictographic
 * @tparam N Size of trailing window
 * @param trailing_window Trailing window of properties, the zeroth index is the 'current' class
 * begin iterated over
 * @param next_class The class after the current one
 * @return True if you should not break between these break classes
 */
template <std::size_t N>
[[nodiscard]] constexpr bool LB30b(const std::array<line_break_class, N>& trailing_window,
                                   line_break_class next_class) noexcept
{
    static_assert(N > 0, "Trailing window must be at least 1 element");
    return (trailing_window[0] == line_break_class::EB) &&  //
           (next_class == line_break_class::EM);
}

/** Array of rule function pointers, just to ease iteration.
 *
 * @tparam N Size of trailing window
 */
template <std::size_t N>
constexpr auto line_break = std::array{
    &LB6<N>,  &LB7<N>,  &LB8a_9<N>, &LB11<N>, &LB12<N>,  &LB12a<N>, &LB13<N>, &LB14<N>,  &LB15<N>,
    &LB16<N>, &LB17<N>, &LB19<N>,   &LB21<N>, &LB21a<N>, &LB21b<N>, &LB22<N>, &LB23<N>,  &LB23a<N>,
    &LB24<N>, &LB25<N>, &LB26<N>,   &LB27<N>, &LB28<N>,  &LB29<N>,  &LB30<N>, &LB30a<N>, &LB30b<N>,
};
}  // namespace no_break_rules

/** Line break class table.
 *
 * Each entry is an inclusive range of code points, with each range's break class value encoded as
 * metadata.
 *
 * This table is generated using scripts/unicode_table_generators.py from
 * https://www.unicode.org/Public/UCD/latest/ucd/LineBreak.txt v14.0.0
 */
constexpr auto line_break_table = std::array<code_point::range, 3500>{{
    {0x000000, 0x000008, 0x000008}, {0x000009, 0x000009, 0x000002}, {0x00000A, 0x00000A, 0x000019},
    {0x00000B, 0x00000C, 0x000005}, {0x00000D, 0x00000D, 0x00000A}, {0x00000E, 0x00001F, 0x000008},
    {0x000020, 0x000020, 0x000022}, {0x000021, 0x000021, 0x00000D}, {0x000022, 0x000022, 0x000020},
    {0x000023, 0x000023, 0x000001}, {0x000024, 0x000024, 0x00001F}, {0x000025, 0x000025, 0x00001E},
    {0x000026, 0x000026, 0x000001}, {0x000027, 0x000027, 0x000020}, {0x000028, 0x000028, 0x00001D},
    {0x000029, 0x000029, 0x000009}, {0x00002A, 0x00002A, 0x000001}, {0x00002B, 0x00002B, 0x00001F},
    {0x00002C, 0x00002C, 0x000015}, {0x00002D, 0x00002D, 0x000011}, {0x00002E, 0x00002E, 0x000015},
    {0x00002F, 0x00002F, 0x000023}, {0x000030, 0x000039, 0x00001C}, {0x00003A, 0x00003B, 0x000015},
    {0x00003C, 0x00003E, 0x000001}, {0x00003F, 0x00003F, 0x00000D}, {0x000040, 0x000040, 0x000001},
    {0x000041, 0x00005A, 0x000001}, {0x00005B, 0x00005B, 0x00001D}, {0x00005C, 0x00005C, 0x00001F},
    {0x00005D, 0x00005D, 0x000009}, {0x00005E, 0x00005E, 0x000001}, {0x00005F, 0x00005F, 0x000001},
    {0x000060, 0x000060, 0x000001}, {0x000061, 0x00007A, 0x000001}, {0x00007B, 0x00007B, 0x00001D},
    {0x00007C, 0x00007C, 0x000002}, {0x00007D, 0x00007D, 0x000007}, {0x00007E, 0x00007E, 0x000001},
    {0x00007F, 0x00007F, 0x000008}, {0x000080, 0x000084, 0x000008}, {0x000085, 0x000085, 0x00001A},
    {0x000086, 0x00009F, 0x000008}, {0x0000A0, 0x0000A0, 0x00000E}, {0x0000A1, 0x0000A1, 0x00001D},
    {0x0000A2, 0x0000A2, 0x00001E}, {0x0000A3, 0x0000A5, 0x00001F}, {0x0000A6, 0x0000A6, 0x000001},
    {0x0000A7, 0x0000A7, 0x000001}, {0x0000A8, 0x0000A8, 0x000001}, {0x0000A9, 0x0000A9, 0x000001},
    {0x0000AA, 0x0000AA, 0x000001}, {0x0000AB, 0x0000AB, 0x000020}, {0x0000AC, 0x0000AC, 0x000001},
    {0x0000AD, 0x0000AD, 0x000002}, {0x0000AE, 0x0000AE, 0x000001}, {0x0000AF, 0x0000AF, 0x000001},
    {0x0000B0, 0x0000B0, 0x00001E}, {0x0000B1, 0x0000B1, 0x00001F}, {0x0000B2, 0x0000B3, 0x000001},
    {0x0000B4, 0x0000B4, 0x000003}, {0x0000B5, 0x0000B5, 0x000001}, {0x0000B6, 0x0000B7, 0x000001},
    {0x0000B8, 0x0000B8, 0x000001}, {0x0000B9, 0x0000B9, 0x000001}, {0x0000BA, 0x0000BA, 0x000001},
    {0x0000BB, 0x0000BB, 0x000020}, {0x0000BC, 0x0000BE, 0x000001}, {0x0000BF, 0x0000BF, 0x00001D},
    {0x0000C0, 0x0000D6, 0x000001}, {0x0000D7, 0x0000D7, 0x000001}, {0x0000D8, 0x0000F6, 0x000001},
    {0x0000F7, 0x0000F7, 0x000001}, {0x0000F8, 0x0000FF, 0x000001}, {0x000100, 0x00017F, 0x000001},
    {0x000180, 0x0001BA, 0x000001}, {0x0001BB, 0x0001BB, 0x000001}, {0x0001BC, 0x0001BF, 0x000001},
    {0x0001C0, 0x0001C3, 0x000001}, {0x0001C4, 0x00024F, 0x000001}, {0x000250, 0x000293, 0x000001},
    {0x000294, 0x000294, 0x000001}, {0x000295, 0x0002AF, 0x000001}, {0x0002B0, 0x0002C1, 0x000001},
    {0x0002C2, 0x0002C5, 0x000001}, {0x0002C6, 0x0002C6, 0x000001}, {0x0002C7, 0x0002C7, 0x000001},
    {0x0002C8, 0x0002C8, 0x000003}, {0x0002C9, 0x0002CB, 0x000001}, {0x0002CC, 0x0002CC, 0x000003},
    {0x0002CD, 0x0002CD, 0x000001}, {0x0002CE, 0x0002CF, 0x000001}, {0x0002D0, 0x0002D0, 0x000001},
    {0x0002D1, 0x0002D1, 0x000001}, {0x0002D2, 0x0002D7, 0x000001}, {0x0002D8, 0x0002DB, 0x000001},
    {0x0002DC, 0x0002DC, 0x000001}, {0x0002DD, 0x0002DD, 0x000001}, {0x0002DE, 0x0002DE, 0x000001},
    {0x0002DF, 0x0002DF, 0x000003}, {0x0002E0, 0x0002E4, 0x000001}, {0x0002E5, 0x0002EB, 0x000001},
    {0x0002EC, 0x0002EC, 0x000001}, {0x0002ED, 0x0002ED, 0x000001}, {0x0002EE, 0x0002EE, 0x000001},
    {0x0002EF, 0x0002FF, 0x000001}, {0x000300, 0x00034E, 0x000008}, {0x00034F, 0x00034F, 0x00000E},
    {0x000350, 0x00035B, 0x000008}, {0x00035C, 0x000362, 0x00000E}, {0x000363, 0x00036F, 0x000008},
    {0x000370, 0x000373, 0x000001}, {0x000374, 0x000374, 0x000001}, {0x000375, 0x000375, 0x000001},
    {0x000376, 0x000377, 0x000001}, {0x00037A, 0x00037A, 0x000001}, {0x00037B, 0x00037D, 0x000001},
    {0x00037E, 0x00037E, 0x000015}, {0x00037F, 0x00037F, 0x000001}, {0x000384, 0x000385, 0x000001},
    {0x000386, 0x000386, 0x000001}, {0x000387, 0x000387, 0x000001}, {0x000388, 0x00038A, 0x000001},
    {0x00038C, 0x00038C, 0x000001}, {0x00038E, 0x0003A1, 0x000001}, {0x0003A3, 0x0003F5, 0x000001},
    {0x0003F6, 0x0003F6, 0x000001}, {0x0003F7, 0x0003FF, 0x000001}, {0x000400, 0x000481, 0x000001},
    {0x000482, 0x000482, 0x000001}, {0x000483, 0x000487, 0x000008}, {0x000488, 0x000489, 0x000008},
    {0x00048A, 0x0004FF, 0x000001}, {0x000500, 0x00052F, 0x000001}, {0x000531, 0x000556, 0x000001},
    {0x000559, 0x000559, 0x000001}, {0x00055A, 0x00055F, 0x000001}, {0x000560, 0x000588, 0x000001},
    {0x000589, 0x000589, 0x000015}, {0x00058A, 0x00058A, 0x000002}, {0x00058D, 0x00058E, 0x000001},
    {0x00058F, 0x00058F, 0x00001F}, {0x000591, 0x0005BD, 0x000008}, {0x0005BE, 0x0005BE, 0x000002},
    {0x0005BF, 0x0005BF, 0x000008}, {0x0005C0, 0x0005C0, 0x000001}, {0x0005C1, 0x0005C2, 0x000008},
    {0x0005C3, 0x0005C3, 0x000001}, {0x0005C4, 0x0005C5, 0x000008}, {0x0005C6, 0x0005C6, 0x00000D},
    {0x0005C7, 0x0005C7, 0x000008}, {0x0005D0, 0x0005EA, 0x000013}, {0x0005EF, 0x0005F2, 0x000013},
    {0x0005F3, 0x0005F4, 0x000001}, {0x000600, 0x000605, 0x000001}, {0x000606, 0x000608, 0x000001},
    {0x000609, 0x00060A, 0x00001E}, {0x00060B, 0x00060B, 0x00001E}, {0x00060C, 0x00060D, 0x000015},
    {0x00060E, 0x00060F, 0x000001}, {0x000610, 0x00061A, 0x000008}, {0x00061B, 0x00061B, 0x00000D},
    {0x00061C, 0x00061C, 0x000008}, {0x00061D, 0x00061F, 0x00000D}, {0x000620, 0x00063F, 0x000001},
    {0x000640, 0x000640, 0x000001}, {0x000641, 0x00064A, 0x000001}, {0x00064B, 0x00065F, 0x000008},
    {0x000660, 0x000669, 0x00001C}, {0x00066A, 0x00066A, 0x00001E}, {0x00066B, 0x00066C, 0x00001C},
    {0x00066D, 0x00066D, 0x000001}, {0x00066E, 0x00066F, 0x000001}, {0x000670, 0x000670, 0x000008},
    {0x000671, 0x0006D3, 0x000001}, {0x0006D4, 0x0006D4, 0x00000D}, {0x0006D5, 0x0006D5, 0x000001},
    {0x0006D6, 0x0006DC, 0x000008}, {0x0006DD, 0x0006DD, 0x000001}, {0x0006DE, 0x0006DE, 0x000001},
    {0x0006DF, 0x0006E4, 0x000008}, {0x0006E5, 0x0006E6, 0x000001}, {0x0006E7, 0x0006E8, 0x000008},
    {0x0006E9, 0x0006E9, 0x000001}, {0x0006EA, 0x0006ED, 0x000008}, {0x0006EE, 0x0006EF, 0x000001},
    {0x0006F0, 0x0006F9, 0x00001C}, {0x0006FA, 0x0006FC, 0x000001}, {0x0006FD, 0x0006FE, 0x000001},
    {0x0006FF, 0x0006FF, 0x000001}, {0x000700, 0x00070D, 0x000001}, {0x00070F, 0x00070F, 0x000001},
    {0x000710, 0x000710, 0x000001}, {0x000711, 0x000711, 0x000008}, {0x000712, 0x00072F, 0x000001},
    {0x000730, 0x00074A, 0x000008}, {0x00074D, 0x00074F, 0x000001}, {0x000750, 0x00077F, 0x000001},
    {0x000780, 0x0007A5, 0x000001}, {0x0007A6, 0x0007B0, 0x000008}, {0x0007B1, 0x0007B1, 0x000001},
    {0x0007C0, 0x0007C9, 0x00001C}, {0x0007CA, 0x0007EA, 0x000001}, {0x0007EB, 0x0007F3, 0x000008},
    {0x0007F4, 0x0007F5, 0x000001}, {0x0007F6, 0x0007F6, 0x000001}, {0x0007F7, 0x0007F7, 0x000001},
    {0x0007F8, 0x0007F8, 0x000015}, {0x0007F9, 0x0007F9, 0x00000D}, {0x0007FA, 0x0007FA, 0x000001},
    {0x0007FD, 0x0007FD, 0x000008}, {0x0007FE, 0x0007FF, 0x00001F}, {0x000800, 0x000815, 0x000001},
    {0x000816, 0x000819, 0x000008}, {0x00081A, 0x00081A, 0x000001}, {0x00081B, 0x000823, 0x000008},
    {0x000824, 0x000824, 0x000001}, {0x000825, 0x000827, 0x000008}, {0x000828, 0x000828, 0x000001},
    {0x000829, 0x00082D, 0x000008}, {0x000830, 0x00083E, 0x000001}, {0x000840, 0x000858, 0x000001},
    {0x000859, 0x00085B, 0x000008}, {0x00085E, 0x00085E, 0x000001}, {0x000860, 0x00086A, 0x000001},
    {0x000870, 0x000887, 0x000001}, {0x000888, 0x000888, 0x000001}, {0x000889, 0x00088E, 0x000001},
    {0x000890, 0x000891, 0x000001}, {0x000898, 0x00089F, 0x000008}, {0x0008A0, 0x0008C8, 0x000001},
    {0x0008C9, 0x0008C9, 0x000001}, {0x0008CA, 0x0008E1, 0x000008}, {0x0008E2, 0x0008E2, 0x000001},
    {0x0008E3, 0x0008FF, 0x000008}, {0x000900, 0x000902, 0x000008}, {0x000903, 0x000903, 0x000008},
    {0x000904, 0x000939, 0x000001}, {0x00093A, 0x00093A, 0x000008}, {0x00093B, 0x00093B, 0x000008},
    {0x00093C, 0x00093C, 0x000008}, {0x00093D, 0x00093D, 0x000001}, {0x00093E, 0x000940, 0x000008},
    {0x000941, 0x000948, 0x000008}, {0x000949, 0x00094C, 0x000008}, {0x00094D, 0x00094D, 0x000008},
    {0x00094E, 0x00094F, 0x000008}, {0x000950, 0x000950, 0x000001}, {0x000951, 0x000957, 0x000008},
    {0x000958, 0x000961, 0x000001}, {0x000962, 0x000963, 0x000008}, {0x000964, 0x000965, 0x000002},
    {0x000966, 0x00096F, 0x00001C}, {0x000970, 0x000970, 0x000001}, {0x000971, 0x000971, 0x000001},
    {0x000972, 0x00097F, 0x000001}, {0x000980, 0x000980, 0x000001}, {0x000981, 0x000981, 0x000008},
    {0x000982, 0x000983, 0x000008}, {0x000985, 0x00098C, 0x000001}, {0x00098F, 0x000990, 0x000001},
    {0x000993, 0x0009A8, 0x000001}, {0x0009AA, 0x0009B0, 0x000001}, {0x0009B2, 0x0009B2, 0x000001},
    {0x0009B6, 0x0009B9, 0x000001}, {0x0009BC, 0x0009BC, 0x000008}, {0x0009BD, 0x0009BD, 0x000001},
    {0x0009BE, 0x0009C0, 0x000008}, {0x0009C1, 0x0009C4, 0x000008}, {0x0009C7, 0x0009C8, 0x000008},
    {0x0009CB, 0x0009CC, 0x000008}, {0x0009CD, 0x0009CD, 0x000008}, {0x0009CE, 0x0009CE, 0x000001},
    {0x0009D7, 0x0009D7, 0x000008}, {0x0009DC, 0x0009DD, 0x000001}, {0x0009DF, 0x0009E1, 0x000001},
    {0x0009E2, 0x0009E3, 0x000008}, {0x0009E6, 0x0009EF, 0x00001C}, {0x0009F0, 0x0009F1, 0x000001},
    {0x0009F2, 0x0009F3, 0x00001E}, {0x0009F4, 0x0009F8, 0x000001}, {0x0009F9, 0x0009F9, 0x00001E},
    {0x0009FA, 0x0009FA, 0x000001}, {0x0009FB, 0x0009FB, 0x00001F}, {0x0009FC, 0x0009FC, 0x000001},
    {0x0009FD, 0x0009FD, 0x000001}, {0x0009FE, 0x0009FE, 0x000008}, {0x000A01, 0x000A02, 0x000008},
    {0x000A03, 0x000A03, 0x000008}, {0x000A05, 0x000A0A, 0x000001}, {0x000A0F, 0x000A10, 0x000001},
    {0x000A13, 0x000A28, 0x000001}, {0x000A2A, 0x000A30, 0x000001}, {0x000A32, 0x000A33, 0x000001},
    {0x000A35, 0x000A36, 0x000001}, {0x000A38, 0x000A39, 0x000001}, {0x000A3C, 0x000A3C, 0x000008},
    {0x000A3E, 0x000A40, 0x000008}, {0x000A41, 0x000A42, 0x000008}, {0x000A47, 0x000A48, 0x000008},
    {0x000A4B, 0x000A4D, 0x000008}, {0x000A51, 0x000A51, 0x000008}, {0x000A59, 0x000A5C, 0x000001},
    {0x000A5E, 0x000A5E, 0x000001}, {0x000A66, 0x000A6F, 0x00001C}, {0x000A70, 0x000A71, 0x000008},
    {0x000A72, 0x000A74, 0x000001}, {0x000A75, 0x000A75, 0x000008}, {0x000A76, 0x000A76, 0x000001},
    {0x000A81, 0x000A82, 0x000008}, {0x000A83, 0x000A83, 0x000008}, {0x000A85, 0x000A8D, 0x000001},
    {0x000A8F, 0x000A91, 0x000001}, {0x000A93, 0x000AA8, 0x000001}, {0x000AAA, 0x000AB0, 0x000001},
    {0x000AB2, 0x000AB3, 0x000001}, {0x000AB5, 0x000AB9, 0x000001}, {0x000ABC, 0x000ABC, 0x000008},
    {0x000ABD, 0x000ABD, 0x000001}, {0x000ABE, 0x000AC0, 0x000008}, {0x000AC1, 0x000AC5, 0x000008},
    {0x000AC7, 0x000AC8, 0x000008}, {0x000AC9, 0x000AC9, 0x000008}, {0x000ACB, 0x000ACC, 0x000008},
    {0x000ACD, 0x000ACD, 0x000008}, {0x000AD0, 0x000AD0, 0x000001}, {0x000AE0, 0x000AE1, 0x000001},
    {0x000AE2, 0x000AE3, 0x000008}, {0x000AE6, 0x000AEF, 0x00001C}, {0x000AF0, 0x000AF0, 0x000001},
    {0x000AF1, 0x000AF1, 0x00001F}, {0x000AF9, 0x000AF9, 0x000001}, {0x000AFA, 0x000AFF, 0x000008},
    {0x000B01, 0x000B01, 0x000008}, {0x000B02, 0x000B03, 0x000008}, {0x000B05, 0x000B0C, 0x000001},
    {0x000B0F, 0x000B10, 0x000001}, {0x000B13, 0x000B28, 0x000001}, {0x000B2A, 0x000B30, 0x000001},
    {0x000B32, 0x000B33, 0x000001}, {0x000B35, 0x000B39, 0x000001}, {0x000B3C, 0x000B3C, 0x000008},
    {0x000B3D, 0x000B3D, 0x000001}, {0x000B3E, 0x000B3E, 0x000008}, {0x000B3F, 0x000B3F, 0x000008},
    {0x000B40, 0x000B40, 0x000008}, {0x000B41, 0x000B44, 0x000008}, {0x000B47, 0x000B48, 0x000008},
    {0x000B4B, 0x000B4C, 0x000008}, {0x000B4D, 0x000B4D, 0x000008}, {0x000B55, 0x000B56, 0x000008},
    {0x000B57, 0x000B57, 0x000008}, {0x000B5C, 0x000B5D, 0x000001}, {0x000B5F, 0x000B61, 0x000001},
    {0x000B62, 0x000B63, 0x000008}, {0x000B66, 0x000B6F, 0x00001C}, {0x000B70, 0x000B70, 0x000001},
    {0x000B71, 0x000B71, 0x000001}, {0x000B72, 0x000B77, 0x000001}, {0x000B82, 0x000B82, 0x000008},
    {0x000B83, 0x000B83, 0x000001}, {0x000B85, 0x000B8A, 0x000001}, {0x000B8E, 0x000B90, 0x000001},
    {0x000B92, 0x000B95, 0x000001}, {0x000B99, 0x000B9A, 0x000001}, {0x000B9C, 0x000B9C, 0x000001},
    {0x000B9E, 0x000B9F, 0x000001}, {0x000BA3, 0x000BA4, 0x000001}, {0x000BA8, 0x000BAA, 0x000001},
    {0x000BAE, 0x000BB9, 0x000001}, {0x000BBE, 0x000BBF, 0x000008}, {0x000BC0, 0x000BC0, 0x000008},
    {0x000BC1, 0x000BC2, 0x000008}, {0x000BC6, 0x000BC8, 0x000008}, {0x000BCA, 0x000BCC, 0x000008},
    {0x000BCD, 0x000BCD, 0x000008}, {0x000BD0, 0x000BD0, 0x000001}, {0x000BD7, 0x000BD7, 0x000008},
    {0x000BE6, 0x000BEF, 0x00001C}, {0x000BF0, 0x000BF2, 0x000001}, {0x000BF3, 0x000BF8, 0x000001},
    {0x000BF9, 0x000BF9, 0x00001F}, {0x000BFA, 0x000BFA, 0x000001}, {0x000C00, 0x000C00, 0x000008},
    {0x000C01, 0x000C03, 0x000008}, {0x000C04, 0x000C04, 0x000008}, {0x000C05, 0x000C0C, 0x000001},
    {0x000C0E, 0x000C10, 0x000001}, {0x000C12, 0x000C28, 0x000001}, {0x000C2A, 0x000C39, 0x000001},
    {0x000C3C, 0x000C3C, 0x000008}, {0x000C3D, 0x000C3D, 0x000001}, {0x000C3E, 0x000C40, 0x000008},
    {0x000C41, 0x000C44, 0x000008}, {0x000C46, 0x000C48, 0x000008}, {0x000C4A, 0x000C4D, 0x000008},
    {0x000C55, 0x000C56, 0x000008}, {0x000C58, 0x000C5A, 0x000001}, {0x000C5D, 0x000C5D, 0x000001},
    {0x000C60, 0x000C61, 0x000001}, {0x000C62, 0x000C63, 0x000008}, {0x000C66, 0x000C6F, 0x00001C},
    {0x000C77, 0x000C77, 0x000003}, {0x000C78, 0x000C7E, 0x000001}, {0x000C7F, 0x000C7F, 0x000001},
    {0x000C80, 0x000C80, 0x000001}, {0x000C81, 0x000C81, 0x000008}, {0x000C82, 0x000C83, 0x000008},
    {0x000C84, 0x000C84, 0x000003}, {0x000C85, 0x000C8C, 0x000001}, {0x000C8E, 0x000C90, 0x000001},
    {0x000C92, 0x000CA8, 0x000001}, {0x000CAA, 0x000CB3, 0x000001}, {0x000CB5, 0x000CB9, 0x000001},
    {0x000CBC, 0x000CBC, 0x000008}, {0x000CBD, 0x000CBD, 0x000001}, {0x000CBE, 0x000CBE, 0x000008},
    {0x000CBF, 0x000CBF, 0x000008}, {0x000CC0, 0x000CC4, 0x000008}, {0x000CC6, 0x000CC6, 0x000008},
    {0x000CC7, 0x000CC8, 0x000008}, {0x000CCA, 0x000CCB, 0x000008}, {0x000CCC, 0x000CCD, 0x000008},
    {0x000CD5, 0x000CD6, 0x000008}, {0x000CDD, 0x000CDE, 0x000001}, {0x000CE0, 0x000CE1, 0x000001},
    {0x000CE2, 0x000CE3, 0x000008}, {0x000CE6, 0x000CEF, 0x00001C}, {0x000CF1, 0x000CF2, 0x000001},
    {0x000D00, 0x000D01, 0x000008}, {0x000D02, 0x000D03, 0x000008}, {0x000D04, 0x000D0C, 0x000001},
    {0x000D0E, 0x000D10, 0x000001}, {0x000D12, 0x000D3A, 0x000001}, {0x000D3B, 0x000D3C, 0x000008},
    {0x000D3D, 0x000D3D, 0x000001}, {0x000D3E, 0x000D40, 0x000008}, {0x000D41, 0x000D44, 0x000008},
    {0x000D46, 0x000D48, 0x000008}, {0x000D4A, 0x000D4C, 0x000008}, {0x000D4D, 0x000D4D, 0x000008},
    {0x000D4E, 0x000D4E, 0x000001}, {0x000D4F, 0x000D4F, 0x000001}, {0x000D54, 0x000D56, 0x000001},
    {0x000D57, 0x000D57, 0x000008}, {0x000D58, 0x000D5E, 0x000001}, {0x000D5F, 0x000D61, 0x000001},
    {0x000D62, 0x000D63, 0x000008}, {0x000D66, 0x000D6F, 0x00001C}, {0x000D70, 0x000D78, 0x000001},
    {0x000D79, 0x000D79, 0x00001E}, {0x000D7A, 0x000D7F, 0x000001}, {0x000D81, 0x000D81, 0x000008},
    {0x000D82, 0x000D83, 0x000008}, {0x000D85, 0x000D96, 0x000001}, {0x000D9A, 0x000DB1, 0x000001},
    {0x000DB3, 0x000DBB, 0x000001}, {0x000DBD, 0x000DBD, 0x000001}, {0x000DC0, 0x000DC6, 0x000001},
    {0x000DCA, 0x000DCA, 0x000008}, {0x000DCF, 0x000DD1, 0x000008}, {0x000DD2, 0x000DD4, 0x000008},
    {0x000DD6, 0x000DD6, 0x000008}, {0x000DD8, 0x000DDF, 0x000008}, {0x000DE6, 0x000DEF, 0x00001C},
    {0x000DF2, 0x000DF3, 0x000008}, {0x000DF4, 0x000DF4, 0x000001}, {0x000E01, 0x000E30, 0x000001},
    {0x000E31, 0x000E31, 0x000008}, {0x000E32, 0x000E33, 0x000001}, {0x000E34, 0x000E3A, 0x000008},
    {0x000E3F, 0x000E3F, 0x00001F}, {0x000E40, 0x000E45, 0x000001}, {0x000E46, 0x000E46, 0x000001},
    {0x000E47, 0x000E4E, 0x000008}, {0x000E4F, 0x000E4F, 0x000001}, {0x000E50, 0x000E59, 0x00001C},
    {0x000E5A, 0x000E5B, 0x000002}, {0x000E81, 0x000E82, 0x000001}, {0x000E84, 0x000E84, 0x000001},
    {0x000E86, 0x000E8A, 0x000001}, {0x000E8C, 0x000EA3, 0x000001}, {0x000EA5, 0x000EA5, 0x000001},
    {0x000EA7, 0x000EB0, 0x000001}, {0x000EB1, 0x000EB1, 0x000008}, {0x000EB2, 0x000EB3, 0x000001},
    {0x000EB4, 0x000EBC, 0x000008}, {0x000EBD, 0x000EBD, 0x000001}, {0x000EC0, 0x000EC4, 0x000001},
    {0x000EC6, 0x000EC6, 0x000001}, {0x000EC8, 0x000ECD, 0x000008}, {0x000ED0, 0x000ED9, 0x00001C},
    {0x000EDC, 0x000EDF, 0x000001}, {0x000F00, 0x000F00, 0x000001}, {0x000F01, 0x000F03, 0x000003},
    {0x000F04, 0x000F04, 0x000003}, {0x000F05, 0x000F05, 0x000001}, {0x000F06, 0x000F07, 0x000003},
    {0x000F08, 0x000F08, 0x00000E}, {0x000F09, 0x000F0A, 0x000003}, {0x000F0B, 0x000F0B, 0x000002},
    {0x000F0C, 0x000F0C, 0x00000E}, {0x000F0D, 0x000F11, 0x00000D}, {0x000F12, 0x000F12, 0x00000E},
    {0x000F13, 0x000F13, 0x000001}, {0x000F14, 0x000F14, 0x00000D}, {0x000F15, 0x000F17, 0x000001},
    {0x000F18, 0x000F19, 0x000008}, {0x000F1A, 0x000F1F, 0x000001}, {0x000F20, 0x000F29, 0x00001C},
    {0x000F2A, 0x000F33, 0x000001}, {0x000F34, 0x000F34, 0x000002}, {0x000F35, 0x000F35, 0x000008},
    {0x000F36, 0x000F36, 0x000001}, {0x000F37, 0x000F37, 0x000008}, {0x000F38, 0x000F38, 0x000001},
    {0x000F39, 0x000F39, 0x000008}, {0x000F3A, 0x000F3A, 0x00001D}, {0x000F3B, 0x000F3B, 0x000007},
    {0x000F3C, 0x000F3C, 0x00001D}, {0x000F3D, 0x000F3D, 0x000007}, {0x000F3E, 0x000F3F, 0x000008},
    {0x000F40, 0x000F47, 0x000001}, {0x000F49, 0x000F6C, 0x000001}, {0x000F71, 0x000F7E, 0x000008},
    {0x000F7F, 0x000F7F, 0x000002}, {0x000F80, 0x000F84, 0x000008}, {0x000F85, 0x000F85, 0x000002},
    {0x000F86, 0x000F87, 0x000008}, {0x000F88, 0x000F8C, 0x000001}, {0x000F8D, 0x000F97, 0x000008},
    {0x000F99, 0x000FBC, 0x000008}, {0x000FBE, 0x000FBF, 0x000002}, {0x000FC0, 0x000FC5, 0x000001},
    {0x000FC6, 0x000FC6, 0x000008}, {0x000FC7, 0x000FCC, 0x000001}, {0x000FCE, 0x000FCF, 0x000001},
    {0x000FD0, 0x000FD1, 0x000003}, {0x000FD2, 0x000FD2, 0x000002}, {0x000FD3, 0x000FD3, 0x000003},
    {0x000FD4, 0x000FD4, 0x000001}, {0x000FD5, 0x000FD8, 0x000001}, {0x000FD9, 0x000FDA, 0x00000E},
    {0x001000, 0x00102A, 0x000001}, {0x00102B, 0x00102C, 0x000008}, {0x00102D, 0x001030, 0x000008},
    {0x001031, 0x001031, 0x000008}, {0x001032, 0x001037, 0x000008}, {0x001038, 0x001038, 0x000008},
    {0x001039, 0x00103A, 0x000008}, {0x00103B, 0x00103C, 0x000008}, {0x00103D, 0x00103E, 0x000008},
    {0x00103F, 0x00103F, 0x000001}, {0x001040, 0x001049, 0x00001C}, {0x00104A, 0x00104B, 0x000002},
    {0x00104C, 0x00104F, 0x000001}, {0x001050, 0x001055, 0x000001}, {0x001056, 0x001057, 0x000008},
    {0x001058, 0x001059, 0x000008}, {0x00105A, 0x00105D, 0x000001}, {0x00105E, 0x001060, 0x000008},
    {0x001061, 0x001061, 0x000001}, {0x001062, 0x001064, 0x000008}, {0x001065, 0x001066, 0x000001},
    {0x001067, 0x00106D, 0x000008}, {0x00106E, 0x001070, 0x000001}, {0x001071, 0x001074, 0x000008},
    {0x001075, 0x001081, 0x000001}, {0x001082, 0x001082, 0x000008}, {0x001083, 0x001084, 0x000008},
    {0x001085, 0x001086, 0x000008}, {0x001087, 0x00108C, 0x000008}, {0x00108D, 0x00108D, 0x000008},
    {0x00108E, 0x00108E, 0x000001}, {0x00108F, 0x00108F, 0x000008}, {0x001090, 0x001099, 0x00001C},
    {0x00109A, 0x00109C, 0x000008}, {0x00109D, 0x00109D, 0x000008}, {0x00109E, 0x00109F, 0x000001},
    {0x0010A0, 0x0010C5, 0x000001}, {0x0010C7, 0x0010C7, 0x000001}, {0x0010CD, 0x0010CD, 0x000001},
    {0x0010D0, 0x0010FA, 0x000001}, {0x0010FB, 0x0010FB, 0x000001}, {0x0010FC, 0x0010FC, 0x000001},
    {0x0010FD, 0x0010FF, 0x000001}, {0x001100, 0x00115F, 0x000016}, {0x001160, 0x0011A7, 0x000018},
    {0x0011A8, 0x0011FF, 0x000017}, {0x001200, 0x001248, 0x000001}, {0x00124A, 0x00124D, 0x000001},
    {0x001250, 0x001256, 0x000001}, {0x001258, 0x001258, 0x000001}, {0x00125A, 0x00125D, 0x000001},
    {0x001260, 0x001288, 0x000001}, {0x00128A, 0x00128D, 0x000001}, {0x001290, 0x0012B0, 0x000001},
    {0x0012B2, 0x0012B5, 0x000001}, {0x0012B8, 0x0012BE, 0x000001}, {0x0012C0, 0x0012C0, 0x000001},
    {0x0012C2, 0x0012C5, 0x000001}, {0x0012C8, 0x0012D6, 0x000001}, {0x0012D8, 0x001310, 0x000001},
    {0x001312, 0x001315, 0x000001}, {0x001318, 0x00135A, 0x000001}, {0x00135D, 0x00135F, 0x000008},
    {0x001360, 0x001360, 0x000001}, {0x001361, 0x001361, 0x000002}, {0x001362, 0x001368, 0x000001},
    {0x001369, 0x00137C, 0x000001}, {0x001380, 0x00138F, 0x000001}, {0x001390, 0x001399, 0x000001},
    {0x0013A0, 0x0013F5, 0x000001}, {0x0013F8, 0x0013FD, 0x000001}, {0x001400, 0x001400, 0x000002},
    {0x001401, 0x00166C, 0x000001}, {0x00166D, 0x00166D, 0x000001}, {0x00166E, 0x00166E, 0x000001},
    {0x00166F, 0x00167F, 0x000001}, {0x001680, 0x001680, 0x000002}, {0x001681, 0x00169A, 0x000001},
    {0x00169B, 0x00169B, 0x00001D}, {0x00169C, 0x00169C, 0x000007}, {0x0016A0, 0x0016EA, 0x000001},
    {0x0016EB, 0x0016ED, 0x000002}, {0x0016EE, 0x0016F0, 0x000001}, {0x0016F1, 0x0016F8, 0x000001},
    {0x001700, 0x001711, 0x000001}, {0x001712, 0x001714, 0x000008}, {0x001715, 0x001715, 0x000008},
    {0x00171F, 0x00171F, 0x000001}, {0x001720, 0x001731, 0x000001}, {0x001732, 0x001733, 0x000008},
    {0x001734, 0x001734, 0x000008}, {0x001735, 0x001736, 0x000002}, {0x001740, 0x001751, 0x000001},
    {0x001752, 0x001753, 0x000008}, {0x001760, 0x00176C, 0x000001}, {0x00176E, 0x001770, 0x000001},
    {0x001772, 0x001773, 0x000008}, {0x001780, 0x0017B3, 0x000001}, {0x0017B4, 0x0017B5, 0x000008},
    {0x0017B6, 0x0017B6, 0x000008}, {0x0017B7, 0x0017BD, 0x000008}, {0x0017BE, 0x0017C5, 0x000008},
    {0x0017C6, 0x0017C6, 0x000008}, {0x0017C7, 0x0017C8, 0x000008}, {0x0017C9, 0x0017D3, 0x000008},
    {0x0017D4, 0x0017D5, 0x000002}, {0x0017D6, 0x0017D6, 0x00001B}, {0x0017D7, 0x0017D7, 0x000001},
    {0x0017D8, 0x0017D8, 0x000002}, {0x0017D9, 0x0017D9, 0x000001}, {0x0017DA, 0x0017DA, 0x000002},
    {0x0017DB, 0x0017DB, 0x00001F}, {0x0017DC, 0x0017DC, 0x000001}, {0x0017DD, 0x0017DD, 0x000008},
    {0x0017E0, 0x0017E9, 0x00001C}, {0x0017F0, 0x0017F9, 0x000001}, {0x001800, 0x001801, 0x000001},
    {0x001802, 0x001803, 0x00000D}, {0x001804, 0x001805, 0x000002}, {0x001806, 0x001806, 0x000003},
    {0x001807, 0x001807, 0x000001}, {0x001808, 0x001809, 0x00000D}, {0x00180A, 0x00180A, 0x000001},
    {0x00180B, 0x00180D, 0x000008}, {0x00180E, 0x00180E, 0x00000E}, {0x00180F, 0x00180F, 0x000008},
    {0x001810, 0x001819, 0x00001C}, {0x001820, 0x001842, 0x000001}, {0x001843, 0x001843, 0x000001},
    {0x001844, 0x001878, 0x000001}, {0x001880, 0x001884, 0x000001}, {0x001885, 0x001886, 0x000008},
    {0x001887, 0x0018A8, 0x000001}, {0x0018A9, 0x0018A9, 0x000008}, {0x0018AA, 0x0018AA, 0x000001},
    {0x0018B0, 0x0018F5, 0x000001}, {0x001900, 0x00191E, 0x000001}, {0x001920, 0x001922, 0x000008},
    {0x001923, 0x001926, 0x000008}, {0x001927, 0x001928, 0x000008}, {0x001929, 0x00192B, 0x000008},
    {0x001930, 0x001931, 0x000008}, {0x001932, 0x001932, 0x000008}, {0x001933, 0x001938, 0x000008},
    {0x001939, 0x00193B, 0x000008}, {0x001940, 0x001940, 0x000001}, {0x001944, 0x001945, 0x00000D},
    {0x001946, 0x00194F, 0x00001C}, {0x001950, 0x00196D, 0x000001}, {0x001970, 0x001974, 0x000001},
    {0x001980, 0x0019AB, 0x000001}, {0x0019B0, 0x0019C9, 0x000001}, {0x0019D0, 0x0019D9, 0x00001C},
    {0x0019DA, 0x0019DA, 0x000001}, {0x0019DE, 0x0019DF, 0x000001}, {0x0019E0, 0x0019FF, 0x000001},
    {0x001A00, 0x001A16, 0x000001}, {0x001A17, 0x001A18, 0x000008}, {0x001A19, 0x001A1A, 0x000008},
    {0x001A1B, 0x001A1B, 0x000008}, {0x001A1E, 0x001A1F, 0x000001}, {0x001A20, 0x001A54, 0x000001},
    {0x001A55, 0x001A55, 0x000008}, {0x001A56, 0x001A56, 0x000008}, {0x001A57, 0x001A57, 0x000008},
    {0x001A58, 0x001A5E, 0x000008}, {0x001A60, 0x001A60, 0x000008}, {0x001A61, 0x001A61, 0x000008},
    {0x001A62, 0x001A62, 0x000008}, {0x001A63, 0x001A64, 0x000008}, {0x001A65, 0x001A6C, 0x000008},
    {0x001A6D, 0x001A72, 0x000008}, {0x001A73, 0x001A7C, 0x000008}, {0x001A7F, 0x001A7F, 0x000008},
    {0x001A80, 0x001A89, 0x00001C}, {0x001A90, 0x001A99, 0x00001C}, {0x001AA0, 0x001AA6, 0x000001},
    {0x001AA7, 0x001AA7, 0x000001}, {0x001AA8, 0x001AAD, 0x000001}, {0x001AB0, 0x001ABD, 0x000008},
    {0x001ABE, 0x001ABE, 0x000008}, {0x001ABF, 0x001ACE, 0x000008}, {0x001B00, 0x001B03, 0x000008},
    {0x001B04, 0x001B04, 0x000008}, {0x001B05, 0x001B33, 0x000001}, {0x001B34, 0x001B34, 0x000008},
    {0x001B35, 0x001B35, 0x000008}, {0x001B36, 0x001B3A, 0x000008}, {0x001B3B, 0x001B3B, 0x000008},
    {0x001B3C, 0x001B3C, 0x000008}, {0x001B3D, 0x001B41, 0x000008}, {0x001B42, 0x001B42, 0x000008},
    {0x001B43, 0x001B44, 0x000008}, {0x001B45, 0x001B4C, 0x000001}, {0x001B50, 0x001B59, 0x00001C},
    {0x001B5A, 0x001B5B, 0x000002}, {0x001B5C, 0x001B5C, 0x000001}, {0x001B5D, 0x001B60, 0x000002},
    {0x001B61, 0x001B6A, 0x000001}, {0x001B6B, 0x001B73, 0x000008}, {0x001B74, 0x001B7C, 0x000001},
    {0x001B7D, 0x001B7E, 0x000002}, {0x001B80, 0x001B81, 0x000008}, {0x001B82, 0x001B82, 0x000008},
    {0x001B83, 0x001BA0, 0x000001}, {0x001BA1, 0x001BA1, 0x000008}, {0x001BA2, 0x001BA5, 0x000008},
    {0x001BA6, 0x001BA7, 0x000008}, {0x001BA8, 0x001BA9, 0x000008}, {0x001BAA, 0x001BAA, 0x000008},
    {0x001BAB, 0x001BAD, 0x000008}, {0x001BAE, 0x001BAF, 0x000001}, {0x001BB0, 0x001BB9, 0x00001C},
    {0x001BBA, 0x001BBF, 0x000001}, {0x001BC0, 0x001BE5, 0x000001}, {0x001BE6, 0x001BE6, 0x000008},
    {0x001BE7, 0x001BE7, 0x000008}, {0x001BE8, 0x001BE9, 0x000008}, {0x001BEA, 0x001BEC, 0x000008},
    {0x001BED, 0x001BED, 0x000008}, {0x001BEE, 0x001BEE, 0x000008}, {0x001BEF, 0x001BF1, 0x000008},
    {0x001BF2, 0x001BF3, 0x000008}, {0x001BFC, 0x001BFF, 0x000001}, {0x001C00, 0x001C23, 0x000001},
    {0x001C24, 0x001C2B, 0x000008}, {0x001C2C, 0x001C33, 0x000008}, {0x001C34, 0x001C35, 0x000008},
    {0x001C36, 0x001C37, 0x000008}, {0x001C3B, 0x001C3F, 0x000002}, {0x001C40, 0x001C49, 0x00001C},
    {0x001C4D, 0x001C4F, 0x000001}, {0x001C50, 0x001C59, 0x00001C}, {0x001C5A, 0x001C77, 0x000001},
    {0x001C78, 0x001C7D, 0x000001}, {0x001C7E, 0x001C7F, 0x000002}, {0x001C80, 0x001C88, 0x000001},
    {0x001C90, 0x001CBA, 0x000001}, {0x001CBD, 0x001CBF, 0x000001}, {0x001CC0, 0x001CC7, 0x000001},
    {0x001CD0, 0x001CD2, 0x000008}, {0x001CD3, 0x001CD3, 0x000001}, {0x001CD4, 0x001CE0, 0x000008},
    {0x001CE1, 0x001CE1, 0x000008}, {0x001CE2, 0x001CE8, 0x000008}, {0x001CE9, 0x001CEC, 0x000001},
    {0x001CED, 0x001CED, 0x000008}, {0x001CEE, 0x001CF3, 0x000001}, {0x001CF4, 0x001CF4, 0x000008},
    {0x001CF5, 0x001CF6, 0x000001}, {0x001CF7, 0x001CF7, 0x000008}, {0x001CF8, 0x001CF9, 0x000008},
    {0x001CFA, 0x001CFA, 0x000001}, {0x001D00, 0x001D2B, 0x000001}, {0x001D2C, 0x001D6A, 0x000001},
    {0x001D6B, 0x001D77, 0x000001}, {0x001D78, 0x001D78, 0x000001}, {0x001D79, 0x001D7F, 0x000001},
    {0x001D80, 0x001D9A, 0x000001}, {0x001D9B, 0x001DBF, 0x000001}, {0x001DC0, 0x001DFF, 0x000008},
    {0x001E00, 0x001EFF, 0x000001}, {0x001F00, 0x001F15, 0x000001}, {0x001F18, 0x001F1D, 0x000001},
    {0x001F20, 0x001F45, 0x000001}, {0x001F48, 0x001F4D, 0x000001}, {0x001F50, 0x001F57, 0x000001},
    {0x001F59, 0x001F59, 0x000001}, {0x001F5B, 0x001F5B, 0x000001}, {0x001F5D, 0x001F5D, 0x000001},
    {0x001F5F, 0x001F7D, 0x000001}, {0x001F80, 0x001FB4, 0x000001}, {0x001FB6, 0x001FBC, 0x000001},
    {0x001FBD, 0x001FBD, 0x000001}, {0x001FBE, 0x001FBE, 0x000001}, {0x001FBF, 0x001FC1, 0x000001},
    {0x001FC2, 0x001FC4, 0x000001}, {0x001FC6, 0x001FCC, 0x000001}, {0x001FCD, 0x001FCF, 0x000001},
    {0x001FD0, 0x001FD3, 0x000001}, {0x001FD6, 0x001FDB, 0x000001}, {0x001FDD, 0x001FDF, 0x000001},
    {0x001FE0, 0x001FEC, 0x000001}, {0x001FED, 0x001FEF, 0x000001}, {0x001FF2, 0x001FF4, 0x000001},
    {0x001FF6, 0x001FFC, 0x000001}, {0x001FFD, 0x001FFD, 0x000003}, {0x001FFE, 0x001FFE, 0x000001},
    {0x002000, 0x002006, 0x000002}, {0x002007, 0x002007, 0x00000E}, {0x002008, 0x00200A, 0x000002},
    {0x00200B, 0x00200B, 0x000025}, {0x00200C, 0x00200C, 0x000008}, {0x00200D, 0x00200D, 0x000026},
    {0x00200E, 0x00200F, 0x000008}, {0x002010, 0x002010, 0x000002}, {0x002011, 0x002011, 0x00000E},
    {0x002012, 0x002013, 0x000002}, {0x002014, 0x002014, 0x000004}, {0x002015, 0x002015, 0x000001},
    {0x002016, 0x002016, 0x000001}, {0x002017, 0x002017, 0x000001}, {0x002018, 0x002018, 0x000020},
    {0x002019, 0x002019, 0x000020}, {0x00201A, 0x00201A, 0x00001D}, {0x00201B, 0x00201C, 0x000020},
    {0x00201D, 0x00201D, 0x000020}, {0x00201E, 0x00201E, 0x00001D}, {0x00201F, 0x00201F, 0x000020},
    {0x002020, 0x002021, 0x000001}, {0x002022, 0x002023, 0x000001}, {0x002024, 0x002026, 0x000014},
    {0x002027, 0x002027, 0x000002}, {0x002028, 0x002028, 0x000005}, {0x002029, 0x002029, 0x000005},
    {0x00202A, 0x00202E, 0x000008}, {0x00202F, 0x00202F, 0x00000E}, {0x002030, 0x002037, 0x00001E},
    {0x002038, 0x002038, 0x000001}, {0x002039, 0x002039, 0x000020}, {0x00203A, 0x00203A, 0x000020},
    {0x00203B, 0x00203B, 0x000001}, {0x00203C, 0x00203D, 0x00001B}, {0x00203E, 0x00203E, 0x000001},
    {0x00203F, 0x002040, 0x000001}, {0x002041, 0x002043, 0x000001}, {0x002044, 0x002044, 0x000015},
    {0x002045, 0x002045, 0x00001D}, {0x002046, 0x002046, 0x000007}, {0x002047, 0x002049, 0x00001B},
    {0x00204A, 0x002051, 0x000001}, {0x002052, 0x002052, 0x000001}, {0x002053, 0x002053, 0x000001},
    {0x002054, 0x002054, 0x000001}, {0x002055, 0x002055, 0x000001}, {0x002056, 0x002056, 0x000002},
    {0x002057, 0x002057, 0x000001}, {0x002058, 0x00205B, 0x000002}, {0x00205C, 0x00205C, 0x000001},
    {0x00205D, 0x00205E, 0x000002}, {0x00205F, 0x00205F, 0x000002}, {0x002060, 0x002060, 0x000024},
    {0x002061, 0x002064, 0x000001}, {0x002066, 0x00206F, 0x000008}, {0x002070, 0x002070, 0x000001},
    {0x002071, 0x002071, 0x000001}, {0x002074, 0x002074, 0x000001}, {0x002075, 0x002079, 0x000001},
    {0x00207A, 0x00207C, 0x000001}, {0x00207D, 0x00207D, 0x00001D}, {0x00207E, 0x00207E, 0x000007},
    {0x00207F, 0x00207F, 0x000001}, {0x002080, 0x002080, 0x000001}, {0x002081, 0x002084, 0x000001},
    {0x002085, 0x002089, 0x000001}, {0x00208A, 0x00208C, 0x000001}, {0x00208D, 0x00208D, 0x00001D},
    {0x00208E, 0x00208E, 0x000007}, {0x002090, 0x00209C, 0x000001}, {0x0020A0, 0x0020A6, 0x00001F},
    {0x0020A7, 0x0020A7, 0x00001E}, {0x0020A8, 0x0020B5, 0x00001F}, {0x0020B6, 0x0020B6, 0x00001E},
    {0x0020B7, 0x0020BA, 0x00001F}, {0x0020BB, 0x0020BB, 0x00001E}, {0x0020BC, 0x0020BD, 0x00001F},
    {0x0020BE, 0x0020BE, 0x00001E}, {0x0020BF, 0x0020BF, 0x00001F}, {0x0020C0, 0x0020C0, 0x00001E},
    {0x0020C1, 0x0020CF, 0x00001F}, {0x0020D0, 0x0020DC, 0x000008}, {0x0020DD, 0x0020E0, 0x000008},
    {0x0020E1, 0x0020E1, 0x000008}, {0x0020E2, 0x0020E4, 0x000008}, {0x0020E5, 0x0020F0, 0x000008},
    {0x002100, 0x002101, 0x000001}, {0x002102, 0x002102, 0x000001}, {0x002103, 0x002103, 0x00001E},
    {0x002104, 0x002104, 0x000001}, {0x002105, 0x002105, 0x000001}, {0x002106, 0x002106, 0x000001},
    {0x002107, 0x002107, 0x000001}, {0x002108, 0x002108, 0x000001}, {0x002109, 0x002109, 0x00001E},
    {0x00210A, 0x002112, 0x000001}, {0x002113, 0x002113, 0x000001}, {0x002114, 0x002114, 0x000001},
    {0x002115, 0x002115, 0x000001}, {0x002116, 0x002116, 0x00001F}, {0x002117, 0x002117, 0x000001},
    {0x002118, 0x002118, 0x000001}, {0x002119, 0x00211D, 0x000001}, {0x00211E, 0x002120, 0x000001},
    {0x002121, 0x002122, 0x000001}, {0x002123, 0x002123, 0x000001}, {0x002124, 0x002124, 0x000001},
    {0x002125, 0x002125, 0x000001}, {0x002126, 0x002126, 0x000001}, {0x002127, 0x002127, 0x000001},
    {0x002128, 0x002128, 0x000001}, {0x002129, 0x002129, 0x000001}, {0x00212A, 0x00212A, 0x000001},
    {0x00212B, 0x00212B, 0x000001}, {0x00212C, 0x00212D, 0x000001}, {0x00212E, 0x00212E, 0x000001},
    {0x00212F, 0x002134, 0x000001}, {0x002135, 0x002138, 0x000001}, {0x002139, 0x002139, 0x000001},
    {0x00213A, 0x00213B, 0x000001}, {0x00213C, 0x00213F, 0x000001}, {0x002140, 0x002144, 0x000001},
    {0x002145, 0x002149, 0x000001}, {0x00214A, 0x00214A, 0x000001}, {0x00214B, 0x00214B, 0x000001},
    {0x00214C, 0x00214D, 0x000001}, {0x00214E, 0x00214E, 0x000001}, {0x00214F, 0x00214F, 0x000001},
    {0x002150, 0x002153, 0x000001}, {0x002154, 0x002155, 0x000001}, {0x002156, 0x00215A, 0x000001},
    {0x00215B, 0x00215B, 0x000001}, {0x00215C, 0x00215D, 0x000001}, {0x00215E, 0x00215E, 0x000001},
    {0x00215F, 0x00215F, 0x000001}, {0x002160, 0x00216B, 0x000001}, {0x00216C, 0x00216F, 0x000001},
    {0x002170, 0x002179, 0x000001}, {0x00217A, 0x002182, 0x000001}, {0x002183, 0x002184, 0x000001},
    {0x002185, 0x002188, 0x000001}, {0x002189, 0x002189, 0x000001}, {0x00218A, 0x00218B, 0x000001},
    {0x002190, 0x002194, 0x000001}, {0x002195, 0x002199, 0x000001}, {0x00219A, 0x00219B, 0x000001},
    {0x00219C, 0x00219F, 0x000001}, {0x0021A0, 0x0021A0, 0x000001}, {0x0021A1, 0x0021A2, 0x000001},
    {0x0021A3, 0x0021A3, 0x000001}, {0x0021A4, 0x0021A5, 0x000001}, {0x0021A6, 0x0021A6, 0x000001},
    {0x0021A7, 0x0021AD, 0x000001}, {0x0021AE, 0x0021AE, 0x000001}, {0x0021AF, 0x0021CD, 0x000001},
    {0x0021CE, 0x0021CF, 0x000001}, {0x0021D0, 0x0021D1, 0x000001}, {0x0021D2, 0x0021D2, 0x000001},
    {0x0021D3, 0x0021D3, 0x000001}, {0x0021D4, 0x0021D4, 0x000001}, {0x0021D5, 0x0021F3, 0x000001},
    {0x0021F4, 0x0021FF, 0x000001}, {0x002200, 0x002200, 0x000001}, {0x002201, 0x002201, 0x000001},
    {0x002202, 0x002203, 0x000001}, {0x002204, 0x002206, 0x000001}, {0x002207, 0x002208, 0x000001},
    {0x002209, 0x00220A, 0x000001}, {0x00220B, 0x00220B, 0x000001}, {0x00220C, 0x00220E, 0x000001},
    {0x00220F, 0x00220F, 0x000001}, {0x002210, 0x002210, 0x000001}, {0x002211, 0x002211, 0x000001},
    {0x002212, 0x002213, 0x00001F}, {0x002214, 0x002214, 0x000001}, {0x002215, 0x002215, 0x000001},
    {0x002216, 0x002219, 0x000001}, {0x00221A, 0x00221A, 0x000001}, {0x00221B, 0x00221C, 0x000001},
    {0x00221D, 0x002220, 0x000001}, {0x002221, 0x002222, 0x000001}, {0x002223, 0x002223, 0x000001},
    {0x002224, 0x002224, 0x000001}, {0x002225, 0x002225, 0x000001}, {0x002226, 0x002226, 0x000001},
    {0x002227, 0x00222C, 0x000001}, {0x00222D, 0x00222D, 0x000001}, {0x00222E, 0x00222E, 0x000001},
    {0x00222F, 0x002233, 0x000001}, {0x002234, 0x002237, 0x000001}, {0x002238, 0x00223B, 0x000001},
    {0x00223C, 0x00223D, 0x000001}, {0x00223E, 0x002247, 0x000001}, {0x002248, 0x002248, 0x000001},
    {0x002249, 0x00224B, 0x000001}, {0x00224C, 0x00224C, 0x000001}, {0x00224D, 0x002251, 0x000001},
    {0x002252, 0x002252, 0x000001}, {0x002253, 0x00225F, 0x000001}, {0x002260, 0x002261, 0x000001},
    {0x002262, 0x002263, 0x000001}, {0x002264, 0x002267, 0x000001}, {0x002268, 0x002269, 0x000001},
    {0x00226A, 0x00226B, 0x000001}, {0x00226C, 0x00226D, 0x000001}, {0x00226E, 0x00226F, 0x000001},
    {0x002270, 0x002281, 0x000001}, {0x002282, 0x002283, 0x000001}, {0x002284, 0x002285, 0x000001},
    {0x002286, 0x002287, 0x000001}, {0x002288, 0x002294, 0x000001}, {0x002295, 0x002295, 0x000001},
    {0x002296, 0x002298, 0x000001}, {0x002299, 0x002299, 0x000001}, {0x00229A, 0x0022A4, 0x000001},
    {0x0022A5, 0x0022A5, 0x000001}, {0x0022A6, 0x0022BE, 0x000001}, {0x0022BF, 0x0022BF, 0x000001},
    {0x0022C0, 0x0022EE, 0x000001}, {0x0022EF, 0x0022EF, 0x000014}, {0x0022F0, 0x0022FF, 0x000001},
    {0x002300, 0x002307, 0x000001}, {0x002308, 0x002308, 0x00001D}, {0x002309, 0x002309, 0x000007},
    {0x00230A, 0x00230A, 0x00001D}, {0x00230B, 0x00230B, 0x000007}, {0x00230C, 0x002311, 0x000001},
    {0x002312, 0x002312, 0x000001}, {0x002313, 0x002319, 0x000001}, {0x00231A, 0x00231B, 0x000012},
    {0x00231C, 0x00231F, 0x000001}, {0x002320, 0x002321, 0x000001}, {0x002322, 0x002328, 0x000001},
    {0x002329, 0x002329, 0x00001D}, {0x00232A, 0x00232A, 0x000007}, {0x00232B, 0x00237B, 0x000001},
    {0x00237C, 0x00237C, 0x000001}, {0x00237D, 0x00239A, 0x000001}, {0x00239B, 0x0023B3, 0x000001},
    {0x0023B4, 0x0023DB, 0x000001}, {0x0023DC, 0x0023E1, 0x000001}, {0x0023E2, 0x0023EF, 0x000001},
    {0x0023F0, 0x0023F3, 0x000012}, {0x0023F4, 0x0023FF, 0x000001}, {0x002400, 0x002426, 0x000001},
    {0x002440, 0x00244A, 0x000001}, {0x002460, 0x00249B, 0x000001}, {0x00249C, 0x0024E9, 0x000001},
    {0x0024EA, 0x0024FE, 0x000001}, {0x0024FF, 0x0024FF, 0x000001}, {0x002500, 0x00254B, 0x000001},
    {0x00254C, 0x00254F, 0x000001}, {0x002550, 0x002574, 0x000001}, {0x002575, 0x00257F, 0x000001},
    {0x002580, 0x00258F, 0x000001}, {0x002590, 0x002591, 0x000001}, {0x002592, 0x002595, 0x000001},
    {0x002596, 0x00259F, 0x000001}, {0x0025A0, 0x0025A1, 0x000001}, {0x0025A2, 0x0025A2, 0x000001},
    {0x0025A3, 0x0025A9, 0x000001}, {0x0025AA, 0x0025B1, 0x000001}, {0x0025B2, 0x0025B3, 0x000001},
    {0x0025B4, 0x0025B5, 0x000001}, {0x0025B6, 0x0025B6, 0x000001}, {0x0025B7, 0x0025B7, 0x000001},
    {0x0025B8, 0x0025BB, 0x000001}, {0x0025BC, 0x0025BD, 0x000001}, {0x0025BE, 0x0025BF, 0x000001},
    {0x0025C0, 0x0025C0, 0x000001}, {0x0025C1, 0x0025C1, 0x000001}, {0x0025C2, 0x0025C5, 0x000001},
    {0x0025C6, 0x0025C8, 0x000001}, {0x0025C9, 0x0025CA, 0x000001}, {0x0025CB, 0x0025CB, 0x000001},
    {0x0025CC, 0x0025CD, 0x000001}, {0x0025CE, 0x0025D1, 0x000001}, {0x0025D2, 0x0025E1, 0x000001},
    {0x0025E2, 0x0025E5, 0x000001}, {0x0025E6, 0x0025EE, 0x000001}, {0x0025EF, 0x0025EF, 0x000001},
    {0x0025F0, 0x0025F7, 0x000001}, {0x0025F8, 0x0025FF, 0x000001}, {0x002600, 0x002603, 0x000012},
    {0x002604, 0x002604, 0x000001}, {0x002605, 0x002606, 0x000001}, {0x002607, 0x002608, 0x000001},
    {0x002609, 0x002609, 0x000001}, {0x00260A, 0x00260D, 0x000001}, {0x00260E, 0x00260F, 0x000001},
    {0x002610, 0x002613, 0x000001}, {0x002614, 0x002615, 0x000012}, {0x002616, 0x002617, 0x000001},
    {0x002618, 0x002618, 0x000012}, {0x002619, 0x002619, 0x000001}, {0x00261A, 0x00261C, 0x000012},
    {0x00261D, 0x00261D, 0x00000B}, {0x00261E, 0x00261F, 0x000012}, {0x002620, 0x002638, 0x000001},
    {0x002639, 0x00263B, 0x000012}, {0x00263C, 0x00263F, 0x000001}, {0x002640, 0x002640, 0x000001},
    {0x002641, 0x002641, 0x000001}, {0x002642, 0x002642, 0x000001}, {0x002643, 0x00265F, 0x000001},
    {0x002660, 0x002661, 0x000001}, {0x002662, 0x002662, 0x000001}, {0x002663, 0x002665, 0x000001},
    {0x002666, 0x002666, 0x000001}, {0x002667, 0x002667, 0x000001}, {0x002668, 0x002668, 0x000012},
    {0x002669, 0x00266A, 0x000001}, {0x00266B, 0x00266B, 0x000001}, {0x00266C, 0x00266D, 0x000001},
    {0x00266E, 0x00266E, 0x000001}, {0x00266F, 0x00266F, 0x000001}, {0x002670, 0x00267E, 0x000001},
    {0x00267F, 0x00267F, 0x000012}, {0x002680, 0x00269D, 0x000001}, {0x00269E, 0x00269F, 0x000001},
    {0x0026A0, 0x0026BC, 0x000001}, {0x0026BD, 0x0026C8, 0x000012}, {0x0026C9, 0x0026CC, 0x000001},
    {0x0026CD, 0x0026CD, 0x000012}, {0x0026CE, 0x0026CE, 0x000001}, {0x0026CF, 0x0026D1, 0x000012},
    {0x0026D2, 0x0026D2, 0x000001}, {0x0026D3, 0x0026D4, 0x000012}, {0x0026D5, 0x0026D7, 0x000001},
    {0x0026D8, 0x0026D9, 0x000012}, {0x0026DA, 0x0026DB, 0x000001}, {0x0026DC, 0x0026DC, 0x000012},
    {0x0026DD, 0x0026DE, 0x000001}, {0x0026DF, 0x0026E1, 0x000012}, {0x0026E2, 0x0026E2, 0x000001},
    {0x0026E3, 0x0026E3, 0x000001}, {0x0026E4, 0x0026E7, 0x000001}, {0x0026E8, 0x0026E9, 0x000001},
    {0x0026EA, 0x0026EA, 0x000012}, {0x0026EB, 0x0026F0, 0x000001}, {0x0026F1, 0x0026F5, 0x000012},
    {0x0026F6, 0x0026F6, 0x000001}, {0x0026F7, 0x0026F8, 0x000012}, {0x0026F9, 0x0026F9, 0x00000B},
    {0x0026FA, 0x0026FA, 0x000012}, {0x0026FB, 0x0026FC, 0x000001}, {0x0026FD, 0x0026FF, 0x000012},
    {0x002700, 0x002704, 0x000012}, {0x002705, 0x002707, 0x000001}, {0x002708, 0x002709, 0x000012},
    {0x00270A, 0x00270D, 0x00000B}, {0x00270E, 0x002756, 0x000001}, {0x002757, 0x002757, 0x000001},
    {0x002758, 0x00275A, 0x000001}, {0x00275B, 0x002760, 0x000020}, {0x002761, 0x002761, 0x000001},
    {0x002762, 0x002763, 0x00000D}, {0x002764, 0x002764, 0x000012}, {0x002765, 0x002767, 0x000001},
    {0x002768, 0x002768, 0x00001D}, {0x002769, 0x002769, 0x000007}, {0x00276A, 0x00276A, 0x00001D},
    {0x00276B, 0x00276B, 0x000007}, {0x00276C, 0x00276C, 0x00001D}, {0x00276D, 0x00276D, 0x000007},
    {0x00276E, 0x00276E, 0x00001D}, {0x00276F, 0x00276F, 0x000007}, {0x002770, 0x002770, 0x00001D},
    {0x002771, 0x002771, 0x000007}, {0x002772, 0x002772, 0x00001D}, {0x002773, 0x002773, 0x000007},
    {0x002774, 0x002774, 0x00001D}, {0x002775, 0x002775, 0x000007}, {0x002776, 0x002793, 0x000001},
    {0x002794, 0x0027BF, 0x000001}, {0x0027C0, 0x0027C4, 0x000001}, {0x0027C5, 0x0027C5, 0x00001D},
    {0x0027C6, 0x0027C6, 0x000007}, {0x0027C7, 0x0027E5, 0x000001}, {0x0027E6, 0x0027E6, 0x00001D},
    {0x0027E7, 0x0027E7, 0x000007}, {0x0027E8, 0x0027E8, 0x00001D}, {0x0027E9, 0x0027E9, 0x000007},
    {0x0027EA, 0x0027EA, 0x00001D}, {0x0027EB, 0x0027EB, 0x000007}, {0x0027EC, 0x0027EC, 0x00001D},
    {0x0027ED, 0x0027ED, 0x000007}, {0x0027EE, 0x0027EE, 0x00001D}, {0x0027EF, 0x0027EF, 0x000007},
    {0x0027F0, 0x0027FF, 0x000001}, {0x002800, 0x0028FF, 0x000001}, {0x002900, 0x00297F, 0x000001},
    {0x002980, 0x002982, 0x000001}, {0x002983, 0x002983, 0x00001D}, {0x002984, 0x002984, 0x000007},
    {0x002985, 0x002985, 0x00001D}, {0x002986, 0x002986, 0x000007}, {0x002987, 0x002987, 0x00001D},
    {0x002988, 0x002988, 0x000007}, {0x002989, 0x002989, 0x00001D}, {0x00298A, 0x00298A, 0x000007},
    {0x00298B, 0x00298B, 0x00001D}, {0x00298C, 0x00298C, 0x000007}, {0x00298D, 0x00298D, 0x00001D},
    {0x00298E, 0x00298E, 0x000007}, {0x00298F, 0x00298F, 0x00001D}, {0x002990, 0x002990, 0x000007},
    {0x002991, 0x002991, 0x00001D}, {0x002992, 0x002992, 0x000007}, {0x002993, 0x002993, 0x00001D},
    {0x002994, 0x002994, 0x000007}, {0x002995, 0x002995, 0x00001D}, {0x002996, 0x002996, 0x000007},
    {0x002997, 0x002997, 0x00001D}, {0x002998, 0x002998, 0x000007}, {0x002999, 0x0029D7, 0x000001},
    {0x0029D8, 0x0029D8, 0x00001D}, {0x0029D9, 0x0029D9, 0x000007}, {0x0029DA, 0x0029DA, 0x00001D},
    {0x0029DB, 0x0029DB, 0x000007}, {0x0029DC, 0x0029FB, 0x000001}, {0x0029FC, 0x0029FC, 0x00001D},
    {0x0029FD, 0x0029FD, 0x000007}, {0x0029FE, 0x0029FF, 0x000001}, {0x002A00, 0x002AFF, 0x000001},
    {0x002B00, 0x002B2F, 0x000001}, {0x002B30, 0x002B44, 0x000001}, {0x002B45, 0x002B46, 0x000001},
    {0x002B47, 0x002B4C, 0x000001}, {0x002B4D, 0x002B54, 0x000001}, {0x002B55, 0x002B59, 0x000001},
    {0x002B5A, 0x002B73, 0x000001}, {0x002B76, 0x002B95, 0x000001}, {0x002B97, 0x002BFF, 0x000001},
    {0x002C00, 0x002C5F, 0x000001}, {0x002C60, 0x002C7B, 0x000001}, {0x002C7C, 0x002C7D, 0x000001},
    {0x002C7E, 0x002C7F, 0x000001}, {0x002C80, 0x002CE4, 0x000001}, {0x002CE5, 0x002CEA, 0x000001},
    {0x002CEB, 0x002CEE, 0x000001}, {0x002CEF, 0x002CF1, 0x000008}, {0x002CF2, 0x002CF3, 0x000001},
    {0x002CF9, 0x002CF9, 0x00000D}, {0x002CFA, 0x002CFC, 0x000002}, {0x002CFD, 0x002CFD, 0x000001},
    {0x002CFE, 0x002CFE, 0x00000D}, {0x002CFF, 0x002CFF, 0x000002}, {0x002D00, 0x002D25, 0x000001},
    {0x002D27, 0x002D27, 0x000001}, {0x002D2D, 0x002D2D, 0x000001}, {0x002D30, 0x002D67, 0x000001},
    {0x002D6F, 0x002D6F, 0x000001}, {0x002D70, 0x002D70, 0x000002}, {0x002D7F, 0x002D7F, 0x000008},
    {0x002D80, 0x002D96, 0x000001}, {0x002DA0, 0x002DA6, 0x000001}, {0x002DA8, 0x002DAE, 0x000001},
    {0x002DB0, 0x002DB6, 0x000001}, {0x002DB8, 0x002DBE, 0x000001}, {0x002DC0, 0x002DC6, 0x000001},
    {0x002DC8, 0x002DCE, 0x000001}, {0x002DD0, 0x002DD6, 0x000001}, {0x002DD8, 0x002DDE, 0x000001},
    {0x002DE0, 0x002DFF, 0x000008}, {0x002E00, 0x002E01, 0x000020}, {0x002E02, 0x002E02, 0x000020},
    {0x002E03, 0x002E03, 0x000020}, {0x002E04, 0x002E04, 0x000020}, {0x002E05, 0x002E05, 0x000020},
    {0x002E06, 0x002E08, 0x000020}, {0x002E09, 0x002E09, 0x000020}, {0x002E0A, 0x002E0A, 0x000020},
    {0x002E0B, 0x002E0B, 0x000020}, {0x002E0C, 0x002E0C, 0x000020}, {0x002E0D, 0x002E0D, 0x000020},
    {0x002E0E, 0x002E15, 0x000002}, {0x002E16, 0x002E16, 0x000001}, {0x002E17, 0x002E17, 0x000002},
    {0x002E18, 0x002E18, 0x00001D}, {0x002E19, 0x002E19, 0x000002}, {0x002E1A, 0x002E1A, 0x000001},
    {0x002E1B, 0x002E1B, 0x000001}, {0x002E1C, 0x002E1C, 0x000020}, {0x002E1D, 0x002E1D, 0x000020},
    {0x002E1E, 0x002E1F, 0x000001}, {0x002E20, 0x002E20, 0x000020}, {0x002E21, 0x002E21, 0x000020},
    {0x002E22, 0x002E22, 0x00001D}, {0x002E23, 0x002E23, 0x000007}, {0x002E24, 0x002E24, 0x00001D},
    {0x002E25, 0x002E25, 0x000007}, {0x002E26, 0x002E26, 0x00001D}, {0x002E27, 0x002E27, 0x000007},
    {0x002E28, 0x002E28, 0x00001D}, {0x002E29, 0x002E29, 0x000007}, {0x002E2A, 0x002E2D, 0x000002},
    {0x002E2E, 0x002E2E, 0x00000D}, {0x002E2F, 0x002E2F, 0x000001}, {0x002E30, 0x002E31, 0x000002},
    {0x002E32, 0x002E32, 0x000001}, {0x002E33, 0x002E34, 0x000002}, {0x002E35, 0x002E39, 0x000001},
    {0x002E3A, 0x002E3B, 0x000004}, {0x002E3C, 0x002E3E, 0x000002}, {0x002E3F, 0x002E3F, 0x000001},
    {0x002E40, 0x002E40, 0x000002}, {0x002E41, 0x002E41, 0x000002}, {0x002E42, 0x002E42, 0x00001D},
    {0x002E43, 0x002E4A, 0x000002}, {0x002E4B, 0x002E4B, 0x000001}, {0x002E4C, 0x002E4C, 0x000002},
    {0x002E4D, 0x002E4D, 0x000001}, {0x002E4E, 0x002E4F, 0x000002}, {0x002E50, 0x002E51, 0x000001},
    {0x002E52, 0x002E52, 0x000001}, {0x002E53, 0x002E54, 0x00000D}, {0x002E55, 0x002E55, 0x00001D},
    {0x002E56, 0x002E56, 0x000007}, {0x002E57, 0x002E57, 0x00001D}, {0x002E58, 0x002E58, 0x000007},
    {0x002E59, 0x002E59, 0x00001D}, {0x002E5A, 0x002E5A, 0x000007}, {0x002E5B, 0x002E5B, 0x00001D},
    {0x002E5C, 0x002E5C, 0x000007}, {0x002E5D, 0x002E5D, 0x000002}, {0x002E80, 0x002E99, 0x000012},
    {0x002E9B, 0x002EF3, 0x000012}, {0x002F00, 0x002FD5, 0x000012}, {0x002FF0, 0x002FFB, 0x000012},
    {0x003000, 0x003000, 0x000002}, {0x003001, 0x003002, 0x000007}, {0x003003, 0x003003, 0x000012},
    {0x003004, 0x003004, 0x000012}, {0x003005, 0x003005, 0x00001B}, {0x003006, 0x003006, 0x000012},
    {0x003007, 0x003007, 0x000012}, {0x003008, 0x003008, 0x00001D}, {0x003009, 0x003009, 0x000007},
    {0x00300A, 0x00300A, 0x00001D}, {0x00300B, 0x00300B, 0x000007}, {0x00300C, 0x00300C, 0x00001D},
    {0x00300D, 0x00300D, 0x000007}, {0x00300E, 0x00300E, 0x00001D}, {0x00300F, 0x00300F, 0x000007},
    {0x003010, 0x003010, 0x00001D}, {0x003011, 0x003011, 0x000007}, {0x003012, 0x003013, 0x000012},
    {0x003014, 0x003014, 0x00001D}, {0x003015, 0x003015, 0x000007}, {0x003016, 0x003016, 0x00001D},
    {0x003017, 0x003017, 0x000007}, {0x003018, 0x003018, 0x00001D}, {0x003019, 0x003019, 0x000007},
    {0x00301A, 0x00301A, 0x00001D}, {0x00301B, 0x00301B, 0x000007}, {0x00301C, 0x00301C, 0x00001B},
    {0x00301D, 0x00301D, 0x00001D}, {0x00301E, 0x00301F, 0x000007}, {0x003020, 0x003020, 0x000012},
    {0x003021, 0x003029, 0x000012}, {0x00302A, 0x00302D, 0x000008}, {0x00302E, 0x00302F, 0x000008},
    {0x003030, 0x003030, 0x000012}, {0x003031, 0x003034, 0x000012}, {0x003035, 0x003035, 0x000008},
    {0x003036, 0x003037, 0x000012}, {0x003038, 0x00303A, 0x000012}, {0x00303B, 0x00303B, 0x00001B},
    {0x00303C, 0x00303C, 0x00001B}, {0x00303D, 0x00303D, 0x000012}, {0x00303E, 0x00303F, 0x000012},
    {0x003041, 0x003041, 0x00001B}, {0x003042, 0x003042, 0x000012}, {0x003043, 0x003043, 0x00001B},
    {0x003044, 0x003044, 0x000012}, {0x003045, 0x003045, 0x00001B}, {0x003046, 0x003046, 0x000012},
    {0x003047, 0x003047, 0x00001B}, {0x003048, 0x003048, 0x000012}, {0x003049, 0x003049, 0x00001B},
    {0x00304A, 0x003062, 0x000012}, {0x003063, 0x003063, 0x00001B}, {0x003064, 0x003082, 0x000012},
    {0x003083, 0x003083, 0x00001B}, {0x003084, 0x003084, 0x000012}, {0x003085, 0x003085, 0x00001B},
    {0x003086, 0x003086, 0x000012}, {0x003087, 0x003087, 0x00001B}, {0x003088, 0x00308D, 0x000012},
    {0x00308E, 0x00308E, 0x00001B}, {0x00308F, 0x003094, 0x000012}, {0x003095, 0x003096, 0x00001B},
    {0x003099, 0x00309A, 0x000008}, {0x00309B, 0x00309C, 0x00001B}, {0x00309D, 0x00309E, 0x00001B},
    {0x00309F, 0x00309F, 0x000012}, {0x0030A0, 0x0030A0, 0x00001B}, {0x0030A1, 0x0030A1, 0x00001B},
    {0x0030A2, 0x0030A2, 0x000012}, {0x0030A3, 0x0030A3, 0x00001B}, {0x0030A4, 0x0030A4, 0x000012},
    {0x0030A5, 0x0030A5, 0x00001B}, {0x0030A6, 0x0030A6, 0x000012}, {0x0030A7, 0x0030A7, 0x00001B},
    {0x0030A8, 0x0030A8, 0x000012}, {0x0030A9, 0x0030A9, 0x00001B}, {0x0030AA, 0x0030C2, 0x000012},
    {0x0030C3, 0x0030C3, 0x00001B}, {0x0030C4, 0x0030E2, 0x000012}, {0x0030E3, 0x0030E3, 0x00001B},
    {0x0030E4, 0x0030E4, 0x000012}, {0x0030E5, 0x0030E5, 0x00001B}, {0x0030E6, 0x0030E6, 0x000012},
    {0x0030E7, 0x0030E7, 0x00001B}, {0x0030E8, 0x0030ED, 0x000012}, {0x0030EE, 0x0030EE, 0x00001B},
    {0x0030EF, 0x0030F4, 0x000012}, {0x0030F5, 0x0030F6, 0x00001B}, {0x0030F7, 0x0030FA, 0x000012},
    {0x0030FB, 0x0030FB, 0x00001B}, {0x0030FC, 0x0030FC, 0x00001B}, {0x0030FD, 0x0030FE, 0x00001B},
    {0x0030FF, 0x0030FF, 0x000012}, {0x003105, 0x00312F, 0x000012}, {0x003131, 0x00318E, 0x000012},
    {0x003190, 0x003191, 0x000012}, {0x003192, 0x003195, 0x000012}, {0x003196, 0x00319F, 0x000012},
    {0x0031A0, 0x0031BF, 0x000012}, {0x0031C0, 0x0031E3, 0x000012}, {0x0031F0, 0x0031FF, 0x00001B},
    {0x003200, 0x00321E, 0x000012}, {0x003220, 0x003229, 0x000012}, {0x00322A, 0x003247, 0x000012},
    {0x003248, 0x00324F, 0x000001}, {0x003250, 0x003250, 0x000012}, {0x003251, 0x00325F, 0x000012},
    {0x003260, 0x00327F, 0x000012}, {0x003280, 0x003289, 0x000012}, {0x00328A, 0x0032B0, 0x000012},
    {0x0032B1, 0x0032BF, 0x000012}, {0x0032C0, 0x0032FF, 0x000012}, {0x003300, 0x0033FF, 0x000012},
    {0x003400, 0x004DBF, 0x000012}, {0x004DC0, 0x004DFF, 0x000001}, {0x004E00, 0x009FFF, 0x000012},
    {0x00A000, 0x00A014, 0x000012}, {0x00A015, 0x00A015, 0x00001B}, {0x00A016, 0x00A48C, 0x000012},
    {0x00A490, 0x00A4C6, 0x000012}, {0x00A4D0, 0x00A4F7, 0x000001}, {0x00A4F8, 0x00A4FD, 0x000001},
    {0x00A4FE, 0x00A4FF, 0x000002}, {0x00A500, 0x00A60B, 0x000001}, {0x00A60C, 0x00A60C, 0x000001},
    {0x00A60D, 0x00A60D, 0x000002}, {0x00A60E, 0x00A60E, 0x00000D}, {0x00A60F, 0x00A60F, 0x000002},
    {0x00A610, 0x00A61F, 0x000001}, {0x00A620, 0x00A629, 0x00001C}, {0x00A62A, 0x00A62B, 0x000001},
    {0x00A640, 0x00A66D, 0x000001}, {0x00A66E, 0x00A66E, 0x000001}, {0x00A66F, 0x00A66F, 0x000008},
    {0x00A670, 0x00A672, 0x000008}, {0x00A673, 0x00A673, 0x000001}, {0x00A674, 0x00A67D, 0x000008},
    {0x00A67E, 0x00A67E, 0x000001}, {0x00A67F, 0x00A67F, 0x000001}, {0x00A680, 0x00A69B, 0x000001},
    {0x00A69C, 0x00A69D, 0x000001}, {0x00A69E, 0x00A69F, 0x000008}, {0x00A6A0, 0x00A6E5, 0x000001},
    {0x00A6E6, 0x00A6EF, 0x000001}, {0x00A6F0, 0x00A6F1, 0x000008}, {0x00A6F2, 0x00A6F2, 0x000001},
    {0x00A6F3, 0x00A6F7, 0x000002}, {0x00A700, 0x00A716, 0x000001}, {0x00A717, 0x00A71F, 0x000001},
    {0x00A720, 0x00A721, 0x000001}, {0x00A722, 0x00A76F, 0x000001}, {0x00A770, 0x00A770, 0x000001},
    {0x00A771, 0x00A787, 0x000001}, {0x00A788, 0x00A788, 0x000001}, {0x00A789, 0x00A78A, 0x000001},
    {0x00A78B, 0x00A78E, 0x000001}, {0x00A78F, 0x00A78F, 0x000001}, {0x00A790, 0x00A7CA, 0x000001},
    {0x00A7D0, 0x00A7D1, 0x000001}, {0x00A7D3, 0x00A7D3, 0x000001}, {0x00A7D5, 0x00A7D9, 0x000001},
    {0x00A7F2, 0x00A7F4, 0x000001}, {0x00A7F5, 0x00A7F6, 0x000001}, {0x00A7F7, 0x00A7F7, 0x000001},
    {0x00A7F8, 0x00A7F9, 0x000001}, {0x00A7FA, 0x00A7FA, 0x000001}, {0x00A7FB, 0x00A7FF, 0x000001},
    {0x00A800, 0x00A801, 0x000001}, {0x00A802, 0x00A802, 0x000008}, {0x00A803, 0x00A805, 0x000001},
    {0x00A806, 0x00A806, 0x000008}, {0x00A807, 0x00A80A, 0x000001}, {0x00A80B, 0x00A80B, 0x000008},
    {0x00A80C, 0x00A822, 0x000001}, {0x00A823, 0x00A824, 0x000008}, {0x00A825, 0x00A826, 0x000008},
    {0x00A827, 0x00A827, 0x000008}, {0x00A828, 0x00A82B, 0x000001}, {0x00A82C, 0x00A82C, 0x000008},
    {0x00A830, 0x00A835, 0x000001}, {0x00A836, 0x00A837, 0x000001}, {0x00A838, 0x00A838, 0x00001E},
    {0x00A839, 0x00A839, 0x000001}, {0x00A840, 0x00A873, 0x000001}, {0x00A874, 0x00A875, 0x000003},
    {0x00A876, 0x00A877, 0x00000D}, {0x00A880, 0x00A881, 0x000008}, {0x00A882, 0x00A8B3, 0x000001},
    {0x00A8B4, 0x00A8C3, 0x000008}, {0x00A8C4, 0x00A8C5, 0x000008}, {0x00A8CE, 0x00A8CF, 0x000002},
    {0x00A8D0, 0x00A8D9, 0x00001C}, {0x00A8E0, 0x00A8F1, 0x000008}, {0x00A8F2, 0x00A8F7, 0x000001},
    {0x00A8F8, 0x00A8FA, 0x000001}, {0x00A8FB, 0x00A8FB, 0x000001}, {0x00A8FC, 0x00A8FC, 0x000003},
    {0x00A8FD, 0x00A8FE, 0x000001}, {0x00A8FF, 0x00A8FF, 0x000008}, {0x00A900, 0x00A909, 0x00001C},
    {0x00A90A, 0x00A925, 0x000001}, {0x00A926, 0x00A92D, 0x000008}, {0x00A92E, 0x00A92F, 0x000002},
    {0x00A930, 0x00A946, 0x000001}, {0x00A947, 0x00A951, 0x000008}, {0x00A952, 0x00A953, 0x000008},
    {0x00A95F, 0x00A95F, 0x000001}, {0x00A960, 0x00A97C, 0x000016}, {0x00A980, 0x00A982, 0x000008},
    {0x00A983, 0x00A983, 0x000008}, {0x00A984, 0x00A9B2, 0x000001}, {0x00A9B3, 0x00A9B3, 0x000008},
    {0x00A9B4, 0x00A9B5, 0x000008}, {0x00A9B6, 0x00A9B9, 0x000008}, {0x00A9BA, 0x00A9BB, 0x000008},
    {0x00A9BC, 0x00A9BD, 0x000008}, {0x00A9BE, 0x00A9C0, 0x000008}, {0x00A9C1, 0x00A9C6, 0x000001},
    {0x00A9C7, 0x00A9C9, 0x000002}, {0x00A9CA, 0x00A9CD, 0x000001}, {0x00A9CF, 0x00A9CF, 0x000001},
    {0x00A9D0, 0x00A9D9, 0x00001C}, {0x00A9DE, 0x00A9DF, 0x000001}, {0x00A9E0, 0x00A9E4, 0x000001},
    {0x00A9E5, 0x00A9E5, 0x000008}, {0x00A9E6, 0x00A9E6, 0x000001}, {0x00A9E7, 0x00A9EF, 0x000001},
    {0x00A9F0, 0x00A9F9, 0x00001C}, {0x00A9FA, 0x00A9FE, 0x000001}, {0x00AA00, 0x00AA28, 0x000001},
    {0x00AA29, 0x00AA2E, 0x000008}, {0x00AA2F, 0x00AA30, 0x000008}, {0x00AA31, 0x00AA32, 0x000008},
    {0x00AA33, 0x00AA34, 0x000008}, {0x00AA35, 0x00AA36, 0x000008}, {0x00AA40, 0x00AA42, 0x000001},
    {0x00AA43, 0x00AA43, 0x000008}, {0x00AA44, 0x00AA4B, 0x000001}, {0x00AA4C, 0x00AA4C, 0x000008},
    {0x00AA4D, 0x00AA4D, 0x000008}, {0x00AA50, 0x00AA59, 0x00001C}, {0x00AA5C, 0x00AA5C, 0x000001},
    {0x00AA5D, 0x00AA5F, 0x000002}, {0x00AA60, 0x00AA6F, 0x000001}, {0x00AA70, 0x00AA70, 0x000001},
    {0x00AA71, 0x00AA76, 0x000001}, {0x00AA77, 0x00AA79, 0x000001}, {0x00AA7A, 0x00AA7A, 0x000001},
    {0x00AA7B, 0x00AA7B, 0x000008}, {0x00AA7C, 0x00AA7C, 0x000008}, {0x00AA7D, 0x00AA7D, 0x000008},
    {0x00AA7E, 0x00AA7F, 0x000001}, {0x00AA80, 0x00AAAF, 0x000001}, {0x00AAB0, 0x00AAB0, 0x000008},
    {0x00AAB1, 0x00AAB1, 0x000001}, {0x00AAB2, 0x00AAB4, 0x000008}, {0x00AAB5, 0x00AAB6, 0x000001},
    {0x00AAB7, 0x00AAB8, 0x000008}, {0x00AAB9, 0x00AABD, 0x000001}, {0x00AABE, 0x00AABF, 0x000008},
    {0x00AAC0, 0x00AAC0, 0x000001}, {0x00AAC1, 0x00AAC1, 0x000008}, {0x00AAC2, 0x00AAC2, 0x000001},
    {0x00AADB, 0x00AADC, 0x000001}, {0x00AADD, 0x00AADD, 0x000001}, {0x00AADE, 0x00AADF, 0x000001},
    {0x00AAE0, 0x00AAEA, 0x000001}, {0x00AAEB, 0x00AAEB, 0x000008}, {0x00AAEC, 0x00AAED, 0x000008},
    {0x00AAEE, 0x00AAEF, 0x000008}, {0x00AAF0, 0x00AAF1, 0x000002}, {0x00AAF2, 0x00AAF2, 0x000001},
    {0x00AAF3, 0x00AAF4, 0x000001}, {0x00AAF5, 0x00AAF5, 0x000008}, {0x00AAF6, 0x00AAF6, 0x000008},
    {0x00AB01, 0x00AB06, 0x000001}, {0x00AB09, 0x00AB0E, 0x000001}, {0x00AB11, 0x00AB16, 0x000001},
    {0x00AB20, 0x00AB26, 0x000001}, {0x00AB28, 0x00AB2E, 0x000001}, {0x00AB30, 0x00AB5A, 0x000001},
    {0x00AB5B, 0x00AB5B, 0x000001}, {0x00AB5C, 0x00AB5F, 0x000001}, {0x00AB60, 0x00AB68, 0x000001},
    {0x00AB69, 0x00AB69, 0x000001}, {0x00AB6A, 0x00AB6B, 0x000001}, {0x00AB70, 0x00ABBF, 0x000001},
    {0x00ABC0, 0x00ABE2, 0x000001}, {0x00ABE3, 0x00ABE4, 0x000008}, {0x00ABE5, 0x00ABE5, 0x000008},
    {0x00ABE6, 0x00ABE7, 0x000008}, {0x00ABE8, 0x00ABE8, 0x000008}, {0x00ABE9, 0x00ABEA, 0x000008},
    {0x00ABEB, 0x00ABEB, 0x000002}, {0x00ABEC, 0x00ABEC, 0x000008}, {0x00ABED, 0x00ABED, 0x000008},
    {0x00ABF0, 0x00ABF9, 0x00001C}, {0x00AC00, 0x00AC00, 0x00000F}, {0x00AC01, 0x00AC1B, 0x000010},
    {0x00AC1C, 0x00AC1C, 0x00000F}, {0x00AC1D, 0x00AC37, 0x000010}, {0x00AC38, 0x00AC38, 0x00000F},
    {0x00AC39, 0x00AC53, 0x000010}, {0x00AC54, 0x00AC54, 0x00000F}, {0x00AC55, 0x00AC6F, 0x000010},
    {0x00AC70, 0x00AC70, 0x00000F}, {0x00AC71, 0x00AC8B, 0x000010}, {0x00AC8C, 0x00AC8C, 0x00000F},
    {0x00AC8D, 0x00ACA7, 0x000010}, {0x00ACA8, 0x00ACA8, 0x00000F}, {0x00ACA9, 0x00ACC3, 0x000010},
    {0x00ACC4, 0x00ACC4, 0x00000F}, {0x00ACC5, 0x00ACDF, 0x000010}, {0x00ACE0, 0x00ACE0, 0x00000F},
    {0x00ACE1, 0x00ACFB, 0x000010}, {0x00ACFC, 0x00ACFC, 0x00000F}, {0x00ACFD, 0x00AD17, 0x000010},
    {0x00AD18, 0x00AD18, 0x00000F}, {0x00AD19, 0x00AD33, 0x000010}, {0x00AD34, 0x00AD34, 0x00000F},
    {0x00AD35, 0x00AD4F, 0x000010}, {0x00AD50, 0x00AD50, 0x00000F}, {0x00AD51, 0x00AD6B, 0x000010},
    {0x00AD6C, 0x00AD6C, 0x00000F}, {0x00AD6D, 0x00AD87, 0x000010}, {0x00AD88, 0x00AD88, 0x00000F},
    {0x00AD89, 0x00ADA3, 0x000010}, {0x00ADA4, 0x00ADA4, 0x00000F}, {0x00ADA5, 0x00ADBF, 0x000010},
    {0x00ADC0, 0x00ADC0, 0x00000F}, {0x00ADC1, 0x00ADDB, 0x000010}, {0x00ADDC, 0x00ADDC, 0x00000F},
    {0x00ADDD, 0x00ADF7, 0x000010}, {0x00ADF8, 0x00ADF8, 0x00000F}, {0x00ADF9, 0x00AE13, 0x000010},
    {0x00AE14, 0x00AE14, 0x00000F}, {0x00AE15, 0x00AE2F, 0x000010}, {0x00AE30, 0x00AE30, 0x00000F},
    {0x00AE31, 0x00AE4B, 0x000010}, {0x00AE4C, 0x00AE4C, 0x00000F}, {0x00AE4D, 0x00AE67, 0x000010},
    {0x00AE68, 0x00AE68, 0x00000F}, {0x00AE69, 0x00AE83, 0x000010}, {0x00AE84, 0x00AE84, 0x00000F},
    {0x00AE85, 0x00AE9F, 0x000010}, {0x00AEA0, 0x00AEA0, 0x00000F}, {0x00AEA1, 0x00AEBB, 0x000010},
    {0x00AEBC, 0x00AEBC, 0x00000F}, {0x00AEBD, 0x00AED7, 0x000010}, {0x00AED8, 0x00AED8, 0x00000F},
    {0x00AED9, 0x00AEF3, 0x000010}, {0x00AEF4, 0x00AEF4, 0x00000F}, {0x00AEF5, 0x00AF0F, 0x000010},
    {0x00AF10, 0x00AF10, 0x00000F}, {0x00AF11, 0x00AF2B, 0x000010}, {0x00AF2C, 0x00AF2C, 0x00000F},
    {0x00AF2D, 0x00AF47, 0x000010}, {0x00AF48, 0x00AF48, 0x00000F}, {0x00AF49, 0x00AF63, 0x000010},
    {0x00AF64, 0x00AF64, 0x00000F}, {0x00AF65, 0x00AF7F, 0x000010}, {0x00AF80, 0x00AF80, 0x00000F},
    {0x00AF81, 0x00AF9B, 0x000010}, {0x00AF9C, 0x00AF9C, 0x00000F}, {0x00AF9D, 0x00AFB7, 0x000010},
    {0x00AFB8, 0x00AFB8, 0x00000F}, {0x00AFB9, 0x00AFD3, 0x000010}, {0x00AFD4, 0x00AFD4, 0x00000F},
    {0x00AFD5, 0x00AFEF, 0x000010}, {0x00AFF0, 0x00AFF0, 0x00000F}, {0x00AFF1, 0x00B00B, 0x000010},
    {0x00B00C, 0x00B00C, 0x00000F}, {0x00B00D, 0x00B027, 0x000010}, {0x00B028, 0x00B028, 0x00000F},
    {0x00B029, 0x00B043, 0x000010}, {0x00B044, 0x00B044, 0x00000F}, {0x00B045, 0x00B05F, 0x000010},
    {0x00B060, 0x00B060, 0x00000F}, {0x00B061, 0x00B07B, 0x000010}, {0x00B07C, 0x00B07C, 0x00000F},
    {0x00B07D, 0x00B097, 0x000010}, {0x00B098, 0x00B098, 0x00000F}, {0x00B099, 0x00B0B3, 0x000010},
    {0x00B0B4, 0x00B0B4, 0x00000F}, {0x00B0B5, 0x00B0CF, 0x000010}, {0x00B0D0, 0x00B0D0, 0x00000F},
    {0x00B0D1, 0x00B0EB, 0x000010}, {0x00B0EC, 0x00B0EC, 0x00000F}, {0x00B0ED, 0x00B107, 0x000010},
    {0x00B108, 0x00B108, 0x00000F}, {0x00B109, 0x00B123, 0x000010}, {0x00B124, 0x00B124, 0x00000F},
    {0x00B125, 0x00B13F, 0x000010}, {0x00B140, 0x00B140, 0x00000F}, {0x00B141, 0x00B15B, 0x000010},
    {0x00B15C, 0x00B15C, 0x00000F}, {0x00B15D, 0x00B177, 0x000010}, {0x00B178, 0x00B178, 0x00000F},
    {0x00B179, 0x00B193, 0x000010}, {0x00B194, 0x00B194, 0x00000F}, {0x00B195, 0x00B1AF, 0x000010},
    {0x00B1B0, 0x00B1B0, 0x00000F}, {0x00B1B1, 0x00B1CB, 0x000010}, {0x00B1CC, 0x00B1CC, 0x00000F},
    {0x00B1CD, 0x00B1E7, 0x000010}, {0x00B1E8, 0x00B1E8, 0x00000F}, {0x00B1E9, 0x00B203, 0x000010},
    {0x00B204, 0x00B204, 0x00000F}, {0x00B205, 0x00B21F, 0x000010}, {0x00B220, 0x00B220, 0x00000F},
    {0x00B221, 0x00B23B, 0x000010}, {0x00B23C, 0x00B23C, 0x00000F}, {0x00B23D, 0x00B257, 0x000010},
    {0x00B258, 0x00B258, 0x00000F}, {0x00B259, 0x00B273, 0x000010}, {0x00B274, 0x00B274, 0x00000F},
    {0x00B275, 0x00B28F, 0x000010}, {0x00B290, 0x00B290, 0x00000F}, {0x00B291, 0x00B2AB, 0x000010},
    {0x00B2AC, 0x00B2AC, 0x00000F}, {0x00B2AD, 0x00B2C7, 0x000010}, {0x00B2C8, 0x00B2C8, 0x00000F},
    {0x00B2C9, 0x00B2E3, 0x000010}, {0x00B2E4, 0x00B2E4, 0x00000F}, {0x00B2E5, 0x00B2FF, 0x000010},
    {0x00B300, 0x00B300, 0x00000F}, {0x00B301, 0x00B31B, 0x000010}, {0x00B31C, 0x00B31C, 0x00000F},
    {0x00B31D, 0x00B337, 0x000010}, {0x00B338, 0x00B338, 0x00000F}, {0x00B339, 0x00B353, 0x000010},
    {0x00B354, 0x00B354, 0x00000F}, {0x00B355, 0x00B36F, 0x000010}, {0x00B370, 0x00B370, 0x00000F},
    {0x00B371, 0x00B38B, 0x000010}, {0x00B38C, 0x00B38C, 0x00000F}, {0x00B38D, 0x00B3A7, 0x000010},
    {0x00B3A8, 0x00B3A8, 0x00000F}, {0x00B3A9, 0x00B3C3, 0x000010}, {0x00B3C4, 0x00B3C4, 0x00000F},
    {0x00B3C5, 0x00B3DF, 0x000010}, {0x00B3E0, 0x00B3E0, 0x00000F}, {0x00B3E1, 0x00B3FB, 0x000010},
    {0x00B3FC, 0x00B3FC, 0x00000F}, {0x00B3FD, 0x00B417, 0x000010}, {0x00B418, 0x00B418, 0x00000F},
    {0x00B419, 0x00B433, 0x000010}, {0x00B434, 0x00B434, 0x00000F}, {0x00B435, 0x00B44F, 0x000010},
    {0x00B450, 0x00B450, 0x00000F}, {0x00B451, 0x00B46B, 0x000010}, {0x00B46C, 0x00B46C, 0x00000F},
    {0x00B46D, 0x00B487, 0x000010}, {0x00B488, 0x00B488, 0x00000F}, {0x00B489, 0x00B4A3, 0x000010},
    {0x00B4A4, 0x00B4A4, 0x00000F}, {0x00B4A5, 0x00B4BF, 0x000010}, {0x00B4C0, 0x00B4C0, 0x00000F},
    {0x00B4C1, 0x00B4DB, 0x000010}, {0x00B4DC, 0x00B4DC, 0x00000F}, {0x00B4DD, 0x00B4F7, 0x000010},
    {0x00B4F8, 0x00B4F8, 0x00000F}, {0x00B4F9, 0x00B513, 0x000010}, {0x00B514, 0x00B514, 0x00000F},
    {0x00B515, 0x00B52F, 0x000010}, {0x00B530, 0x00B530, 0x00000F}, {0x00B531, 0x00B54B, 0x000010},
    {0x00B54C, 0x00B54C, 0x00000F}, {0x00B54D, 0x00B567, 0x000010}, {0x00B568, 0x00B568, 0x00000F},
    {0x00B569, 0x00B583, 0x000010}, {0x00B584, 0x00B584, 0x00000F}, {0x00B585, 0x00B59F, 0x000010},
    {0x00B5A0, 0x00B5A0, 0x00000F}, {0x00B5A1, 0x00B5BB, 0x000010}, {0x00B5BC, 0x00B5BC, 0x00000F},
    {0x00B5BD, 0x00B5D7, 0x000010}, {0x00B5D8, 0x00B5D8, 0x00000F}, {0x00B5D9, 0x00B5F3, 0x000010},
    {0x00B5F4, 0x00B5F4, 0x00000F}, {0x00B5F5, 0x00B60F, 0x000010}, {0x00B610, 0x00B610, 0x00000F},
    {0x00B611, 0x00B62B, 0x000010}, {0x00B62C, 0x00B62C, 0x00000F}, {0x00B62D, 0x00B647, 0x000010},
    {0x00B648, 0x00B648, 0x00000F}, {0x00B649, 0x00B663, 0x000010}, {0x00B664, 0x00B664, 0x00000F},
    {0x00B665, 0x00B67F, 0x000010}, {0x00B680, 0x00B680, 0x00000F}, {0x00B681, 0x00B69B, 0x000010},
    {0x00B69C, 0x00B69C, 0x00000F}, {0x00B69D, 0x00B6B7, 0x000010}, {0x00B6B8, 0x00B6B8, 0x00000F},
    {0x00B6B9, 0x00B6D3, 0x000010}, {0x00B6D4, 0x00B6D4, 0x00000F}, {0x00B6D5, 0x00B6EF, 0x000010},
    {0x00B6F0, 0x00B6F0, 0x00000F}, {0x00B6F1, 0x00B70B, 0x000010}, {0x00B70C, 0x00B70C, 0x00000F},
    {0x00B70D, 0x00B727, 0x000010}, {0x00B728, 0x00B728, 0x00000F}, {0x00B729, 0x00B743, 0x000010},
    {0x00B744, 0x00B744, 0x00000F}, {0x00B745, 0x00B75F, 0x000010}, {0x00B760, 0x00B760, 0x00000F},
    {0x00B761, 0x00B77B, 0x000010}, {0x00B77C, 0x00B77C, 0x00000F}, {0x00B77D, 0x00B797, 0x000010},
    {0x00B798, 0x00B798, 0x00000F}, {0x00B799, 0x00B7B3, 0x000010}, {0x00B7B4, 0x00B7B4, 0x00000F},
    {0x00B7B5, 0x00B7CF, 0x000010}, {0x00B7D0, 0x00B7D0, 0x00000F}, {0x00B7D1, 0x00B7EB, 0x000010},
    {0x00B7EC, 0x00B7EC, 0x00000F}, {0x00B7ED, 0x00B807, 0x000010}, {0x00B808, 0x00B808, 0x00000F},
    {0x00B809, 0x00B823, 0x000010}, {0x00B824, 0x00B824, 0x00000F}, {0x00B825, 0x00B83F, 0x000010},
    {0x00B840, 0x00B840, 0x00000F}, {0x00B841, 0x00B85B, 0x000010}, {0x00B85C, 0x00B85C, 0x00000F},
    {0x00B85D, 0x00B877, 0x000010}, {0x00B878, 0x00B878, 0x00000F}, {0x00B879, 0x00B893, 0x000010},
    {0x00B894, 0x00B894, 0x00000F}, {0x00B895, 0x00B8AF, 0x000010}, {0x00B8B0, 0x00B8B0, 0x00000F},
    {0x00B8B1, 0x00B8CB, 0x000010}, {0x00B8CC, 0x00B8CC, 0x00000F}, {0x00B8CD, 0x00B8E7, 0x000010},
    {0x00B8E8, 0x00B8E8, 0x00000F}, {0x00B8E9, 0x00B903, 0x000010}, {0x00B904, 0x00B904, 0x00000F},
    {0x00B905, 0x00B91F, 0x000010}, {0x00B920, 0x00B920, 0x00000F}, {0x00B921, 0x00B93B, 0x000010},
    {0x00B93C, 0x00B93C, 0x00000F}, {0x00B93D, 0x00B957, 0x000010}, {0x00B958, 0x00B958, 0x00000F},
    {0x00B959, 0x00B973, 0x000010}, {0x00B974, 0x00B974, 0x00000F}, {0x00B975, 0x00B98F, 0x000010},
    {0x00B990, 0x00B990, 0x00000F}, {0x00B991, 0x00B9AB, 0x000010}, {0x00B9AC, 0x00B9AC, 0x00000F},
    {0x00B9AD, 0x00B9C7, 0x000010}, {0x00B9C8, 0x00B9C8, 0x00000F}, {0x00B9C9, 0x00B9E3, 0x000010},
    {0x00B9E4, 0x00B9E4, 0x00000F}, {0x00B9E5, 0x00B9FF, 0x000010}, {0x00BA00, 0x00BA00, 0x00000F},
    {0x00BA01, 0x00BA1B, 0x000010}, {0x00BA1C, 0x00BA1C, 0x00000F}, {0x00BA1D, 0x00BA37, 0x000010},
    {0x00BA38, 0x00BA38, 0x00000F}, {0x00BA39, 0x00BA53, 0x000010}, {0x00BA54, 0x00BA54, 0x00000F},
    {0x00BA55, 0x00BA6F, 0x000010}, {0x00BA70, 0x00BA70, 0x00000F}, {0x00BA71, 0x00BA8B, 0x000010},
    {0x00BA8C, 0x00BA8C, 0x00000F}, {0x00BA8D, 0x00BAA7, 0x000010}, {0x00BAA8, 0x00BAA8, 0x00000F},
    {0x00BAA9, 0x00BAC3, 0x000010}, {0x00BAC4, 0x00BAC4, 0x00000F}, {0x00BAC5, 0x00BADF, 0x000010},
    {0x00BAE0, 0x00BAE0, 0x00000F}, {0x00BAE1, 0x00BAFB, 0x000010}, {0x00BAFC, 0x00BAFC, 0x00000F},
    {0x00BAFD, 0x00BB17, 0x000010}, {0x00BB18, 0x00BB18, 0x00000F}, {0x00BB19, 0x00BB33, 0x000010},
    {0x00BB34, 0x00BB34, 0x00000F}, {0x00BB35, 0x00BB4F, 0x000010}, {0x00BB50, 0x00BB50, 0x00000F},
    {0x00BB51, 0x00BB6B, 0x000010}, {0x00BB6C, 0x00BB6C, 0x00000F}, {0x00BB6D, 0x00BB87, 0x000010},
    {0x00BB88, 0x00BB88, 0x00000F}, {0x00BB89, 0x00BBA3, 0x000010}, {0x00BBA4, 0x00BBA4, 0x00000F},
    {0x00BBA5, 0x00BBBF, 0x000010}, {0x00BBC0, 0x00BBC0, 0x00000F}, {0x00BBC1, 0x00BBDB, 0x000010},
    {0x00BBDC, 0x00BBDC, 0x00000F}, {0x00BBDD, 0x00BBF7, 0x000010}, {0x00BBF8, 0x00BBF8, 0x00000F},
    {0x00BBF9, 0x00BC13, 0x000010}, {0x00BC14, 0x00BC14, 0x00000F}, {0x00BC15, 0x00BC2F, 0x000010},
    {0x00BC30, 0x00BC30, 0x00000F}, {0x00BC31, 0x00BC4B, 0x000010}, {0x00BC4C, 0x00BC4C, 0x00000F},
    {0x00BC4D, 0x00BC67, 0x000010}, {0x00BC68, 0x00BC68, 0x00000F}, {0x00BC69, 0x00BC83, 0x000010},
    {0x00BC84, 0x00BC84, 0x00000F}, {0x00BC85, 0x00BC9F, 0x000010}, {0x00BCA0, 0x00BCA0, 0x00000F},
    {0x00BCA1, 0x00BCBB, 0x000010}, {0x00BCBC, 0x00BCBC, 0x00000F}, {0x00BCBD, 0x00BCD7, 0x000010},
    {0x00BCD8, 0x00BCD8, 0x00000F}, {0x00BCD9, 0x00BCF3, 0x000010}, {0x00BCF4, 0x00BCF4, 0x00000F},
    {0x00BCF5, 0x00BD0F, 0x000010}, {0x00BD10, 0x00BD10, 0x00000F}, {0x00BD11, 0x00BD2B, 0x000010},
    {0x00BD2C, 0x00BD2C, 0x00000F}, {0x00BD2D, 0x00BD47, 0x000010}, {0x00BD48, 0x00BD48, 0x00000F},
    {0x00BD49, 0x00BD63, 0x000010}, {0x00BD64, 0x00BD64, 0x00000F}, {0x00BD65, 0x00BD7F, 0x000010},
    {0x00BD80, 0x00BD80, 0x00000F}, {0x00BD81, 0x00BD9B, 0x000010}, {0x00BD9C, 0x00BD9C, 0x00000F},
    {0x00BD9D, 0x00BDB7, 0x000010}, {0x00BDB8, 0x00BDB8, 0x00000F}, {0x00BDB9, 0x00BDD3, 0x000010},
    {0x00BDD4, 0x00BDD4, 0x00000F}, {0x00BDD5, 0x00BDEF, 0x000010}, {0x00BDF0, 0x00BDF0, 0x00000F},
    {0x00BDF1, 0x00BE0B, 0x000010}, {0x00BE0C, 0x00BE0C, 0x00000F}, {0x00BE0D, 0x00BE27, 0x000010},
    {0x00BE28, 0x00BE28, 0x00000F}, {0x00BE29, 0x00BE43, 0x000010}, {0x00BE44, 0x00BE44, 0x00000F},
    {0x00BE45, 0x00BE5F, 0x000010}, {0x00BE60, 0x00BE60, 0x00000F}, {0x00BE61, 0x00BE7B, 0x000010},
    {0x00BE7C, 0x00BE7C, 0x00000F}, {0x00BE7D, 0x00BE97, 0x000010}, {0x00BE98, 0x00BE98, 0x00000F},
    {0x00BE99, 0x00BEB3, 0x000010}, {0x00BEB4, 0x00BEB4, 0x00000F}, {0x00BEB5, 0x00BECF, 0x000010},
    {0x00BED0, 0x00BED0, 0x00000F}, {0x00BED1, 0x00BEEB, 0x000010}, {0x00BEEC, 0x00BEEC, 0x00000F},
    {0x00BEED, 0x00BF07, 0x000010}, {0x00BF08, 0x00BF08, 0x00000F}, {0x00BF09, 0x00BF23, 0x000010},
    {0x00BF24, 0x00BF24, 0x00000F}, {0x00BF25, 0x00BF3F, 0x000010}, {0x00BF40, 0x00BF40, 0x00000F},
    {0x00BF41, 0x00BF5B, 0x000010}, {0x00BF5C, 0x00BF5C, 0x00000F}, {0x00BF5D, 0x00BF77, 0x000010},
    {0x00BF78, 0x00BF78, 0x00000F}, {0x00BF79, 0x00BF93, 0x000010}, {0x00BF94, 0x00BF94, 0x00000F},
    {0x00BF95, 0x00BFAF, 0x000010}, {0x00BFB0, 0x00BFB0, 0x00000F}, {0x00BFB1, 0x00BFCB, 0x000010},
    {0x00BFCC, 0x00BFCC, 0x00000F}, {0x00BFCD, 0x00BFE7, 0x000010}, {0x00BFE8, 0x00BFE8, 0x00000F},
    {0x00BFE9, 0x00C003, 0x000010}, {0x00C004, 0x00C004, 0x00000F}, {0x00C005, 0x00C01F, 0x000010},
    {0x00C020, 0x00C020, 0x00000F}, {0x00C021, 0x00C03B, 0x000010}, {0x00C03C, 0x00C03C, 0x00000F},
    {0x00C03D, 0x00C057, 0x000010}, {0x00C058, 0x00C058, 0x00000F}, {0x00C059, 0x00C073, 0x000010},
    {0x00C074, 0x00C074, 0x00000F}, {0x00C075, 0x00C08F, 0x000010}, {0x00C090, 0x00C090, 0x00000F},
    {0x00C091, 0x00C0AB, 0x000010}, {0x00C0AC, 0x00C0AC, 0x00000F}, {0x00C0AD, 0x00C0C7, 0x000010},
    {0x00C0C8, 0x00C0C8, 0x00000F}, {0x00C0C9, 0x00C0E3, 0x000010}, {0x00C0E4, 0x00C0E4, 0x00000F},
    {0x00C0E5, 0x00C0FF, 0x000010}, {0x00C100, 0x00C100, 0x00000F}, {0x00C101, 0x00C11B, 0x000010},
    {0x00C11C, 0x00C11C, 0x00000F}, {0x00C11D, 0x00C137, 0x000010}, {0x00C138, 0x00C138, 0x00000F},
    {0x00C139, 0x00C153, 0x000010}, {0x00C154, 0x00C154, 0x00000F}, {0x00C155, 0x00C16F, 0x000010},
    {0x00C170, 0x00C170, 0x00000F}, {0x00C171, 0x00C18B, 0x000010}, {0x00C18C, 0x00C18C, 0x00000F},
    {0x00C18D, 0x00C1A7, 0x000010}, {0x00C1A8, 0x00C1A8, 0x00000F}, {0x00C1A9, 0x00C1C3, 0x000010},
    {0x00C1C4, 0x00C1C4, 0x00000F}, {0x00C1C5, 0x00C1DF, 0x000010}, {0x00C1E0, 0x00C1E0, 0x00000F},
    {0x00C1E1, 0x00C1FB, 0x000010}, {0x00C1FC, 0x00C1FC, 0x00000F}, {0x00C1FD, 0x00C217, 0x000010},
    {0x00C218, 0x00C218, 0x00000F}, {0x00C219, 0x00C233, 0x000010}, {0x00C234, 0x00C234, 0x00000F},
    {0x00C235, 0x00C24F, 0x000010}, {0x00C250, 0x00C250, 0x00000F}, {0x00C251, 0x00C26B, 0x000010},
    {0x00C26C, 0x00C26C, 0x00000F}, {0x00C26D, 0x00C287, 0x000010}, {0x00C288, 0x00C288, 0x00000F},
    {0x00C289, 0x00C2A3, 0x000010}, {0x00C2A4, 0x00C2A4, 0x00000F}, {0x00C2A5, 0x00C2BF, 0x000010},
    {0x00C2C0, 0x00C2C0, 0x00000F}, {0x00C2C1, 0x00C2DB, 0x000010}, {0x00C2DC, 0x00C2DC, 0x00000F},
    {0x00C2DD, 0x00C2F7, 0x000010}, {0x00C2F8, 0x00C2F8, 0x00000F}, {0x00C2F9, 0x00C313, 0x000010},
    {0x00C314, 0x00C314, 0x00000F}, {0x00C315, 0x00C32F, 0x000010}, {0x00C330, 0x00C330, 0x00000F},
    {0x00C331, 0x00C34B, 0x000010}, {0x00C34C, 0x00C34C, 0x00000F}, {0x00C34D, 0x00C367, 0x000010},
    {0x00C368, 0x00C368, 0x00000F}, {0x00C369, 0x00C383, 0x000010}, {0x00C384, 0x00C384, 0x00000F},
    {0x00C385, 0x00C39F, 0x000010}, {0x00C3A0, 0x00C3A0, 0x00000F}, {0x00C3A1, 0x00C3BB, 0x000010},
    {0x00C3BC, 0x00C3BC, 0x00000F}, {0x00C3BD, 0x00C3D7, 0x000010}, {0x00C3D8, 0x00C3D8, 0x00000F},
    {0x00C3D9, 0x00C3F3, 0x000010}, {0x00C3F4, 0x00C3F4, 0x00000F}, {0x00C3F5, 0x00C40F, 0x000010},
    {0x00C410, 0x00C410, 0x00000F}, {0x00C411, 0x00C42B, 0x000010}, {0x00C42C, 0x00C42C, 0x00000F},
    {0x00C42D, 0x00C447, 0x000010}, {0x00C448, 0x00C448, 0x00000F}, {0x00C449, 0x00C463, 0x000010},
    {0x00C464, 0x00C464, 0x00000F}, {0x00C465, 0x00C47F, 0x000010}, {0x00C480, 0x00C480, 0x00000F},
    {0x00C481, 0x00C49B, 0x000010}, {0x00C49C, 0x00C49C, 0x00000F}, {0x00C49D, 0x00C4B7, 0x000010},
    {0x00C4B8, 0x00C4B8, 0x00000F}, {0x00C4B9, 0x00C4D3, 0x000010}, {0x00C4D4, 0x00C4D4, 0x00000F},
    {0x00C4D5, 0x00C4EF, 0x000010}, {0x00C4F0, 0x00C4F0, 0x00000F}, {0x00C4F1, 0x00C50B, 0x000010},
    {0x00C50C, 0x00C50C, 0x00000F}, {0x00C50D, 0x00C527, 0x000010}, {0x00C528, 0x00C528, 0x00000F},
    {0x00C529, 0x00C543, 0x000010}, {0x00C544, 0x00C544, 0x00000F}, {0x00C545, 0x00C55F, 0x000010},
    {0x00C560, 0x00C560, 0x00000F}, {0x00C561, 0x00C57B, 0x000010}, {0x00C57C, 0x00C57C, 0x00000F},
    {0x00C57D, 0x00C597, 0x000010}, {0x00C598, 0x00C598, 0x00000F}, {0x00C599, 0x00C5B3, 0x000010},
    {0x00C5B4, 0x00C5B4, 0x00000F}, {0x00C5B5, 0x00C5CF, 0x000010}, {0x00C5D0, 0x00C5D0, 0x00000F},
    {0x00C5D1, 0x00C5EB, 0x000010}, {0x00C5EC, 0x00C5EC, 0x00000F}, {0x00C5ED, 0x00C607, 0x000010},
    {0x00C608, 0x00C608, 0x00000F}, {0x00C609, 0x00C623, 0x000010}, {0x00C624, 0x00C624, 0x00000F},
    {0x00C625, 0x00C63F, 0x000010}, {0x00C640, 0x00C640, 0x00000F}, {0x00C641, 0x00C65B, 0x000010},
    {0x00C65C, 0x00C65C, 0x00000F}, {0x00C65D, 0x00C677, 0x000010}, {0x00C678, 0x00C678, 0x00000F},
    {0x00C679, 0x00C693, 0x000010}, {0x00C694, 0x00C694, 0x00000F}, {0x00C695, 0x00C6AF, 0x000010},
    {0x00C6B0, 0x00C6B0, 0x00000F}, {0x00C6B1, 0x00C6CB, 0x000010}, {0x00C6CC, 0x00C6CC, 0x00000F},
    {0x00C6CD, 0x00C6E7, 0x000010}, {0x00C6E8, 0x00C6E8, 0x00000F}, {0x00C6E9, 0x00C703, 0x000010},
    {0x00C704, 0x00C704, 0x00000F}, {0x00C705, 0x00C71F, 0x000010}, {0x00C720, 0x00C720, 0x00000F},
    {0x00C721, 0x00C73B, 0x000010}, {0x00C73C, 0x00C73C, 0x00000F}, {0x00C73D, 0x00C757, 0x000010},
    {0x00C758, 0x00C758, 0x00000F}, {0x00C759, 0x00C773, 0x000010}, {0x00C774, 0x00C774, 0x00000F},
    {0x00C775, 0x00C78F, 0x000010}, {0x00C790, 0x00C790, 0x00000F}, {0x00C791, 0x00C7AB, 0x000010},
    {0x00C7AC, 0x00C7AC, 0x00000F}, {0x00C7AD, 0x00C7C7, 0x000010}, {0x00C7C8, 0x00C7C8, 0x00000F},
    {0x00C7C9, 0x00C7E3, 0x000010}, {0x00C7E4, 0x00C7E4, 0x00000F}, {0x00C7E5, 0x00C7FF, 0x000010},
    {0x00C800, 0x00C800, 0x00000F}, {0x00C801, 0x00C81B, 0x000010}, {0x00C81C, 0x00C81C, 0x00000F},
    {0x00C81D, 0x00C837, 0x000010}, {0x00C838, 0x00C838, 0x00000F}, {0x00C839, 0x00C853, 0x000010},
    {0x00C854, 0x00C854, 0x00000F}, {0x00C855, 0x00C86F, 0x000010}, {0x00C870, 0x00C870, 0x00000F},
    {0x00C871, 0x00C88B, 0x000010}, {0x00C88C, 0x00C88C, 0x00000F}, {0x00C88D, 0x00C8A7, 0x000010},
    {0x00C8A8, 0x00C8A8, 0x00000F}, {0x00C8A9, 0x00C8C3, 0x000010}, {0x00C8C4, 0x00C8C4, 0x00000F},
    {0x00C8C5, 0x00C8DF, 0x000010}, {0x00C8E0, 0x00C8E0, 0x00000F}, {0x00C8E1, 0x00C8FB, 0x000010},
    {0x00C8FC, 0x00C8FC, 0x00000F}, {0x00C8FD, 0x00C917, 0x000010}, {0x00C918, 0x00C918, 0x00000F},
    {0x00C919, 0x00C933, 0x000010}, {0x00C934, 0x00C934, 0x00000F}, {0x00C935, 0x00C94F, 0x000010},
    {0x00C950, 0x00C950, 0x00000F}, {0x00C951, 0x00C96B, 0x000010}, {0x00C96C, 0x00C96C, 0x00000F},
    {0x00C96D, 0x00C987, 0x000010}, {0x00C988, 0x00C988, 0x00000F}, {0x00C989, 0x00C9A3, 0x000010},
    {0x00C9A4, 0x00C9A4, 0x00000F}, {0x00C9A5, 0x00C9BF, 0x000010}, {0x00C9C0, 0x00C9C0, 0x00000F},
    {0x00C9C1, 0x00C9DB, 0x000010}, {0x00C9DC, 0x00C9DC, 0x00000F}, {0x00C9DD, 0x00C9F7, 0x000010},
    {0x00C9F8, 0x00C9F8, 0x00000F}, {0x00C9F9, 0x00CA13, 0x000010}, {0x00CA14, 0x00CA14, 0x00000F},
    {0x00CA15, 0x00CA2F, 0x000010}, {0x00CA30, 0x00CA30, 0x00000F}, {0x00CA31, 0x00CA4B, 0x000010},
    {0x00CA4C, 0x00CA4C, 0x00000F}, {0x00CA4D, 0x00CA67, 0x000010}, {0x00CA68, 0x00CA68, 0x00000F},
    {0x00CA69, 0x00CA83, 0x000010}, {0x00CA84, 0x00CA84, 0x00000F}, {0x00CA85, 0x00CA9F, 0x000010},
    {0x00CAA0, 0x00CAA0, 0x00000F}, {0x00CAA1, 0x00CABB, 0x000010}, {0x00CABC, 0x00CABC, 0x00000F},
    {0x00CABD, 0x00CAD7, 0x000010}, {0x00CAD8, 0x00CAD8, 0x00000F}, {0x00CAD9, 0x00CAF3, 0x000010},
    {0x00CAF4, 0x00CAF4, 0x00000F}, {0x00CAF5, 0x00CB0F, 0x000010}, {0x00CB10, 0x00CB10, 0x00000F},
    {0x00CB11, 0x00CB2B, 0x000010}, {0x00CB2C, 0x00CB2C, 0x00000F}, {0x00CB2D, 0x00CB47, 0x000010},
    {0x00CB48, 0x00CB48, 0x00000F}, {0x00CB49, 0x00CB63, 0x000010}, {0x00CB64, 0x00CB64, 0x00000F},
    {0x00CB65, 0x00CB7F, 0x000010}, {0x00CB80, 0x00CB80, 0x00000F}, {0x00CB81, 0x00CB9B, 0x000010},
    {0x00CB9C, 0x00CB9C, 0x00000F}, {0x00CB9D, 0x00CBB7, 0x000010}, {0x00CBB8, 0x00CBB8, 0x00000F},
    {0x00CBB9, 0x00CBD3, 0x000010}, {0x00CBD4, 0x00CBD4, 0x00000F}, {0x00CBD5, 0x00CBEF, 0x000010},
    {0x00CBF0, 0x00CBF0, 0x00000F}, {0x00CBF1, 0x00CC0B, 0x000010}, {0x00CC0C, 0x00CC0C, 0x00000F},
    {0x00CC0D, 0x00CC27, 0x000010}, {0x00CC28, 0x00CC28, 0x00000F}, {0x00CC29, 0x00CC43, 0x000010},
    {0x00CC44, 0x00CC44, 0x00000F}, {0x00CC45, 0x00CC5F, 0x000010}, {0x00CC60, 0x00CC60, 0x00000F},
    {0x00CC61, 0x00CC7B, 0x000010}, {0x00CC7C, 0x00CC7C, 0x00000F}, {0x00CC7D, 0x00CC97, 0x000010},
    {0x00CC98, 0x00CC98, 0x00000F}, {0x00CC99, 0x00CCB3, 0x000010}, {0x00CCB4, 0x00CCB4, 0x00000F},
    {0x00CCB5, 0x00CCCF, 0x000010}, {0x00CCD0, 0x00CCD0, 0x00000F}, {0x00CCD1, 0x00CCEB, 0x000010},
    {0x00CCEC, 0x00CCEC, 0x00000F}, {0x00CCED, 0x00CD07, 0x000010}, {0x00CD08, 0x00CD08, 0x00000F},
    {0x00CD09, 0x00CD23, 0x000010}, {0x00CD24, 0x00CD24, 0x00000F}, {0x00CD25, 0x00CD3F, 0x000010},
    {0x00CD40, 0x00CD40, 0x00000F}, {0x00CD41, 0x00CD5B, 0x000010}, {0x00CD5C, 0x00CD5C, 0x00000F},
    {0x00CD5D, 0x00CD77, 0x000010}, {0x00CD78, 0x00CD78, 0x00000F}, {0x00CD79, 0x00CD93, 0x000010},
    {0x00CD94, 0x00CD94, 0x00000F}, {0x00CD95, 0x00CDAF, 0x000010}, {0x00CDB0, 0x00CDB0, 0x00000F},
    {0x00CDB1, 0x00CDCB, 0x000010}, {0x00CDCC, 0x00CDCC, 0x00000F}, {0x00CDCD, 0x00CDE7, 0x000010},
    {0x00CDE8, 0x00CDE8, 0x00000F}, {0x00CDE9, 0x00CE03, 0x000010}, {0x00CE04, 0x00CE04, 0x00000F},
    {0x00CE05, 0x00CE1F, 0x000010}, {0x00CE20, 0x00CE20, 0x00000F}, {0x00CE21, 0x00CE3B, 0x000010},
    {0x00CE3C, 0x00CE3C, 0x00000F}, {0x00CE3D, 0x00CE57, 0x000010}, {0x00CE58, 0x00CE58, 0x00000F},
    {0x00CE59, 0x00CE73, 0x000010}, {0x00CE74, 0x00CE74, 0x00000F}, {0x00CE75, 0x00CE8F, 0x000010},
    {0x00CE90, 0x00CE90, 0x00000F}, {0x00CE91, 0x00CEAB, 0x000010}, {0x00CEAC, 0x00CEAC, 0x00000F},
    {0x00CEAD, 0x00CEC7, 0x000010}, {0x00CEC8, 0x00CEC8, 0x00000F}, {0x00CEC9, 0x00CEE3, 0x000010},
    {0x00CEE4, 0x00CEE4, 0x00000F}, {0x00CEE5, 0x00CEFF, 0x000010}, {0x00CF00, 0x00CF00, 0x00000F},
    {0x00CF01, 0x00CF1B, 0x000010}, {0x00CF1C, 0x00CF1C, 0x00000F}, {0x00CF1D, 0x00CF37, 0x000010},
    {0x00CF38, 0x00CF38, 0x00000F}, {0x00CF39, 0x00CF53, 0x000010}, {0x00CF54, 0x00CF54, 0x00000F},
    {0x00CF55, 0x00CF6F, 0x000010}, {0x00CF70, 0x00CF70, 0x00000F}, {0x00CF71, 0x00CF8B, 0x000010},
    {0x00CF8C, 0x00CF8C, 0x00000F}, {0x00CF8D, 0x00CFA7, 0x000010}, {0x00CFA8, 0x00CFA8, 0x00000F},
    {0x00CFA9, 0x00CFC3, 0x000010}, {0x00CFC4, 0x00CFC4, 0x00000F}, {0x00CFC5, 0x00CFDF, 0x000010},
    {0x00CFE0, 0x00CFE0, 0x00000F}, {0x00CFE1, 0x00CFFB, 0x000010}, {0x00CFFC, 0x00CFFC, 0x00000F},
    {0x00CFFD, 0x00D017, 0x000010}, {0x00D018, 0x00D018, 0x00000F}, {0x00D019, 0x00D033, 0x000010},
    {0x00D034, 0x00D034, 0x00000F}, {0x00D035, 0x00D04F, 0x000010}, {0x00D050, 0x00D050, 0x00000F},
    {0x00D051, 0x00D06B, 0x000010}, {0x00D06C, 0x00D06C, 0x00000F}, {0x00D06D, 0x00D087, 0x000010},
    {0x00D088, 0x00D088, 0x00000F}, {0x00D089, 0x00D0A3, 0x000010}, {0x00D0A4, 0x00D0A4, 0x00000F},
    {0x00D0A5, 0x00D0BF, 0x000010}, {0x00D0C0, 0x00D0C0, 0x00000F}, {0x00D0C1, 0x00D0DB, 0x000010},
    {0x00D0DC, 0x00D0DC, 0x00000F}, {0x00D0DD, 0x00D0F7, 0x000010}, {0x00D0F8, 0x00D0F8, 0x00000F},
    {0x00D0F9, 0x00D113, 0x000010}, {0x00D114, 0x00D114, 0x00000F}, {0x00D115, 0x00D12F, 0x000010},
    {0x00D130, 0x00D130, 0x00000F}, {0x00D131, 0x00D14B, 0x000010}, {0x00D14C, 0x00D14C, 0x00000F},
    {0x00D14D, 0x00D167, 0x000010}, {0x00D168, 0x00D168, 0x00000F}, {0x00D169, 0x00D183, 0x000010},
    {0x00D184, 0x00D184, 0x00000F}, {0x00D185, 0x00D19F, 0x000010}, {0x00D1A0, 0x00D1A0, 0x00000F},
    {0x00D1A1, 0x00D1BB, 0x000010}, {0x00D1BC, 0x00D1BC, 0x00000F}, {0x00D1BD, 0x00D1D7, 0x000010},
    {0x00D1D8, 0x00D1D8, 0x00000F}, {0x00D1D9, 0x00D1F3, 0x000010}, {0x00D1F4, 0x00D1F4, 0x00000F},
    {0x00D1F5, 0x00D20F, 0x000010}, {0x00D210, 0x00D210, 0x00000F}, {0x00D211, 0x00D22B, 0x000010},
    {0x00D22C, 0x00D22C, 0x00000F}, {0x00D22D, 0x00D247, 0x000010}, {0x00D248, 0x00D248, 0x00000F},
    {0x00D249, 0x00D263, 0x000010}, {0x00D264, 0x00D264, 0x00000F}, {0x00D265, 0x00D27F, 0x000010},
    {0x00D280, 0x00D280, 0x00000F}, {0x00D281, 0x00D29B, 0x000010}, {0x00D29C, 0x00D29C, 0x00000F},
    {0x00D29D, 0x00D2B7, 0x000010}, {0x00D2B8, 0x00D2B8, 0x00000F}, {0x00D2B9, 0x00D2D3, 0x000010},
    {0x00D2D4, 0x00D2D4, 0x00000F}, {0x00D2D5, 0x00D2EF, 0x000010}, {0x00D2F0, 0x00D2F0, 0x00000F},
    {0x00D2F1, 0x00D30B, 0x000010}, {0x00D30C, 0x00D30C, 0x00000F}, {0x00D30D, 0x00D327, 0x000010},
    {0x00D328, 0x00D328, 0x00000F}, {0x00D329, 0x00D343, 0x000010}, {0x00D344, 0x00D344, 0x00000F},
    {0x00D345, 0x00D35F, 0x000010}, {0x00D360, 0x00D360, 0x00000F}, {0x00D361, 0x00D37B, 0x000010},
    {0x00D37C, 0x00D37C, 0x00000F}, {0x00D37D, 0x00D397, 0x000010}, {0x00D398, 0x00D398, 0x00000F},
    {0x00D399, 0x00D3B3, 0x000010}, {0x00D3B4, 0x00D3B4, 0x00000F}, {0x00D3B5, 0x00D3CF, 0x000010},
    {0x00D3D0, 0x00D3D0, 0x00000F}, {0x00D3D1, 0x00D3EB, 0x000010}, {0x00D3EC, 0x00D3EC, 0x00000F},
    {0x00D3ED, 0x00D407, 0x000010}, {0x00D408, 0x00D408, 0x00000F}, {0x00D409, 0x00D423, 0x000010},
    {0x00D424, 0x00D424, 0x00000F}, {0x00D425, 0x00D43F, 0x000010}, {0x00D440, 0x00D440, 0x00000F},
    {0x00D441, 0x00D45B, 0x000010}, {0x00D45C, 0x00D45C, 0x00000F}, {0x00D45D, 0x00D477, 0x000010},
    {0x00D478, 0x00D478, 0x00000F}, {0x00D479, 0x00D493, 0x000010}, {0x00D494, 0x00D494, 0x00000F},
    {0x00D495, 0x00D4AF, 0x000010}, {0x00D4B0, 0x00D4B0, 0x00000F}, {0x00D4B1, 0x00D4CB, 0x000010},
    {0x00D4CC, 0x00D4CC, 0x00000F}, {0x00D4CD, 0x00D4E7, 0x000010}, {0x00D4E8, 0x00D4E8, 0x00000F},
    {0x00D4E9, 0x00D503, 0x000010}, {0x00D504, 0x00D504, 0x00000F}, {0x00D505, 0x00D51F, 0x000010},
    {0x00D520, 0x00D520, 0x00000F}, {0x00D521, 0x00D53B, 0x000010}, {0x00D53C, 0x00D53C, 0x00000F},
    {0x00D53D, 0x00D557, 0x000010}, {0x00D558, 0x00D558, 0x00000F}, {0x00D559, 0x00D573, 0x000010},
    {0x00D574, 0x00D574, 0x00000F}, {0x00D575, 0x00D58F, 0x000010}, {0x00D590, 0x00D590, 0x00000F},
    {0x00D591, 0x00D5AB, 0x000010}, {0x00D5AC, 0x00D5AC, 0x00000F}, {0x00D5AD, 0x00D5C7, 0x000010},
    {0x00D5C8, 0x00D5C8, 0x00000F}, {0x00D5C9, 0x00D5E3, 0x000010}, {0x00D5E4, 0x00D5E4, 0x00000F},
    {0x00D5E5, 0x00D5FF, 0x000010}, {0x00D600, 0x00D600, 0x00000F}, {0x00D601, 0x00D61B, 0x000010},
    {0x00D61C, 0x00D61C, 0x00000F}, {0x00D61D, 0x00D637, 0x000010}, {0x00D638, 0x00D638, 0x00000F},
    {0x00D639, 0x00D653, 0x000010}, {0x00D654, 0x00D654, 0x00000F}, {0x00D655, 0x00D66F, 0x000010},
    {0x00D670, 0x00D670, 0x00000F}, {0x00D671, 0x00D68B, 0x000010}, {0x00D68C, 0x00D68C, 0x00000F},
    {0x00D68D, 0x00D6A7, 0x000010}, {0x00D6A8, 0x00D6A8, 0x00000F}, {0x00D6A9, 0x00D6C3, 0x000010},
    {0x00D6C4, 0x00D6C4, 0x00000F}, {0x00D6C5, 0x00D6DF, 0x000010}, {0x00D6E0, 0x00D6E0, 0x00000F},
    {0x00D6E1, 0x00D6FB, 0x000010}, {0x00D6FC, 0x00D6FC, 0x00000F}, {0x00D6FD, 0x00D717, 0x000010},
    {0x00D718, 0x00D718, 0x00000F}, {0x00D719, 0x00D733, 0x000010}, {0x00D734, 0x00D734, 0x00000F},
    {0x00D735, 0x00D74F, 0x000010}, {0x00D750, 0x00D750, 0x00000F}, {0x00D751, 0x00D76B, 0x000010},
    {0x00D76C, 0x00D76C, 0x00000F}, {0x00D76D, 0x00D787, 0x000010}, {0x00D788, 0x00D788, 0x00000F},
    {0x00D789, 0x00D7A3, 0x000010}, {0x00D7B0, 0x00D7C6, 0x000018}, {0x00D7CB, 0x00D7FB, 0x000017},
    {0x00D800, 0x00DB7F, 0x000001}, {0x00DB80, 0x00DBFF, 0x000001}, {0x00DC00, 0x00DFFF, 0x000001},
    {0x00E000, 0x00F8FF, 0x000001}, {0x00F900, 0x00FA6D, 0x000012}, {0x00FA6E, 0x00FA6F, 0x000012},
    {0x00FA70, 0x00FAD9, 0x000012}, {0x00FADA, 0x00FAFF, 0x000012}, {0x00FB00, 0x00FB06, 0x000001},
    {0x00FB13, 0x00FB17, 0x000001}, {0x00FB1D, 0x00FB1D, 0x000013}, {0x00FB1E, 0x00FB1E, 0x000008},
    {0x00FB1F, 0x00FB28, 0x000013}, {0x00FB29, 0x00FB29, 0x000001}, {0x00FB2A, 0x00FB36, 0x000013},
    {0x00FB38, 0x00FB3C, 0x000013}, {0x00FB3E, 0x00FB3E, 0x000013}, {0x00FB40, 0x00FB41, 0x000013},
    {0x00FB43, 0x00FB44, 0x000013}, {0x00FB46, 0x00FB4F, 0x000013}, {0x00FB50, 0x00FBB1, 0x000001},
    {0x00FBB2, 0x00FBC2, 0x000001}, {0x00FBD3, 0x00FD3D, 0x000001}, {0x00FD3E, 0x00FD3E, 0x000007},
    {0x00FD3F, 0x00FD3F, 0x00001D}, {0x00FD40, 0x00FD4F, 0x000001}, {0x00FD50, 0x00FD8F, 0x000001},
    {0x00FD92, 0x00FDC7, 0x000001}, {0x00FDCF, 0x00FDCF, 0x000001}, {0x00FDF0, 0x00FDFB, 0x000001},
    {0x00FDFC, 0x00FDFC, 0x00001E}, {0x00FDFD, 0x00FDFF, 0x000001}, {0x00FE00, 0x00FE0F, 0x000008},
    {0x00FE10, 0x00FE10, 0x000015}, {0x00FE11, 0x00FE12, 0x000007}, {0x00FE13, 0x00FE14, 0x000015},
    {0x00FE15, 0x00FE16, 0x00000D}, {0x00FE17, 0x00FE17, 0x00001D}, {0x00FE18, 0x00FE18, 0x000007},
    {0x00FE19, 0x00FE19, 0x000014}, {0x00FE20, 0x00FE2F, 0x000008}, {0x00FE30, 0x00FE30, 0x000012},
    {0x00FE31, 0x00FE32, 0x000012}, {0x00FE33, 0x00FE34, 0x000012}, {0x00FE35, 0x00FE35, 0x00001D},
    {0x00FE36, 0x00FE36, 0x000007}, {0x00FE37, 0x00FE37, 0x00001D}, {0x00FE38, 0x00FE38, 0x000007},
    {0x00FE39, 0x00FE39, 0x00001D}, {0x00FE3A, 0x00FE3A, 0x000007}, {0x00FE3B, 0x00FE3B, 0x00001D},
    {0x00FE3C, 0x00FE3C, 0x000007}, {0x00FE3D, 0x00FE3D, 0x00001D}, {0x00FE3E, 0x00FE3E, 0x000007},
    {0x00FE3F, 0x00FE3F, 0x00001D}, {0x00FE40, 0x00FE40, 0x000007}, {0x00FE41, 0x00FE41, 0x00001D},
    {0x00FE42, 0x00FE42, 0x000007}, {0x00FE43, 0x00FE43, 0x00001D}, {0x00FE44, 0x00FE44, 0x000007},
    {0x00FE45, 0x00FE46, 0x000012}, {0x00FE47, 0x00FE47, 0x00001D}, {0x00FE48, 0x00FE48, 0x000007},
    {0x00FE49, 0x00FE4C, 0x000012}, {0x00FE4D, 0x00FE4F, 0x000012}, {0x00FE50, 0x00FE50, 0x000007},
    {0x00FE51, 0x00FE51, 0x000012}, {0x00FE52, 0x00FE52, 0x000007}, {0x00FE54, 0x00FE55, 0x00001B},
    {0x00FE56, 0x00FE57, 0x00000D}, {0x00FE58, 0x00FE58, 0x000012}, {0x00FE59, 0x00FE59, 0x00001D},
    {0x00FE5A, 0x00FE5A, 0x000007}, {0x00FE5B, 0x00FE5B, 0x00001D}, {0x00FE5C, 0x00FE5C, 0x000007},
    {0x00FE5D, 0x00FE5D, 0x00001D}, {0x00FE5E, 0x00FE5E, 0x000007}, {0x00FE5F, 0x00FE61, 0x000012},
    {0x00FE62, 0x00FE62, 0x000012}, {0x00FE63, 0x00FE63, 0x000012}, {0x00FE64, 0x00FE66, 0x000012},
    {0x00FE68, 0x00FE68, 0x000012}, {0x00FE69, 0x00FE69, 0x00001F}, {0x00FE6A, 0x00FE6A, 0x00001E},
    {0x00FE6B, 0x00FE6B, 0x000012}, {0x00FE70, 0x00FE74, 0x000001}, {0x00FE76, 0x00FEFC, 0x000001},
    {0x00FEFF, 0x00FEFF, 0x000024}, {0x00FF01, 0x00FF01, 0x00000D}, {0x00FF02, 0x00FF03, 0x000012},
    {0x00FF04, 0x00FF04, 0x00001F}, {0x00FF05, 0x00FF05, 0x00001E}, {0x00FF06, 0x00FF07, 0x000012},
    {0x00FF08, 0x00FF08, 0x00001D}, {0x00FF09, 0x00FF09, 0x000007}, {0x00FF0A, 0x00FF0A, 0x000012},
    {0x00FF0B, 0x00FF0B, 0x000012}, {0x00FF0C, 0x00FF0C, 0x000007}, {0x00FF0D, 0x00FF0D, 0x000012},
    {0x00FF0E, 0x00FF0E, 0x000007}, {0x00FF0F, 0x00FF0F, 0x000012}, {0x00FF10, 0x00FF19, 0x000012},
    {0x00FF1A, 0x00FF1B, 0x00001B}, {0x00FF1C, 0x00FF1E, 0x000012}, {0x00FF1F, 0x00FF1F, 0x00000D},
    {0x00FF20, 0x00FF20, 0x000012}, {0x00FF21, 0x00FF3A, 0x000012}, {0x00FF3B, 0x00FF3B, 0x00001D},
    {0x00FF3C, 0x00FF3C, 0x000012}, {0x00FF3D, 0x00FF3D, 0x000007}, {0x00FF3E, 0x00FF3E, 0x000012},
    {0x00FF3F, 0x00FF3F, 0x000012}, {0x00FF40, 0x00FF40, 0x000012}, {0x00FF41, 0x00FF5A, 0x000012},
    {0x00FF5B, 0x00FF5B, 0x00001D}, {0x00FF5C, 0x00FF5C, 0x000012}, {0x00FF5D, 0x00FF5D, 0x000007},
    {0x00FF5E, 0x00FF5E, 0x000012}, {0x00FF5F, 0x00FF5F, 0x00001D}, {0x00FF60, 0x00FF60, 0x000007},
    {0x00FF61, 0x00FF61, 0x000007}, {0x00FF62, 0x00FF62, 0x00001D}, {0x00FF63, 0x00FF63, 0x000007},
    {0x00FF64, 0x00FF64, 0x000007}, {0x00FF65, 0x00FF65, 0x00001B}, {0x00FF66, 0x00FF66, 0x000012},
    {0x00FF67, 0x00FF6F, 0x00001B}, {0x00FF70, 0x00FF70, 0x00001B}, {0x00FF71, 0x00FF9D, 0x000012},
    {0x00FF9E, 0x00FF9F, 0x00001B}, {0x00FFA0, 0x00FFBE, 0x000012}, {0x00FFC2, 0x00FFC7, 0x000012},
    {0x00FFCA, 0x00FFCF, 0x000012}, {0x00FFD2, 0x00FFD7, 0x000012}, {0x00FFDA, 0x00FFDC, 0x000012},
    {0x00FFE0, 0x00FFE0, 0x00001E}, {0x00FFE1, 0x00FFE1, 0x00001F}, {0x00FFE2, 0x00FFE2, 0x000012},
    {0x00FFE3, 0x00FFE3, 0x000012}, {0x00FFE4, 0x00FFE4, 0x000012}, {0x00FFE5, 0x00FFE6, 0x00001F},
    {0x00FFE8, 0x00FFE8, 0x000001}, {0x00FFE9, 0x00FFEC, 0x000001}, {0x00FFED, 0x00FFEE, 0x000001},
    {0x00FFF9, 0x00FFFB, 0x000008}, {0x00FFFC, 0x00FFFC, 0x000006}, {0x00FFFD, 0x00FFFD, 0x000001},
    {0x010000, 0x01000B, 0x000001}, {0x01000D, 0x010026, 0x000001}, {0x010028, 0x01003A, 0x000001},
    {0x01003C, 0x01003D, 0x000001}, {0x01003F, 0x01004D, 0x000001}, {0x010050, 0x01005D, 0x000001},
    {0x010080, 0x0100FA, 0x000001}, {0x010100, 0x010102, 0x000002}, {0x010107, 0x010133, 0x000001},
    {0x010137, 0x01013F, 0x000001}, {0x010140, 0x010174, 0x000001}, {0x010175, 0x010178, 0x000001},
    {0x010179, 0x010189, 0x000001}, {0x01018A, 0x01018B, 0x000001}, {0x01018C, 0x01018E, 0x000001},
    {0x010190, 0x01019C, 0x000001}, {0x0101A0, 0x0101A0, 0x000001}, {0x0101D0, 0x0101FC, 0x000001},
    {0x0101FD, 0x0101FD, 0x000008}, {0x010280, 0x01029C, 0x000001}, {0x0102A0, 0x0102D0, 0x000001},
    {0x0102E0, 0x0102E0, 0x000008}, {0x0102E1, 0x0102FB, 0x000001}, {0x010300, 0x01031F, 0x000001},
    {0x010320, 0x010323, 0x000001}, {0x01032D, 0x01032F, 0x000001}, {0x010330, 0x010340, 0x000001},
    {0x010341, 0x010341, 0x000001}, {0x010342, 0x010349, 0x000001}, {0x01034A, 0x01034A, 0x000001},
    {0x010350, 0x010375, 0x000001}, {0x010376, 0x01037A, 0x000008}, {0x010380, 0x01039D, 0x000001},
    {0x01039F, 0x01039F, 0x000002}, {0x0103A0, 0x0103C3, 0x000001}, {0x0103C8, 0x0103CF, 0x000001},
    {0x0103D0, 0x0103D0, 0x000002}, {0x0103D1, 0x0103D5, 0x000001}, {0x010400, 0x01044F, 0x000001},
    {0x010450, 0x01047F, 0x000001}, {0x010480, 0x01049D, 0x000001}, {0x0104A0, 0x0104A9, 0x00001C},
    {0x0104B0, 0x0104D3, 0x000001}, {0x0104D8, 0x0104FB, 0x000001}, {0x010500, 0x010527, 0x000001},
    {0x010530, 0x010563, 0x000001}, {0x01056F, 0x01056F, 0x000001}, {0x010570, 0x01057A, 0x000001},
    {0x01057C, 0x01058A, 0x000001}, {0x01058C, 0x010592, 0x000001}, {0x010594, 0x010595, 0x000001},
    {0x010597, 0x0105A1, 0x000001}, {0x0105A3, 0x0105B1, 0x000001}, {0x0105B3, 0x0105B9, 0x000001},
    {0x0105BB, 0x0105BC, 0x000001}, {0x010600, 0x010736, 0x000001}, {0x010740, 0x010755, 0x000001},
    {0x010760, 0x010767, 0x000001}, {0x010780, 0x010785, 0x000001}, {0x010787, 0x0107B0, 0x000001},
    {0x0107B2, 0x0107BA, 0x000001}, {0x010800, 0x010805, 0x000001}, {0x010808, 0x010808, 0x000001},
    {0x01080A, 0x010835, 0x000001}, {0x010837, 0x010838, 0x000001}, {0x01083C, 0x01083C, 0x000001},
    {0x01083F, 0x01083F, 0x000001}, {0x010840, 0x010855, 0x000001}, {0x010857, 0x010857, 0x000002},
    {0x010858, 0x01085F, 0x000001}, {0x010860, 0x010876, 0x000001}, {0x010877, 0x010878, 0x000001},
    {0x010879, 0x01087F, 0x000001}, {0x010880, 0x01089E, 0x000001}, {0x0108A7, 0x0108AF, 0x000001},
    {0x0108E0, 0x0108F2, 0x000001}, {0x0108F4, 0x0108F5, 0x000001}, {0x0108FB, 0x0108FF, 0x000001},
    {0x010900, 0x010915, 0x000001}, {0x010916, 0x01091B, 0x000001}, {0x01091F, 0x01091F, 0x000002},
    {0x010920, 0x010939, 0x000001}, {0x01093F, 0x01093F, 0x000001}, {0x010980, 0x01099F, 0x000001},
    {0x0109A0, 0x0109B7, 0x000001}, {0x0109BC, 0x0109BD, 0x000001}, {0x0109BE, 0x0109BF, 0x000001},
    {0x0109C0, 0x0109CF, 0x000001}, {0x0109D2, 0x0109FF, 0x000001}, {0x010A00, 0x010A00, 0x000001},
    {0x010A01, 0x010A03, 0x000008}, {0x010A05, 0x010A06, 0x000008}, {0x010A0C, 0x010A0F, 0x000008},
    {0x010A10, 0x010A13, 0x000001}, {0x010A15, 0x010A17, 0x000001}, {0x010A19, 0x010A35, 0x000001},
    {0x010A38, 0x010A3A, 0x000008}, {0x010A3F, 0x010A3F, 0x000008}, {0x010A40, 0x010A48, 0x000001},
    {0x010A50, 0x010A57, 0x000002}, {0x010A58, 0x010A58, 0x000001}, {0x010A60, 0x010A7C, 0x000001},
    {0x010A7D, 0x010A7E, 0x000001}, {0x010A7F, 0x010A7F, 0x000001}, {0x010A80, 0x010A9C, 0x000001},
    {0x010A9D, 0x010A9F, 0x000001}, {0x010AC0, 0x010AC7, 0x000001}, {0x010AC8, 0x010AC8, 0x000001},
    {0x010AC9, 0x010AE4, 0x000001}, {0x010AE5, 0x010AE6, 0x000008}, {0x010AEB, 0x010AEF, 0x000001},
    {0x010AF0, 0x010AF5, 0x000002}, {0x010AF6, 0x010AF6, 0x000014}, {0x010B00, 0x010B35, 0x000001},
    {0x010B39, 0x010B3F, 0x000002}, {0x010B40, 0x010B55, 0x000001}, {0x010B58, 0x010B5F, 0x000001},
    {0x010B60, 0x010B72, 0x000001}, {0x010B78, 0x010B7F, 0x000001}, {0x010B80, 0x010B91, 0x000001},
    {0x010B99, 0x010B9C, 0x000001}, {0x010BA9, 0x010BAF, 0x000001}, {0x010C00, 0x010C48, 0x000001},
    {0x010C80, 0x010CB2, 0x000001}, {0x010CC0, 0x010CF2, 0x000001}, {0x010CFA, 0x010CFF, 0x000001},
    {0x010D00, 0x010D23, 0x000001}, {0x010D24, 0x010D27, 0x000008}, {0x010D30, 0x010D39, 0x00001C},
    {0x010E60, 0x010E7E, 0x000001}, {0x010E80, 0x010EA9, 0x000001}, {0x010EAB, 0x010EAC, 0x000008},
    {0x010EAD, 0x010EAD, 0x000002}, {0x010EB0, 0x010EB1, 0x000001}, {0x010F00, 0x010F1C, 0x000001},
    {0x010F1D, 0x010F26, 0x000001}, {0x010F27, 0x010F27, 0x000001}, {0x010F30, 0x010F45, 0x000001},
    {0x010F46, 0x010F50, 0x000008}, {0x010F51, 0x010F54, 0x000001}, {0x010F55, 0x010F59, 0x000001},
    {0x010F70, 0x010F81, 0x000001}, {0x010F82, 0x010F85, 0x000008}, {0x010F86, 0x010F89, 0x000001},
    {0x010FB0, 0x010FC4, 0x000001}, {0x010FC5, 0x010FCB, 0x000001}, {0x010FE0, 0x010FF6, 0x000001},
    {0x011000, 0x011000, 0x000008}, {0x011001, 0x011001, 0x000008}, {0x011002, 0x011002, 0x000008},
    {0x011003, 0x011037, 0x000001}, {0x011038, 0x011046, 0x000008}, {0x011047, 0x011048, 0x000002},
    {0x011049, 0x01104D, 0x000001}, {0x011052, 0x011065, 0x000001}, {0x011066, 0x01106F, 0x00001C},
    {0x011070, 0x011070, 0x000008}, {0x011071, 0x011072, 0x000001}, {0x011073, 0x011074, 0x000008},
    {0x011075, 0x011075, 0x000001}, {0x01107F, 0x01107F, 0x000008}, {0x011080, 0x011081, 0x000008},
    {0x011082, 0x011082, 0x000008}, {0x011083, 0x0110AF, 0x000001}, {0x0110B0, 0x0110B2, 0x000008},
    {0x0110B3, 0x0110B6, 0x000008}, {0x0110B7, 0x0110B8, 0x000008}, {0x0110B9, 0x0110BA, 0x000008},
    {0x0110BB, 0x0110BC, 0x000001}, {0x0110BD, 0x0110BD, 0x000001}, {0x0110BE, 0x0110C1, 0x000002},
    {0x0110C2, 0x0110C2, 0x000008}, {0x0110CD, 0x0110CD, 0x000001}, {0x0110D0, 0x0110E8, 0x000001},
    {0x0110F0, 0x0110F9, 0x00001C}, {0x011100, 0x011102, 0x000008}, {0x011103, 0x011126, 0x000001},
    {0x011127, 0x01112B, 0x000008}, {0x01112C, 0x01112C, 0x000008}, {0x01112D, 0x011134, 0x000008},
    {0x011136, 0x01113F, 0x00001C}, {0x011140, 0x011143, 0x000002}, {0x011144, 0x011144, 0x000001},
    {0x011145, 0x011146, 0x000008}, {0x011147, 0x011147, 0x000001}, {0x011150, 0x011172, 0x000001},
    {0x011173, 0x011173, 0x000008}, {0x011174, 0x011174, 0x000001}, {0x011175, 0x011175, 0x000003},
    {0x011176, 0x011176, 0x000001}, {0x011180, 0x011181, 0x000008}, {0x011182, 0x011182, 0x000008},
    {0x011183, 0x0111B2, 0x000001}, {0x0111B3, 0x0111B5, 0x000008}, {0x0111B6, 0x0111BE, 0x000008},
    {0x0111BF, 0x0111C0, 0x000008}, {0x0111C1, 0x0111C4, 0x000001}, {0x0111C5, 0x0111C6, 0x000002},
    {0x0111C7, 0x0111C7, 0x000001}, {0x0111C8, 0x0111C8, 0x000002}, {0x0111C9, 0x0111CC, 0x000008},
    {0x0111CD, 0x0111CD, 0x000001}, {0x0111CE, 0x0111CE, 0x000008}, {0x0111CF, 0x0111CF, 0x000008},
    {0x0111D0, 0x0111D9, 0x00001C}, {0x0111DA, 0x0111DA, 0x000001}, {0x0111DB, 0x0111DB, 0x000003},
    {0x0111DC, 0x0111DC, 0x000001}, {0x0111DD, 0x0111DF, 0x000002}, {0x0111E1, 0x0111F4, 0x000001},
    {0x011200, 0x011211, 0x000001}, {0x011213, 0x01122B, 0x000001}, {0x01122C, 0x01122E, 0x000008},
    {0x01122F, 0x011231, 0x000008}, {0x011232, 0x011233, 0x000008}, {0x011234, 0x011234, 0x000008},
    {0x011235, 0x011235, 0x000008}, {0x011236, 0x011237, 0x000008}, {0x011238, 0x011239, 0x000002},
    {0x01123A, 0x01123A, 0x000001}, {0x01123B, 0x01123C, 0x000002}, {0x01123D, 0x01123D, 0x000001},
    {0x01123E, 0x01123E, 0x000008}, {0x011280, 0x011286, 0x000001}, {0x011288, 0x011288, 0x000001},
    {0x01128A, 0x01128D, 0x000001}, {0x01128F, 0x01129D, 0x000001}, {0x01129F, 0x0112A8, 0x000001},
    {0x0112A9, 0x0112A9, 0x000002}, {0x0112B0, 0x0112DE, 0x000001}, {0x0112DF, 0x0112DF, 0x000008},
    {0x0112E0, 0x0112E2, 0x000008}, {0x0112E3, 0x0112EA, 0x000008}, {0x0112F0, 0x0112F9, 0x00001C},
    {0x011300, 0x011301, 0x000008}, {0x011302, 0x011303, 0x000008}, {0x011305, 0x01130C, 0x000001},
    {0x01130F, 0x011310, 0x000001}, {0x011313, 0x011328, 0x000001}, {0x01132A, 0x011330, 0x000001},
    {0x011332, 0x011333, 0x000001}, {0x011335, 0x011339, 0x000001}, {0x01133B, 0x01133C, 0x000008},
    {0x01133D, 0x01133D, 0x000001}, {0x01133E, 0x01133F, 0x000008}, {0x011340, 0x011340, 0x000008},
    {0x011341, 0x011344, 0x000008}, {0x011347, 0x011348, 0x000008}, {0x01134B, 0x01134D, 0x000008},
    {0x011350, 0x011350, 0x000001}, {0x011357, 0x011357, 0x000008}, {0x01135D, 0x011361, 0x000001},
    {0x011362, 0x011363, 0x000008}, {0x011366, 0x01136C, 0x000008}, {0x011370, 0x011374, 0x000008},
    {0x011400, 0x011434, 0x000001}, {0x011435, 0x011437, 0x000008}, {0x011438, 0x01143F, 0x000008},
    {0x011440, 0x011441, 0x000008}, {0x011442, 0x011444, 0x000008}, {0x011445, 0x011445, 0x000008},
    {0x011446, 0x011446, 0x000008}, {0x011447, 0x01144A, 0x000001}, {0x01144B, 0x01144E, 0x000002},
    {0x01144F, 0x01144F, 0x000001}, {0x011450, 0x011459, 0x00001C}, {0x01145A, 0x01145B, 0x000002},
    {0x01145D, 0x01145D, 0x000001}, {0x01145E, 0x01145E, 0x000008}, {0x01145F, 0x011461, 0x000001},
    {0x011480, 0x0114AF, 0x000001}, {0x0114B0, 0x0114B2, 0x000008}, {0x0114B3, 0x0114B8, 0x000008},
    {0x0114B9, 0x0114B9, 0x000008}, {0x0114BA, 0x0114BA, 0x000008}, {0x0114BB, 0x0114BE, 0x000008},
    {0x0114BF, 0x0114C0, 0x000008}, {0x0114C1, 0x0114C1, 0x000008}, {0x0114C2, 0x0114C3, 0x000008},
    {0x0114C4, 0x0114C5, 0x000001}, {0x0114C6, 0x0114C6, 0x000001}, {0x0114C7, 0x0114C7, 0x000001},
    {0x0114D0, 0x0114D9, 0x00001C}, {0x011580, 0x0115AE, 0x000001}, {0x0115AF, 0x0115B1, 0x000008},
    {0x0115B2, 0x0115B5, 0x000008}, {0x0115B8, 0x0115BB, 0x000008}, {0x0115BC, 0x0115BD, 0x000008},
    {0x0115BE, 0x0115BE, 0x000008}, {0x0115BF, 0x0115C0, 0x000008}, {0x0115C1, 0x0115C1, 0x000003},
    {0x0115C2, 0x0115C3, 0x000002}, {0x0115C4, 0x0115C5, 0x00000D}, {0x0115C6, 0x0115C8, 0x000001},
    {0x0115C9, 0x0115D7, 0x000002}, {0x0115D8, 0x0115DB, 0x000001}, {0x0115DC, 0x0115DD, 0x000008},
    {0x011600, 0x01162F, 0x000001}, {0x011630, 0x011632, 0x000008}, {0x011633, 0x01163A, 0x000008},
    {0x01163B, 0x01163C, 0x000008}, {0x01163D, 0x01163D, 0x000008}, {0x01163E, 0x01163E, 0x000008},
    {0x01163F, 0x011640, 0x000008}, {0x011641, 0x011642, 0x000002}, {0x011643, 0x011643, 0x000001},
    {0x011644, 0x011644, 0x000001}, {0x011650, 0x011659, 0x00001C}, {0x011660, 0x01166C, 0x000003},
    {0x011680, 0x0116AA, 0x000001}, {0x0116AB, 0x0116AB, 0x000008}, {0x0116AC, 0x0116AC, 0x000008},
    {0x0116AD, 0x0116AD, 0x000008}, {0x0116AE, 0x0116AF, 0x000008}, {0x0116B0, 0x0116B5, 0x000008},
    {0x0116B6, 0x0116B6, 0x000008}, {0x0116B7, 0x0116B7, 0x000008}, {0x0116B8, 0x0116B8, 0x000001},
    {0x0116B9, 0x0116B9, 0x000001}, {0x0116C0, 0x0116C9, 0x00001C}, {0x011700, 0x01171A, 0x000001},
    {0x01171D, 0x01171F, 0x000008}, {0x011720, 0x011721, 0x000008}, {0x011722, 0x011725, 0x000008},
    {0x011726, 0x011726, 0x000008}, {0x011727, 0x01172B, 0x000008}, {0x011730, 0x011739, 0x00001C},
    {0x01173A, 0x01173B, 0x000001}, {0x01173C, 0x01173E, 0x000002}, {0x01173F, 0x01173F, 0x000001},
    {0x011740, 0x011746, 0x000001}, {0x011800, 0x01182B, 0x000001}, {0x01182C, 0x01182E, 0x000008},
    {0x01182F, 0x011837, 0x000008}, {0x011838, 0x011838, 0x000008}, {0x011839, 0x01183A, 0x000008},
    {0x01183B, 0x01183B, 0x000001}, {0x0118A0, 0x0118DF, 0x000001}, {0x0118E0, 0x0118E9, 0x00001C},
    {0x0118EA, 0x0118F2, 0x000001}, {0x0118FF, 0x0118FF, 0x000001}, {0x011900, 0x011906, 0x000001},
    {0x011909, 0x011909, 0x000001}, {0x01190C, 0x011913, 0x000001}, {0x011915, 0x011916, 0x000001},
    {0x011918, 0x01192F, 0x000001}, {0x011930, 0x011935, 0x000008}, {0x011937, 0x011938, 0x000008},
    {0x01193B, 0x01193C, 0x000008}, {0x01193D, 0x01193D, 0x000008}, {0x01193E, 0x01193E, 0x000008},
    {0x01193F, 0x01193F, 0x000001}, {0x011940, 0x011940, 0x000008}, {0x011941, 0x011941, 0x000001},
    {0x011942, 0x011942, 0x000008}, {0x011943, 0x011943, 0x000008}, {0x011944, 0x011946, 0x000002},
    {0x011950, 0x011959, 0x00001C}, {0x0119A0, 0x0119A7, 0x000001}, {0x0119AA, 0x0119D0, 0x000001},
    {0x0119D1, 0x0119D3, 0x000008}, {0x0119D4, 0x0119D7, 0x000008}, {0x0119DA, 0x0119DB, 0x000008},
    {0x0119DC, 0x0119DF, 0x000008}, {0x0119E0, 0x0119E0, 0x000008}, {0x0119E1, 0x0119E1, 0x000001},
    {0x0119E2, 0x0119E2, 0x000003}, {0x0119E3, 0x0119E3, 0x000001}, {0x0119E4, 0x0119E4, 0x000008},
    {0x011A00, 0x011A00, 0x000001}, {0x011A01, 0x011A0A, 0x000008}, {0x011A0B, 0x011A32, 0x000001},
    {0x011A33, 0x011A38, 0x000008}, {0x011A39, 0x011A39, 0x000008}, {0x011A3A, 0x011A3A, 0x000001},
    {0x011A3B, 0x011A3E, 0x000008}, {0x011A3F, 0x011A3F, 0x000003}, {0x011A40, 0x011A40, 0x000001},
    {0x011A41, 0x011A44, 0x000002}, {0x011A45, 0x011A45, 0x000003}, {0x011A46, 0x011A46, 0x000001},
    {0x011A47, 0x011A47, 0x000008}, {0x011A50, 0x011A50, 0x000001}, {0x011A51, 0x011A56, 0x000008},
    {0x011A57, 0x011A58, 0x000008}, {0x011A59, 0x011A5B, 0x000008}, {0x011A5C, 0x011A89, 0x000001},
    {0x011A8A, 0x011A96, 0x000008}, {0x011A97, 0x011A97, 0x000008}, {0x011A98, 0x011A99, 0x000008},
    {0x011A9A, 0x011A9C, 0x000002}, {0x011A9D, 0x011A9D, 0x000001}, {0x011A9E, 0x011AA0, 0x000003},
    {0x011AA1, 0x011AA2, 0x000002}, {0x011AB0, 0x011ABF, 0x000001}, {0x011AC0, 0x011AF8, 0x000001},
    {0x011C00, 0x011C08, 0x000001}, {0x011C0A, 0x011C2E, 0x000001}, {0x011C2F, 0x011C2F, 0x000008},
    {0x011C30, 0x011C36, 0x000008}, {0x011C38, 0x011C3D, 0x000008}, {0x011C3E, 0x011C3E, 0x000008},
    {0x011C3F, 0x011C3F, 0x000008}, {0x011C40, 0x011C40, 0x000001}, {0x011C41, 0x011C45, 0x000002},
    {0x011C50, 0x011C59, 0x00001C}, {0x011C5A, 0x011C6C, 0x000001}, {0x011C70, 0x011C70, 0x000003},
    {0x011C71, 0x011C71, 0x00000D}, {0x011C72, 0x011C8F, 0x000001}, {0x011C92, 0x011CA7, 0x000008},
    {0x011CA9, 0x011CA9, 0x000008}, {0x011CAA, 0x011CB0, 0x000008}, {0x011CB1, 0x011CB1, 0x000008},
    {0x011CB2, 0x011CB3, 0x000008}, {0x011CB4, 0x011CB4, 0x000008}, {0x011CB5, 0x011CB6, 0x000008},
    {0x011D00, 0x011D06, 0x000001}, {0x011D08, 0x011D09, 0x000001}, {0x011D0B, 0x011D30, 0x000001},
    {0x011D31, 0x011D36, 0x000008}, {0x011D3A, 0x011D3A, 0x000008}, {0x011D3C, 0x011D3D, 0x000008},
    {0x011D3F, 0x011D45, 0x000008}, {0x011D46, 0x011D46, 0x000001}, {0x011D47, 0x011D47, 0x000008},
    {0x011D50, 0x011D59, 0x00001C}, {0x011D60, 0x011D65, 0x000001}, {0x011D67, 0x011D68, 0x000001},
    {0x011D6A, 0x011D89, 0x000001}, {0x011D8A, 0x011D8E, 0x000008}, {0x011D90, 0x011D91, 0x000008},
    {0x011D93, 0x011D94, 0x000008}, {0x011D95, 0x011D95, 0x000008}, {0x011D96, 0x011D96, 0x000008},
    {0x011D97, 0x011D97, 0x000008}, {0x011D98, 0x011D98, 0x000001}, {0x011DA0, 0x011DA9, 0x00001C},
    {0x011EE0, 0x011EF2, 0x000001}, {0x011EF3, 0x011EF4, 0x000008}, {0x011EF5, 0x011EF6, 0x000008},
    {0x011EF7, 0x011EF8, 0x000001}, {0x011FB0, 0x011FB0, 0x000001}, {0x011FC0, 0x011FD4, 0x000001},
    {0x011FD5, 0x011FDC, 0x000001}, {0x011FDD, 0x011FE0, 0x00001E}, {0x011FE1, 0x011FF1, 0x000001},
    {0x011FFF, 0x011FFF, 0x000002}, {0x012000, 0x012399, 0x000001}, {0x012400, 0x01246E, 0x000001},
    {0x012470, 0x012474, 0x000002}, {0x012480, 0x012543, 0x000001}, {0x012F90, 0x012FF0, 0x000001},
    {0x012FF1, 0x012FF2, 0x000001}, {0x013000, 0x013257, 0x000001}, {0x013258, 0x01325A, 0x00001D},
    {0x01325B, 0x01325D, 0x000007}, {0x01325E, 0x013281, 0x000001}, {0x013282, 0x013282, 0x000007},
    {0x013283, 0x013285, 0x000001}, {0x013286, 0x013286, 0x00001D}, {0x013287, 0x013287, 0x000007},
    {0x013288, 0x013288, 0x00001D}, {0x013289, 0x013289, 0x000007}, {0x01328A, 0x013378, 0x000001},
    {0x013379, 0x013379, 0x00001D}, {0x01337A, 0x01337B, 0x000007}, {0x01337C, 0x01342E, 0x000001},
    {0x013430, 0x013436, 0x00000E}, {0x013437, 0x013437, 0x00001D}, {0x013438, 0x013438, 0x000007},
    {0x014400, 0x0145CD, 0x000001}, {0x0145CE, 0x0145CE, 0x00001D}, {0x0145CF, 0x0145CF, 0x000007},
    {0x0145D0, 0x014646, 0x000001}, {0x016800, 0x016A38, 0x000001}, {0x016A40, 0x016A5E, 0x000001},
    {0x016A60, 0x016A69, 0x00001C}, {0x016A6E, 0x016A6F, 0x000002}, {0x016A70, 0x016ABE, 0x000001},
    {0x016AC0, 0x016AC9, 0x00001C}, {0x016AD0, 0x016AED, 0x000001}, {0x016AF0, 0x016AF4, 0x000008},
    {0x016AF5, 0x016AF5, 0x000002}, {0x016B00, 0x016B2F, 0x000001}, {0x016B30, 0x016B36, 0x000008},
    {0x016B37, 0x016B39, 0x000002}, {0x016B3A, 0x016B3B, 0x000001}, {0x016B3C, 0x016B3F, 0x000001},
    {0x016B40, 0x016B43, 0x000001}, {0x016B44, 0x016B44, 0x000002}, {0x016B45, 0x016B45, 0x000001},
    {0x016B50, 0x016B59, 0x00001C}, {0x016B5B, 0x016B61, 0x000001}, {0x016B63, 0x016B77, 0x000001},
    {0x016B7D, 0x016B8F, 0x000001}, {0x016E40, 0x016E7F, 0x000001}, {0x016E80, 0x016E96, 0x000001},
    {0x016E97, 0x016E98, 0x000002}, {0x016E99, 0x016E9A, 0x000001}, {0x016F00, 0x016F4A, 0x000001},
    {0x016F4F, 0x016F4F, 0x000008}, {0x016F50, 0x016F50, 0x000001}, {0x016F51, 0x016F87, 0x000008},
    {0x016F8F, 0x016F92, 0x000008}, {0x016F93, 0x016F9F, 0x000001}, {0x016FE0, 0x016FE1, 0x00001B},
    {0x016FE2, 0x016FE2, 0x00001B}, {0x016FE3, 0x016FE3, 0x00001B}, {0x016FE4, 0x016FE4, 0x00000E},
    {0x016FF0, 0x016FF1, 0x000008}, {0x017000, 0x0187F7, 0x000012}, {0x018800, 0x018AFF, 0x000012},
    {0x018B00, 0x018CD5, 0x000001}, {0x018D00, 0x018D08, 0x000012}, {0x01AFF0, 0x01AFF3, 0x000001},
    {0x01AFF5, 0x01AFFB, 0x000001}, {0x01AFFD, 0x01AFFE, 0x000001}, {0x01B000, 0x01B0FF, 0x000012},
    {0x01B100, 0x01B122, 0x000012}, {0x01B150, 0x01B152, 0x00001B}, {0x01B164, 0x01B167, 0x00001B},
    {0x01B170, 0x01B2FB, 0x000012}, {0x01BC00, 0x01BC6A, 0x000001}, {0x01BC70, 0x01BC7C, 0x000001},
    {0x01BC80, 0x01BC88, 0x000001}, {0x01BC90, 0x01BC99, 0x000001}, {0x01BC9C, 0x01BC9C, 0x000001},
    {0x01BC9D, 0x01BC9E, 0x000008}, {0x01BC9F, 0x01BC9F, 0x000002}, {0x01BCA0, 0x01BCA3, 0x000008},
    {0x01CF00, 0x01CF2D, 0x000008}, {0x01CF30, 0x01CF46, 0x000008}, {0x01CF50, 0x01CFC3, 0x000001},
    {0x01D000, 0x01D0F5, 0x000001}, {0x01D100, 0x01D126, 0x000001}, {0x01D129, 0x01D164, 0x000001},
    {0x01D165, 0x01D166, 0x000008}, {0x01D167, 0x01D169, 0x000008}, {0x01D16A, 0x01D16C, 0x000001},
    {0x01D16D, 0x01D172, 0x000008}, {0x01D173, 0x01D17A, 0x000008}, {0x01D17B, 0x01D182, 0x000008},
    {0x01D183, 0x01D184, 0x000001}, {0x01D185, 0x01D18B, 0x000008}, {0x01D18C, 0x01D1A9, 0x000001},
    {0x01D1AA, 0x01D1AD, 0x000008}, {0x01D1AE, 0x01D1EA, 0x000001}, {0x01D200, 0x01D241, 0x000001},
    {0x01D242, 0x01D244, 0x000008}, {0x01D245, 0x01D245, 0x000001}, {0x01D2E0, 0x01D2F3, 0x000001},
    {0x01D300, 0x01D356, 0x000001}, {0x01D360, 0x01D378, 0x000001}, {0x01D400, 0x01D454, 0x000001},
    {0x01D456, 0x01D49C, 0x000001}, {0x01D49E, 0x01D49F, 0x000001}, {0x01D4A2, 0x01D4A2, 0x000001},
    {0x01D4A5, 0x01D4A6, 0x000001}, {0x01D4A9, 0x01D4AC, 0x000001}, {0x01D4AE, 0x01D4B9, 0x000001},
    {0x01D4BB, 0x01D4BB, 0x000001}, {0x01D4BD, 0x01D4C3, 0x000001}, {0x01D4C5, 0x01D505, 0x000001},
    {0x01D507, 0x01D50A, 0x000001}, {0x01D50D, 0x01D514, 0x000001}, {0x01D516, 0x01D51C, 0x000001},
    {0x01D51E, 0x01D539, 0x000001}, {0x01D53B, 0x01D53E, 0x000001}, {0x01D540, 0x01D544, 0x000001},
    {0x01D546, 0x01D546, 0x000001}, {0x01D54A, 0x01D550, 0x000001}, {0x01D552, 0x01D6A5, 0x000001},
    {0x01D6A8, 0x01D6C0, 0x000001}, {0x01D6C1, 0x01D6C1, 0x000001}, {0x01D6C2, 0x01D6DA, 0x000001},
    {0x01D6DB, 0x01D6DB, 0x000001}, {0x01D6DC, 0x01D6FA, 0x000001}, {0x01D6FB, 0x01D6FB, 0x000001},
    {0x01D6FC, 0x01D714, 0x000001}, {0x01D715, 0x01D715, 0x000001}, {0x01D716, 0x01D734, 0x000001},
    {0x01D735, 0x01D735, 0x000001}, {0x01D736, 0x01D74E, 0x000001}, {0x01D74F, 0x01D74F, 0x000001},
    {0x01D750, 0x01D76E, 0x000001}, {0x01D76F, 0x01D76F, 0x000001}, {0x01D770, 0x01D788, 0x000001},
    {0x01D789, 0x01D789, 0x000001}, {0x01D78A, 0x01D7A8, 0x000001}, {0x01D7A9, 0x01D7A9, 0x000001},
    {0x01D7AA, 0x01D7C2, 0x000001}, {0x01D7C3, 0x01D7C3, 0x000001}, {0x01D7C4, 0x01D7CB, 0x000001},
    {0x01D7CE, 0x01D7FF, 0x00001C}, {0x01D800, 0x01D9FF, 0x000001}, {0x01DA00, 0x01DA36, 0x000008},
    {0x01DA37, 0x01DA3A, 0x000001}, {0x01DA3B, 0x01DA6C, 0x000008}, {0x01DA6D, 0x01DA74, 0x000001},
    {0x01DA75, 0x01DA75, 0x000008}, {0x01DA76, 0x01DA83, 0x000001}, {0x01DA84, 0x01DA84, 0x000008},
    {0x01DA85, 0x01DA86, 0x000001}, {0x01DA87, 0x01DA8A, 0x000002}, {0x01DA8B, 0x01DA8B, 0x000001},
    {0x01DA9B, 0x01DA9F, 0x000008}, {0x01DAA1, 0x01DAAF, 0x000008}, {0x01DF00, 0x01DF09, 0x000001},
    {0x01DF0A, 0x01DF0A, 0x000001}, {0x01DF0B, 0x01DF1E, 0x000001}, {0x01E000, 0x01E006, 0x000008},
    {0x01E008, 0x01E018, 0x000008}, {0x01E01B, 0x01E021, 0x000008}, {0x01E023, 0x01E024, 0x000008},
    {0x01E026, 0x01E02A, 0x000008}, {0x01E100, 0x01E12C, 0x000001}, {0x01E130, 0x01E136, 0x000008},
    {0x01E137, 0x01E13D, 0x000001}, {0x01E140, 0x01E149, 0x00001C}, {0x01E14E, 0x01E14E, 0x000001},
    {0x01E14F, 0x01E14F, 0x000001}, {0x01E290, 0x01E2AD, 0x000001}, {0x01E2AE, 0x01E2AE, 0x000008},
    {0x01E2C0, 0x01E2EB, 0x000001}, {0x01E2EC, 0x01E2EF, 0x000008}, {0x01E2F0, 0x01E2F9, 0x00001C},
    {0x01E2FF, 0x01E2FF, 0x00001F}, {0x01E7E0, 0x01E7E6, 0x000001}, {0x01E7E8, 0x01E7EB, 0x000001},
    {0x01E7ED, 0x01E7EE, 0x000001}, {0x01E7F0, 0x01E7FE, 0x000001}, {0x01E800, 0x01E8C4, 0x000001},
    {0x01E8C7, 0x01E8CF, 0x000001}, {0x01E8D0, 0x01E8D6, 0x000008}, {0x01E900, 0x01E943, 0x000001},
    {0x01E944, 0x01E94A, 0x000008}, {0x01E94B, 0x01E94B, 0x000001}, {0x01E950, 0x01E959, 0x00001C},
    {0x01E95E, 0x01E95F, 0x00001D}, {0x01EC71, 0x01ECAB, 0x000001}, {0x01ECAC, 0x01ECAC, 0x00001E},
    {0x01ECAD, 0x01ECAF, 0x000001}, {0x01ECB0, 0x01ECB0, 0x00001E}, {0x01ECB1, 0x01ECB4, 0x000001},
    {0x01ED01, 0x01ED2D, 0x000001}, {0x01ED2E, 0x01ED2E, 0x000001}, {0x01ED2F, 0x01ED3D, 0x000001},
    {0x01EE00, 0x01EE03, 0x000001}, {0x01EE05, 0x01EE1F, 0x000001}, {0x01EE21, 0x01EE22, 0x000001},
    {0x01EE24, 0x01EE24, 0x000001}, {0x01EE27, 0x01EE27, 0x000001}, {0x01EE29, 0x01EE32, 0x000001},
    {0x01EE34, 0x01EE37, 0x000001}, {0x01EE39, 0x01EE39, 0x000001}, {0x01EE3B, 0x01EE3B, 0x000001},
    {0x01EE42, 0x01EE42, 0x000001}, {0x01EE47, 0x01EE47, 0x000001}, {0x01EE49, 0x01EE49, 0x000001},
    {0x01EE4B, 0x01EE4B, 0x000001}, {0x01EE4D, 0x01EE4F, 0x000001}, {0x01EE51, 0x01EE52, 0x000001},
    {0x01EE54, 0x01EE54, 0x000001}, {0x01EE57, 0x01EE57, 0x000001}, {0x01EE59, 0x01EE59, 0x000001},
    {0x01EE5B, 0x01EE5B, 0x000001}, {0x01EE5D, 0x01EE5D, 0x000001}, {0x01EE5F, 0x01EE5F, 0x000001},
    {0x01EE61, 0x01EE62, 0x000001}, {0x01EE64, 0x01EE64, 0x000001}, {0x01EE67, 0x01EE6A, 0x000001},
    {0x01EE6C, 0x01EE72, 0x000001}, {0x01EE74, 0x01EE77, 0x000001}, {0x01EE79, 0x01EE7C, 0x000001},
    {0x01EE7E, 0x01EE7E, 0x000001}, {0x01EE80, 0x01EE89, 0x000001}, {0x01EE8B, 0x01EE9B, 0x000001},
    {0x01EEA1, 0x01EEA3, 0x000001}, {0x01EEA5, 0x01EEA9, 0x000001}, {0x01EEAB, 0x01EEBB, 0x000001},
    {0x01EEF0, 0x01EEF1, 0x000001}, {0x01F000, 0x01F02B, 0x000012}, {0x01F02C, 0x01F02F, 0x000012},
    {0x01F030, 0x01F093, 0x000012}, {0x01F094, 0x01F09F, 0x000012}, {0x01F0A0, 0x01F0AE, 0x000012},
    {0x01F0AF, 0x01F0B0, 0x000012}, {0x01F0B1, 0x01F0BF, 0x000012}, {0x01F0C0, 0x01F0C0, 0x000012},
    {0x01F0C1, 0x01F0CF, 0x000012}, {0x01F0D0, 0x01F0D0, 0x000012}, {0x01F0D1, 0x01F0F5, 0x000012},
    {0x01F0F6, 0x01F0FF, 0x000012}, {0x01F100, 0x01F10C, 0x000001}, {0x01F10D, 0x01F10F, 0x000012},
    {0x01F110, 0x01F12D, 0x000001}, {0x01F12E, 0x01F12F, 0x000001}, {0x01F130, 0x01F169, 0x000001},
    {0x01F16A, 0x01F16C, 0x000001}, {0x01F16D, 0x01F16F, 0x000012}, {0x01F170, 0x01F1AC, 0x000001},
    {0x01F1AD, 0x01F1AD, 0x000012}, {0x01F1AE, 0x01F1E5, 0x000012}, {0x01F1E6, 0x01F1FF, 0x000021},
    {0x01F200, 0x01F202, 0x000012}, {0x01F203, 0x01F20F, 0x000012}, {0x01F210, 0x01F23B, 0x000012},
    {0x01F23C, 0x01F23F, 0x000012}, {0x01F240, 0x01F248, 0x000012}, {0x01F249, 0x01F24F, 0x000012},
    {0x01F250, 0x01F251, 0x000012}, {0x01F252, 0x01F25F, 0x000012}, {0x01F260, 0x01F265, 0x000012},
    {0x01F266, 0x01F2FF, 0x000012}, {0x01F300, 0x01F384, 0x000012}, {0x01F385, 0x01F385, 0x00000B},
    {0x01F386, 0x01F39B, 0x000012}, {0x01F39C, 0x01F39D, 0x000001}, {0x01F39E, 0x01F3B4, 0x000012},
    {0x01F3B5, 0x01F3B6, 0x000001}, {0x01F3B7, 0x01F3BB, 0x000012}, {0x01F3BC, 0x01F3BC, 0x000001},
    {0x01F3BD, 0x01F3C1, 0x000012}, {0x01F3C2, 0x01F3C4, 0x00000B}, {0x01F3C5, 0x01F3C6, 0x000012},
    {0x01F3C7, 0x01F3C7, 0x00000B}, {0x01F3C8, 0x01F3C9, 0x000012}, {0x01F3CA, 0x01F3CC, 0x00000B},
    {0x01F3CD, 0x01F3FA, 0x000012}, {0x01F3FB, 0x01F3FF, 0x00000C}, {0x01F400, 0x01F441, 0x000012},
    {0x01F442, 0x01F443, 0x00000B}, {0x01F444, 0x01F445, 0x000012}, {0x01F446, 0x01F450, 0x00000B},
    {0x01F451, 0x01F465, 0x000012}, {0x01F466, 0x01F478, 0x00000B}, {0x01F479, 0x01F47B, 0x000012},
    {0x01F47C, 0x01F47C, 0x00000B}, {0x01F47D, 0x01F480, 0x000012}, {0x01F481, 0x01F483, 0x00000B},
    {0x01F484, 0x01F484, 0x000012}, {0x01F485, 0x01F487, 0x00000B}, {0x01F488, 0x01F48E, 0x000012},
    {0x01F48F, 0x01F48F, 0x00000B}, {0x01F490, 0x01F490, 0x000012}, {0x01F491, 0x01F491, 0x00000B},
    {0x01F492, 0x01F49F, 0x000012}, {0x01F4A0, 0x01F4A0, 0x000001}, {0x01F4A1, 0x01F4A1, 0x000012},
    {0x01F4A2, 0x01F4A2, 0x000001}, {0x01F4A3, 0x01F4A3, 0x000012}, {0x01F4A4, 0x01F4A4, 0x000001},
    {0x01F4A5, 0x01F4A9, 0x000012}, {0x01F4AA, 0x01F4AA, 0x00000B}, {0x01F4AB, 0x01F4AE, 0x000012},
    {0x01F4AF, 0x01F4AF, 0x000001}, {0x01F4B0, 0x01F4B0, 0x000012}, {0x01F4B1, 0x01F4B2, 0x000001},
    {0x01F4B3, 0x01F4FF, 0x000012}, {0x01F500, 0x01F506, 0x000001}, {0x01F507, 0x01F516, 0x000012},
    {0x01F517, 0x01F524, 0x000001}, {0x01F525, 0x01F531, 0x000012}, {0x01F532, 0x01F549, 0x000001},
    {0x01F54A, 0x01F573, 0x000012}, {0x01F574, 0x01F575, 0x00000B}, {0x01F576, 0x01F579, 0x000012},
    {0x01F57A, 0x01F57A, 0x00000B}, {0x01F57B, 0x01F58F, 0x000012}, {0x01F590, 0x01F590, 0x00000B},
    {0x01F591, 0x01F594, 0x000012}, {0x01F595, 0x01F596, 0x00000B}, {0x01F597, 0x01F5D3, 0x000012},
    {0x01F5D4, 0x01F5DB, 0x000001}, {0x01F5DC, 0x01F5F3, 0x000012}, {0x01F5F4, 0x01F5F9, 0x000001},
    {0x01F5FA, 0x01F5FF, 0x000012}, {0x01F600, 0x01F644, 0x000012}, {0x01F645, 0x01F647, 0x00000B},
    {0x01F648, 0x01F64A, 0x000012}, {0x01F64B, 0x01F64F, 0x00000B}, {0x01F650, 0x01F675, 0x000001},
    {0x01F676, 0x01F678, 0x000020}, {0x01F679, 0x01F67B, 0x00001B}, {0x01F67C, 0x01F67F, 0x000001},
    {0x01F680, 0x01F6A2, 0x000012}, {0x01F6A3, 0x01F6A3, 0x00000B}, {0x01F6A4, 0x01F6B3, 0x000012},
    {0x01F6B4, 0x01F6B6, 0x00000B}, {0x01F6B7, 0x01F6BF, 0x000012}, {0x01F6C0, 0x01F6C0, 0x00000B},
    {0x01F6C1, 0x01F6CB, 0x000012}, {0x01F6CC, 0x01F6CC, 0x00000B}, {0x01F6CD, 0x01F6D7, 0x000012},
    {0x01F6D8, 0x01F6DC, 0x000012}, {0x01F6DD, 0x01F6EC, 0x000012}, {0x01F6ED, 0x01F6EF, 0x000012},
    {0x01F6F0, 0x01F6FC, 0x000012}, {0x01F6FD, 0x01F6FF, 0x000012}, {0x01F700, 0x01F773, 0x000001},
    {0x01F774, 0x01F77F, 0x000012}, {0x01F780, 0x01F7D4, 0x000001}, {0x01F7D5, 0x01F7D8, 0x000012},
    {0x01F7D9, 0x01F7DF, 0x000012}, {0x01F7E0, 0x01F7EB, 0x000012}, {0x01F7EC, 0x01F7EF, 0x000012},
    {0x01F7F0, 0x01F7F0, 0x000012}, {0x01F7F1, 0x01F7FF, 0x000012}, {0x01F800, 0x01F80B, 0x000001},
    {0x01F80C, 0x01F80F, 0x000012}, {0x01F810, 0x01F847, 0x000001}, {0x01F848, 0x01F84F, 0x000012},
    {0x01F850, 0x01F859, 0x000001}, {0x01F85A, 0x01F85F, 0x000012}, {0x01F860, 0x01F887, 0x000001},
    {0x01F888, 0x01F88F, 0x000012}, {0x01F890, 0x01F8AD, 0x000001}, {0x01F8AE, 0x01F8AF, 0x000012},
    {0x01F8B0, 0x01F8B1, 0x000012}, {0x01F8B2, 0x01F8FF, 0x000012}, {0x01F900, 0x01F90B, 0x000001},
    {0x01F90C, 0x01F90C, 0x00000B}, {0x01F90D, 0x01F90E, 0x000012}, {0x01F90F, 0x01F90F, 0x00000B},
    {0x01F910, 0x01F917, 0x000012}, {0x01F918, 0x01F91F, 0x00000B}, {0x01F920, 0x01F925, 0x000012},
    {0x01F926, 0x01F926, 0x00000B}, {0x01F927, 0x01F92F, 0x000012}, {0x01F930, 0x01F939, 0x00000B},
    {0x01F93A, 0x01F93B, 0x000012}, {0x01F93C, 0x01F93E, 0x00000B}, {0x01F93F, 0x01F976, 0x000012},
    {0x01F977, 0x01F977, 0x00000B}, {0x01F978, 0x01F9B4, 0x000012}, {0x01F9B5, 0x01F9B6, 0x00000B},
    {0x01F9B7, 0x01F9B7, 0x000012}, {0x01F9B8, 0x01F9B9, 0x00000B}, {0x01F9BA, 0x01F9BA, 0x000012},
    {0x01F9BB, 0x01F9BB, 0x00000B}, {0x01F9BC, 0x01F9CC, 0x000012}, {0x01F9CD, 0x01F9CF, 0x00000B},
    {0x01F9D0, 0x01F9D0, 0x000012}, {0x01F9D1, 0x01F9DD, 0x00000B}, {0x01F9DE, 0x01F9FF, 0x000012},
    {0x01FA00, 0x01FA53, 0x000001}, {0x01FA54, 0x01FA5F, 0x000012}, {0x01FA60, 0x01FA6D, 0x000012},
    {0x01FA6E, 0x01FA6F, 0x000012}, {0x01FA70, 0x01FA74, 0x000012}, {0x01FA75, 0x01FA77, 0x000012},
    {0x01FA78, 0x01FA7C, 0x000012}, {0x01FA7D, 0x01FA7F, 0x000012}, {0x01FA80, 0x01FA86, 0x000012},
    {0x01FA87, 0x01FA8F, 0x000012}, {0x01FA90, 0x01FAAC, 0x000012}, {0x01FAAD, 0x01FAAF, 0x000012},
    {0x01FAB0, 0x01FABA, 0x000012}, {0x01FABB, 0x01FABF, 0x000012}, {0x01FAC0, 0x01FAC2, 0x000012},
    {0x01FAC3, 0x01FAC5, 0x00000B}, {0x01FAC6, 0x01FACF, 0x000012}, {0x01FAD0, 0x01FAD9, 0x000012},
    {0x01FADA, 0x01FADF, 0x000012}, {0x01FAE0, 0x01FAE7, 0x000012}, {0x01FAE8, 0x01FAEF, 0x000012},
    {0x01FAF0, 0x01FAF6, 0x00000B}, {0x01FAF7, 0x01FAFF, 0x000012}, {0x01FB00, 0x01FB92, 0x000001},
    {0x01FB94, 0x01FBCA, 0x000001}, {0x01FBF0, 0x01FBF9, 0x00001C}, {0x01FC00, 0x01FFFD, 0x000012},
    {0x020000, 0x02A6DF, 0x000012}, {0x02A6E0, 0x02A6FF, 0x000012}, {0x02A700, 0x02B738, 0x000012},
    {0x02B739, 0x02B73F, 0x000012}, {0x02B740, 0x02B81D, 0x000012}, {0x02B81E, 0x02B81F, 0x000012},
    {0x02B820, 0x02CEA1, 0x000012}, {0x02CEA2, 0x02CEAF, 0x000012}, {0x02CEB0, 0x02EBE0, 0x000012},
    {0x02EBE1, 0x02F7FF, 0x000012}, {0x02F800, 0x02FA1D, 0x000012}, {0x02FA1E, 0x02FA1F, 0x000012},
    {0x02FA20, 0x02FFFD, 0x000012}, {0x030000, 0x03134A, 0x000012}, {0x03134B, 0x03FFFD, 0x000012},
    {0x0E0001, 0x0E0001, 0x000008}, {0x0E0020, 0x0E007F, 0x000008}, {0x0E0100, 0x0E01EF, 0x000008},
    {0x0F0000, 0x0FFFFD, 0x000001}, {0x100000, 0x10FFFD, 0x000001},
}};
}  // namespace arg_router::utility::utf8
